#include <dni_nvram.h>
#include <dni_shutils.h>

#include <unistd.h> //For execv function

int sem_id;		/*semaphore identifier*/
void *pointer;	/*pointer to current poisition in share memory*/
void *ptr_start;	/*pointer to start poisition of share memory*/
static char nullstr[1];         /* zero length string */
char shm_flag=1;	/*check to attach share memory */
char realloc_flag=0;		/*prevent re_alloc function causing deadlock*/

int *var_start;			/*start address of hash table*/

//static inline void* ckmalloc (int sz)          { return xmalloc(sz);     }
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
};
inline void set_sem(int semid)
{
	union semun sem_union;
	sem_union.val=1;
	if (semctl(semid,0,SETVAL,sem_union)==-1)
		printf ("set sem error\n");
}
inline void sem_up(int semid)
{
	struct sembuf sem_b;

	sem_b.sem_num=0;
	sem_b.sem_op=-1;
	sem_b.sem_flg=SEM_UNDO;
	if (semop(semid,&sem_b,1)==-1)
		printf ("semp p process error\n");
}

inline void sem_down(int semid)
{
	struct sembuf sem_b;

	sem_b.sem_num=0;
	sem_b.sem_op=1;
	sem_b.sem_flg=SEM_UNDO;
	if (semop(semid,&sem_b,1)==-1)
		printf ("semp v process error\n");
}

inline void detach_shm(void)
{
	/* detach the share memory*/
	if (shmdt(ptr_start)==-1)
		perror ("shmdt1");
	shm_flag=1;
}

void attach_share_memory()
{
	int shmid;
	/*create share memory*/
	shmid=shmget((key_t)NVRAMKEY,(size_t)SHARESIZE,IPC_CREAT|IPC_EXCL|0666);
	if(shmid==-1)
	{
//		perror("shmget 1");
		if ((shmid=shmget((key_t)NVRAMKEY,(size_t)SHARESIZE,0666))==-1)
			perror ("attach_share_memory failed, shmid is -1");
	}
	pointer=shmat(shmid,(void *)0, 0);
	
	shm_flag=0;
	
	ptr_start=pointer;
	/*set semaphore*/
	if ((sem_id = semget(NVRAMKEY, 1, IPC_CREAT | 0666)) == -1)
	{
		perror("semget error");
	}
	
	nvram_init();
}

void *get_curr_pos()
{
	char *tmp=(char *)ptr_start;
	int offset;
	/*skip MAGIC_ID*/
	tmp+=6;
	
	offset=atol(tmp);

	return (void *)(ptr_start+offset);
}

void *ckmalloc(size_t size)
{
	char *p,*ptr;
	char *tmp=ptr_start;
	char line[8];
	int i;
	ptr=(char *)get_curr_pos();
	/* add 1 byte for NULL character*/
	p=ptr+size+1;
	/*if out of share memory then re-allocate share memory*/
	if (p>=tmp+SHARESIZE)
	{
		//printf ("out of memory...realloc memory....\n");
		re_alloc();
		ptr=(char *)get_curr_pos();
		p=ptr+size+1;
		/*if still out of share memory then return NULL*/
		if (p>=tmp+SHARESIZE)
		{
			//printf ("out of memory...can't alloc memory...\n");
			return NULL;
		}
	}
	
	tmp+=(strlen(MAGIC_ID)+1);
	i=atol(tmp);
	i+=(size+1);
	sprintf(line,"%d-",i);
	strcpy(tmp,line);
//	p=(char *)pointer;
	pointer=(void *)p;

	return (void *)ptr;
}

inline void *get_addr(int offset)
{
//	printf ("get_addr :: offset %x\n",offset);
	char *ptr;
	ptr=(char *)ptr_start;
	if (offset > (SHARESIZE-1)||offset<=0)
		return NULL;
	return (void *)(ptr+offset);
}

inline int get_offset(void *ptr)
{
	int i;
	char *tmp1,*tmp2;
	tmp1=(char *)ptr;
	tmp2=(char *)ptr_start;
	i=tmp1-tmp2;
//	printf ("get_offset :: %x\n",i);
	if ((i>0)&&(i<SHARESIZE))
		return i;
	else
		return 0;
}

void
clear_end(char *value)
{
	while(*value)
	{
		if((*value=='\r')||(*value=='\n')||(*value=='\0'))
		{
			*value='\0';
			break;
		}
		value++;
	}
	return;
}

void
hashvar(const char *p,int **ptr)
{
	unsigned int hashval;

	hashval = ((unsigned char) *p) << 4;
	while (*p && *p != '=')
		hashval += (unsigned char) *p++;
	*ptr=&var_start[hashval % VTABSIZE];
//	printf ("hashvar :: ptr[%x]\n",(int)ptr);
	return;
}

static int
varequal(const char *p, const char *q)
{
	if (p==NULL)
		return 0;
	while (*p == *q++) {
		if (*p++ == '=')
			return 1;
	}
	if (*p == '=' && *(q - 1) == '\0')
		return 1;
	return 0;
}

static struct varinit *
findvar(struct varinit *vp, const char *name)
{
	for (; vp; vp = (struct varinit *)get_addr((vp->next_offset))) {
		if (varequal((char *)get_addr(vp->name_offset), name)) {
			break;
		}
	}
	return vp;
}

char *
nvram_get(name)
	const char *name;
{
	int *offset;
	struct varinit *v,*vp;
	if (shm_flag)
		attach_share_memory();
	INTOFF;
	hashvar(name,&offset);
	vp=(struct varinit*)get_addr(*offset);
	if ((v = findvar(vp, name))) {
		INTON;
//		printf("nvram_get found %s :: %s\n",name,(char *)get_addr(v->text_offset));
		return (char *)get_addr(v->text_offset);
	}
//	printf ("nvram_get found no var\n");
	INTON;
	return NULL;
}


int
nvram_set(name, val)
	const char *name, *val;
{
//	const char *p;
	int len,*ptr;
	int namelen;
	char *line,*nameeq;
	int name_offset,value_offset;
	struct varinit *vp, *vpp;

	if (shm_flag)
		attach_share_memory();
	
//	p = name;
//	p=strchr(p,'\0');
	namelen = strlen(name);
	if (val)
		len  = strlen(val);
	else
		len=0;

	line=malloc(namelen+len+2);

	if (line==NULL)
	{
		//printf("nvram_set :: malloc error...\n");
		return ERR_NO_MEM;
	}
	
	sprintf(line,"%s=%s",name,val);
	if (!realloc_flag)
		INTOFF;
	hashvar(line,&ptr);
	vpp=(struct varinit*)get_addr(*ptr);
	vp = findvar(vpp, line);

	if (vp)
	{
//		cprintf ("nvram_set found var :: set %s=%s\n",name,val);
		char *tmp=(char *)get_addr(vp->text_offset);
		/*if the length of new value is larger than old one, it will allocate a new memory area*/
		if (strlen(tmp)>=len)
		{
			if (len)
			{
				memcpy(tmp , val, len);
				tmp[len]='\0';
			}
//			else
				tmp[len]='\0';
		}
		else
		{
			nameeq = ckmalloc(len);
			if (nameeq==NULL)
			{
				if (!realloc_flag)
					INTON;
				return ERR_NO_MEM;
			}
			value_offset=get_offset(nameeq);

			if (len) 
				memcpy(nameeq , val, len);
//			else 
				nameeq[len] = '\0';
			vp->text_offset = value_offset;
		}
	}
	else
	{
		/*allocate memory for value and get the offset*/
		nameeq = ckmalloc(len);
		if (nameeq==NULL)
		{
			if (!realloc_flag)
				INTON;
			return ERR_NO_MEM;
		}
		value_offset=get_offset(nameeq);
		if (len) 
			memcpy(nameeq , val, len);
//		else 
			nameeq[len] = '\0';

		/*allocate memory for name and get the offset*/
		len = namelen + 2;              /* 2 is space for '=' and '\0' */
		nameeq=ckmalloc(len);
		if (nameeq==NULL)
		{
			if (!realloc_flag)
				INTON;
			return ERR_NO_MEM;
		}
		name_offset=get_offset(nameeq);
		memcpy(nameeq, name, namelen);
		nameeq[namelen] = '=';

		/*allocate memory for varinit and get the offset*/
		vp = ckmalloc(sizeof (*vp));
		if (vp==NULL)
		{
			if (!realloc_flag)
				INTON;
			return ERR_NO_MEM;
		}
		vp->name_offset = name_offset;
		vp->text_offset = value_offset;
		vp->next_offset = get_offset(vpp);
		
		/*record current varinit offset to the entry of hash table*/
		*ptr = get_offset(vp);
//		cprintf ("nvram_set found no var :: set %s=%s\n",(char *)get_addr(name_offset),
//			(char *)get_addr(value_offset));
	}
	if (!realloc_flag)
		INTON;
	free(line);
	return 0;

}

int
nvram_unset(const char *s)
{
	/*pointer to current varinit struct*/
	struct varinit *vp;
	/*pointer to previous varinit struct*/
	struct varinit *vpp;
	int *ptr;

	if (shm_flag)
		attach_share_memory();

	INTOFF;

	hashvar(s,&ptr);
	vp=(struct varinit *)get_addr(*ptr);

	if (varequal((char *)get_addr(vp->name_offset), s))
	{
		*ptr=vp->next_offset;
	}
	else
	{
		for (vpp=vp; vp; vpp=vp,vp = (struct varinit *)get_addr((vp->next_offset))) {
			if (varequal((char *)get_addr(vp->name_offset), s)) {
				break;
			}
		}

		if (vp)
		{
			vpp->next_offset=vp->next_offset;
		}
	}
	INTON;

	return (0);
}

extern void
nvram_show()
{
 
	struct varinit *vp;
	int i=0;

	if (shm_flag)
		attach_share_memory();
	
	INTOFF;
	//printf ("call nvram_show\n");
	for ( i=0; i < VTABSIZE ; i++)
	{
		vp = (struct varinit *)get_addr(var_start[i]);

		for ( ;vp ; vp = (struct varinit*)get_addr(vp->next_offset))
		{
			char *name,*value;
			int len;
			name=(char *)get_addr(vp->name_offset);
			value=(char *)get_addr(vp->text_offset);
			len = strlen(value);
			printf("%s%s\n", name,value);
		}
	}
	INTON;

}
int
nvram_commit()
{
	struct varinit *vp;
	int i;
	const char *sep = nullstr;
	FILE *fp=NULL;
	char *argv[]={ "commit",NULL,NULL,(char *)0};

	if (shm_flag)
		attach_share_memory();

	//fp=fopen(DEVICE_PATH,"w+");
	fp=fopen(TMP_FILE_PATH, "w+");
	if (fp==NULL)
	{
		printf ("file open error\n");
		return -1;
	}

	INTOFF;
//	printf ("call nvram commit\n");
	
	for ( i=0; i < VTABSIZE ; i++) {
		vp = (struct varinit *)get_addr(var_start[i]);
		for (; vp ; vp = (struct varinit *)get_addr(vp->next_offset)) {
			char *name,*value;
			name=(char *)get_addr(vp->name_offset);
			value=(char *)get_addr(vp->text_offset);
			fprintf(fp,"%s%s%s\n", sep, name, value);
		}
	}
	INTON;
	fclose(fp);
/***********For Timo Test Only**********/
	execv("/etc/nvram/commit", argv);
/*****End*****/
	return 0;
}

int nvram_getall(char *buf, int count)
{
	int i,len=0;
	char *line;
	struct varinit *vp;
	
	if (count==0)
		return 0;
	
	if (shm_flag)
		attach_share_memory();
	if (!realloc_flag)
		INTOFF;
	for (i=0; i < VTABSIZE ; i++)
	{
		vp = (struct varinit *)get_addr(var_start[i]);
		for (; vp ; vp = (struct varinit *)get_addr(vp->next_offset)) {
			char *name,*value;
			name=(char *)get_addr(vp->name_offset);
			value=(char *)get_addr(vp->text_offset);
			line=malloc(strlen(name)+strlen(value)+1);
			if (line==NULL)
				goto END;
			sprintf(line,"%s%s", name, value);
			strcat(buf,line);
			buf+=strlen(line);
			*buf='\0';
			buf++;
			len+=(strlen(line)+1);
			free(line);
		}
	}
END:
	if (!realloc_flag)
		INTON;
	return len;
}

/*
 * Match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is string equal
 *		to match or FALSE otherwise
 */
int
nvram_match(char *name, char *match) {
	const char *value = BCMINIT(nvram_get)(name);
	return (value && !strcmp(value, match));
}

/*
 * Inversely match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is not string
 *		equal to invmatch or FALSE otherwise
 */
int
nvram_invmatch(char *name, char *invmatch) {
	const char *value = BCMINIT(nvram_get)(name);
	return (value && strcmp(value, invmatch));
}


void dump_mem(void *p,int len)
{
	int i;
	char *ptr;
	ptr=(char *)p;
	for (i=0;i<len;i++)
		printf ("%c",ptr[i]);
	printf ("\n");
}

void init_share_ptr()
{
	char *p=(char *)ptr_start,*ptr;
	int i;

	/*add 1 for NULL character*/
	p+=(strlen(MAGIC_ID)+1);
	ptr=p;
	for (;;ptr++)
		if ((*ptr>='0')&&(*ptr<='9'))
			break;
	/*get current offset*/
	i=atol(ptr);
	ptr=(char *)ptr_start;
	pointer=ptr+i;

//	printf ("get current offset :: %d\n",i);
	p+=8; // 8 bytes for restoring current offset for pointer

	/*hash table is allocated after MAGIC_ID and current offset (8 bytes)*/
	var_start=(int*)p;
	
}

int
nvram_init()
{
	char *value,*name,*p;
	char line[2048];
//	int varid;
	FILE *fp_ptr=NULL;

//	printf ("pointer :: %x\n",(int) pointer);
//	printf("nvram_init\n");
	if (strncmp(pointer,MAGIC_ID,strlen(MAGIC_ID))!=0)
	{
//		printf("start new process...\n");
		memset(pointer,0,SHARESIZE);
		set_sem(sem_id);
//		printf ("set semp ok...\n");
		INTOFF;
		p=ckmalloc(strlen(MAGIC_ID));
		strcpy(p,MAGIC_ID);
		//8 bytes for restoring current offset of pointer
		//ckmalloc will add 1 byte
		ckmalloc(7);
		var_start=ckmalloc(sizeof(int)*VTABSIZE);
		INTON;
//		if (varid==-1)
//			printf ("create share memory(var) error\n");
		
		/*open NVRAM file*/
		fp_ptr=fopen(TMP_FILE_PATH,"r+");
		if (fp_ptr==NULL)
		{
			//printf ("open %s error..\n",DEVICE_PATH);
		}
		else
		{
			while(fgets(line,sizeof(line),fp_ptr))
			{
				value=line;
				name=line;
				strsep(&value,"=");
				if (value)
				{
					clear_end(value);
					nvram_set(name,value);
				}
			}
		}
	}
	else
	{
//		printf ("found MAGIC ID\n");
		init_share_ptr();
	}

	return 0;
}

void
nvram_clean(void)
{
	/*erase MAGIC_ID, current pointer offset, and hash table*/
	int len=(strlen(MAGIC_ID)+1)+8+(sizeof(struct varinit *)*VTABSIZE+1);
	memset(ptr_start,0,len);
}

void re_alloc(void)
{
	char *line,*value,*name;
	char *buf,*ptr;
	int len,total_len;
	if (shm_flag)
		attach_share_memory();
	//printf ("before realloc...poniter [%x]...\n",(int)pointer);
	buf=malloc(SHARESIZE);
	if (buf==NULL)
	{
		perror("malloc");
		return ;
	}
	/*save current settings to buf*/
	realloc_flag=1;
	total_len=nvram_getall(buf,SHARESIZE);
	pointer=ptr_start;

	nvram_clean();

	/*start initial data*/
	ptr=ckmalloc(strlen(MAGIC_ID));
	strcpy(ptr,MAGIC_ID);
	ckmalloc(7);
	var_start=ckmalloc(sizeof(int)*VTABSIZE);
	
	ptr=buf;
	/*restore variables*/
	while(total_len)
	{
		len=strlen(ptr);
		line=malloc(len);
		if (line==NULL)
		{
			//printf("re_alloc :: malloc error....\n");
			break;
		}
		sprintf(line,"%s",ptr);
//		len=strlen(line);
		if (len==0)
			break;
		value=line;
		name=line;
		strsep(&value,"=");
		if (value)
		{
			clear_end(value);
			nvram_set(name,value);
		}
		/*move to next data*/
		len++;
		ptr+=len;
		total_len-=len;
		free(line);
	}
	realloc_flag=0;
	
	free(buf);
	//printf ("after realloc...poniter [%x]...\n",(int)pointer);
}

