/*
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>															
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if_arp.h>
#include <signal.h>

#include <re865x.h>
#include <board.h>

#include <rtl8651_tblDrv.h>

#define _PRE_TYPE_SET

#include <dni_nvram.h>
#include <dni_shutils.h>
//#include <nvparse.h>
#include "rc.h"

extern void AddLocalServer( char *pstrWanLastIP, char *pstrWanNowIP, int nNoDMZ );
extern void DelLocalServer( char *pstrWanIP  );

#define cprintf		printf

#define RTL8651_LL_PPPOE			0x02

static int
add_routes(char *prefix, char *var, char *ifname)
{
	char word[80], *next;
	char *ipaddr, *netmask, *gateway, *metric;
	char tmp[100];
	
	foreach(word, nvram_safe_get(strcat_r(prefix, var, tmp)), next) {
		dprintf("add %s\n", word);
		
		netmask = word;
		ipaddr = strsep(&netmask, ":");
		if (!ipaddr || !netmask)
			continue;
		gateway = netmask;
		netmask = strsep(&gateway, ":");
		if (!netmask || !gateway)
			continue;
		metric = gateway;
		gateway = strsep(&metric, ":");
		if (!gateway || !metric)
			continue;

		dprintf("add %s\n", ifname);
		
		route_add(ifname, atoi(metric) + 1, ipaddr, gateway, netmask);
	}

	return 0;
}

static int
del_routes(char *prefix, char *var, char *ifname)
{
	char word[80], *next;
	char *ipaddr, *netmask, *gateway, *metric;
	char tmp[100];
	
	foreach(word, nvram_safe_get(strcat_r(prefix, var, tmp)), next) {
		dprintf("add %s\n", word);
		
		netmask = word;
		ipaddr = strsep(&netmask, ":");
		if (!ipaddr || !netmask)
			continue;
		gateway = netmask;
		netmask = strsep(&gateway, ":");
		if (!netmask || !gateway)
			continue;
		metric = gateway;
		gateway = strsep(&metric, ":");
		if (!gateway || !metric)
			continue;
		
		dprintf("add %s\n", ifname);
		
		route_del(ifname, atoi(metric) + 1, ipaddr, gateway, netmask);
	}

	return 0;
}

static int
add_lan_routes(char *lan_ifname)
{
	return add_routes("lan_", "route", lan_ifname);
}

static int
del_lan_routes(char *lan_ifname)
{
	return del_routes("lan_", "route", lan_ifname);
}

void
start_lan(void)
{
	//char *lan_ifname = nvram_safe_get("lan_ifname");
	char lan_ifname = "eth1";
	//char name[80], *next;
	char tmp[100];
	int s;
	struct ifreq ifr;
	//char lan_ifnames[32];
	char ipaddr[20], ipmask[20];

	ifconfig(lan_ifname, IFUP, NULL, NULL);

	/* Get current LAN hardware address */
	if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
		char eabuf[32];
		strncpy(ifr.ifr_name, lan_ifname, IFNAMSIZ);
		if (ioctl(s, SIOCGIFHWADDR, &ifr) == 0)
			nvram_set("lan_hwaddr", ether_etoa(ifr.ifr_hwaddr.sa_data, eabuf));
		close(s);
	}

	/* Bring up and configure LAN interface */
	strcpy(ipaddr, nvram_safe_get("lan_ipaddr"));
	strcpy(ipmask, nvram_safe_get("lan_netmask"));	
	ifconfig(lan_ifname, IFUP,
		 //nvram_safe_get("lan_ipaddr"), nvram_safe_get("lan_netmask"));
		 ipaddr, ipmask);

			
	/* Install lan specific static routes */
	//add_lan_routes(lan_ifname);
	
}

void
stop_lan(void)
{
	char *lan_ifname = "eth1";
	char name[80], *next, signal[] = "XXXXXXXX";

	dprintf("%s\n", lan_ifname);

	/* Remove static routes */
	del_lan_routes(lan_ifname);

	/* Bring down LAN interface */
	ifconfig(lan_ifname, 0, NULL, NULL);

	dprintf("done\n");
}

static int
add_wan_routes(char *wan_ifname)
{

	// Add wan private routes
	return add_routes("wan_", "route", wan_ifname);
}

static int
del_wan_routes(char *wan_ifname)
{
	return del_routes("wan_", "route", wan_ifname);
}

void run_ripd()
{
	char *argv[]={
		"ripd",
		NULL
		};
	int pid;

	if (strcmp(nvram_get("rip_enable"),"0")>0){
		if (task_is_running("ripd")!=1)
			_eval(argv, NULL, 0, &pid);
		
	}
}

int
stop_ripd(void)
{
	int  ret;
	unlink("/var/run/ripd.pid");
	ret = eval("killall", "ripd");
	return ret;
}


#define MAX_NVPARSE 1

extern romeCfgParam_t * pRomeCfgParam;

void 
system_restart(void)
{
	/* rome restart */
	send_signal("/var/run/boa.pid",72);		/* sysReinit */	
	dprintf("done\n");	
}	

void
start_wan(void)
{
	char *wan_ifname="eth0";
	char *wan_proto;
	int unit;
	char tmp[100], prefix[] = "wanXXXXXXXXXX_", signal[] = "XXXX";
	char eabuf[32];
	int s;
	struct ifreq ifr;
	pid_t pid;
	char strGateway[20];

	printf("%s\n", __FUNCTION__);
	send_signal("/var/run/boa.pid",71);		/* start wan */	
	dprintf("done\n");	

}

void
stop_wan(void)
{

	char wan_ifname="eth0", *next, signal[] = "XXXX";
	char str_ethaddr1[32];  //org_wan_mac

	send_signal("/var/run/boa.pid",70);		/* stop wan */
	dprintf("done\n");
}

int
add_ns(char *wan_ifname)
{
	FILE *fp;
	char tmp[100], prefix[] = "wan0_";
	char word[100], *next;
	char line[100], i;

	/* Open resolv.conf to read */
	if (!(fp = fopen("/var/resolv.conf", "r+"))) {
		perror("/var/resolv.conf");
		return errno;
	}
	/* Append only those not in the original list */
	//foreach(word, nvram_safe_get(strcat_r(prefix, "dns", tmp)), next) {
	for (i = 0; i < 2; i++)	{
		sprintf(tmp, "wan0_dns%d", i);
		strcpy(word, nvram_safe_get(tmp));
		if (word[0] == 0)
			continue;
		fseek(fp, 0, SEEK_SET);
		while (fgets(line, sizeof(line), fp)) {
			char *token = strtok(line, " \t\n");

			if (!token || strcmp(token, "nameserver") != 0)
				continue;
			if (!(token = strtok(NULL, " \t\n")))
				continue;

			if (!strcmp(token, word))
				break;
		}
		if (feof(fp))
			fprintf(fp, "nameserver %s\n", word);
	}
	fclose(fp);

	/* notify dnsmasq */
	snprintf(tmp, sizeof(tmp), "-%d", SIGHUP);
	eval("killall", tmp, "dnsmasq");
	
	return 0;
}

static int
del_ns(char *wan_ifname)
{
	FILE *fp, *fp2;
	char tmp[100], prefix[] = "wan0_";
	char word[100], *next;
	char line[100];
	int i;

	/* Open resolv.conf to read */
	if (!(fp = fopen("/var/resolv.conf", "r"))) {
		perror("fopen /var/resolv.conf");
		return errno;
	}
	/* Open resolv.tmp to save updated name server list */
	if (!(fp2 = fopen("/var/resolv.tmp", "w"))) {
		perror("fopen /var/resolv.tmp");
		fclose(fp);
		return errno;
	}
	/* Copy updated name servers */
	while (fgets(line, sizeof(line), fp)) {
		char *token = strtok(line, " \t\n");

		if (!token || strcmp(token, "nameserver") != 0)
			continue;
		if (!(token = strtok(NULL, " \t\n")))
			continue;

		//foreach(word, nvram_safe_get(strcat_r(prefix, "dns", tmp)), next)
		//if (!next)
		//	fprintf(fp2, "nameserver %s\n", token);
		
		for (i = 0; i < 2; i++)	{
			sprintf(tmp, "wan0_dns%d", i);
			strcpy(word, nvram_safe_get(tmp));
			if (!strcmp(word, token))
				continue;
			else
				fprintf(fp2, "nameserver %s\n", token);
		}			
	}
	fclose(fp);
	fclose(fp2);
	/* Use updated file as resolv.conf */
	unlink("/var/resolv.conf");
	rename("/var/resolv.tmp", "/var/resolv.conf");
	
	/* notify dnsmasq */
	snprintf(tmp, sizeof(tmp), "-%d", SIGHUP);
	eval("killall", tmp, "dnsmasq");
	
	return 0;
}

void
wan_up(char *wan_ifname)
{
char tmp[100], prefix[] = "wan0_";
char *wan_proto;
char *pstrWanLastIP;
char *pstrWanNowIP;


	wan_proto = nvram_safe_get("wan_proto");
	
	/* Set default route to gateway if specified */
	//if (nvram_match(strcat_r(prefix, "primary", tmp), "1"))
	//if (1) 
	if (0)
		route_add(wan_ifname, 0, "0.0.0.0", 
			nvram_safe_get(strcat_r(prefix, "gateway", tmp)),
			"0.0.0.0");

	/* Install interface dependent static routes */
	add_wan_routes(wan_ifname);

#if 0
	// Process multiple alias IPs here if static
	if( !strcmp( wan_proto, "static" ) )
	{
		char *strAliasIPs;
		char *strIP, *strMask;
		char cIf[8];
		int iIfIndex=0;
		char *ifArg[] = { "ifconfig", "", "", "netmask", "", NULL };	// ifconfig ifname ip netmask x.x.x.x
		char *routeArg[] = { "route", "add", "-host", "", "dev", "", NULL }; 
						// route add -host x.x.x.x dev vlan1:0
		// Get wan alias IPs
		strAliasIPs = strdup(nvram_safe_get(strcat_r(prefix, "aliasip", tmp)) );
		//printf( "%s:: alias IPs [%s]\n", __FUNCTION__, strAliasIPs );
		if( strAliasIPs[0] )		// Do have multiple IPs
		{
			while ( (strMask = strsep(&strAliasIPs, " ")) )
			{
				// Extract individual alias IP
				strIP = strsep( &strMask, "/" );
				// Set to interface
				sprintf(cIf, "%s:%d", nvram_get( "wan0_ifname" ), iIfIndex++ );
				ifArg[1] = cIf;
				ifArg[2] = strIP;
				ifArg[4] = strMask;
				if( _eval( ifArg, NULL, 0, NULL ) )
				{
					printf( "%s:: ifconfig fails, ifname[%s], IP/mask [%s/%s]\n",
						__FUNCTION__, ifArg[1] ,  ifArg[2], ifArg[4] );
				}
				// Add host route
				routeArg[3] = strIP;
				routeArg[5] = cIf;
				if( _eval( routeArg, NULL, 0, NULL ) )
				{
					printf( "%s:: route add fails, ifname[%s], IP [%s]\n",
						__FUNCTION__, routeArg[5],  routeArg[3] );
				}

			}
		}
		// Free buffer of wan alias IPs 
		free( strAliasIPs );
	}
#endif
	/* Add dns servers to resolv.conf */
	add_ns(wan_ifname);

	/* Sync time */
	start_ntpc();
#if 0	
	pstrWanLastIP = file2str("/tmp/WanLastIPAddr");
	pstrWanNowIP = nvram_get( "wan0_ipaddr");

	if( pstrWanLastIP && strcmp( pstrWanLastIP, pstrWanNowIP ) )
	{
		char *ddns_enable=nvram_safe_get("ddns_enable");
		
		str2file( "/tmp/WanLastIPAddr", pstrWanNowIP );
		if (!strcmp(ddns_enable,"1")) {
			if (ddns(pstrWanNowIP))
				cprintf("ddns update error \n");
		}	
	}	
	free( pstrWanLastIP );
#endif	
	cprintf("%s:: done\n", __FUNCTION__);
}


void
wan_down(char *wan_ifname)
{
	char tmp[100], prefix[] = "wan0_";
	char *wan_proto;
	char *pWanLastIP;

	//wan_proto = nvram_safe_get(strcat_r(prefix, "proto", tmp));
	wan_proto = nvram_safe_get("wan_proto");	
	
	/* Remove default route to gateway if specified */
	//if (nvram_match(strcat_r(prefix, "primary", tmp), "1"))
	if (1)
		route_del(wan_ifname, 0, "0.0.0.0", 
			nvram_safe_get(strcat_r(prefix, "gateway", tmp)),
			"0.0.0.0");

	/* Remove interface dependent static routes */
	del_wan_routes(wan_ifname);

	/* Update resolv.conf */
	del_ns(wan_ifname);

	if( (pWanLastIP =  file2str( "/tmp/WanLastIPAddr")) )
	{
		strcpy( pWanLastIP, "0.0.0.0" );
		str2file("/tmp/WanLastIPAddr", pWanLastIP);
		free( pWanLastIP );
	}
	
	cprintf("%s:: done\n", __FUNCTION__);
}


int
preset_wan_routes(char *wan_ifname)
{
	char tmp[100], prefix[] = "wan_";

	/* Set default route to gateway if specified */
	//if (nvram_match(strcat_r(prefix, "primary", tmp), "1"))
	if (1)
		route_add(wan_ifname, 0, "0.0.0.0", "0.0.0.0", "0.0.0.0");

	/* Install interface dependent static routes */
	add_wan_routes(wan_ifname);
	return 0;
}
