/*
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <re865x.h>
#include <board.h>

#define _PRE_TYPE_SET

#include <dni_nvram.h>
#include <dni_netconf.h>
#include <dni_shutils.h>
#include "rc.h"
#include <sys/ioctl.h>

#include <rtl8651_tblDrv.h>

extern int shmid;
extern int geted;
extern romeCfgParam_t * pRomeCfgParam;

int
start_dhcpd(void)
{
	FILE *fp;
	char name[100];
	int leaseTime;
	char leaseTimeStr[20];
	int i;

	if (nvram_invmatch("lan_proto", "dhcp"))
		return 0;

	dprintf("%s %s %s %s\n",
		nvram_safe_get("lan_ifname"),
		nvram_safe_get("dhcp_start"),
		nvram_safe_get("dhcp_end"),
		nvram_safe_get("dhcp_lease"));

/*	
echo "Create /var/udhcpd.leases"	
/bin/touch /var/udhcpd.leases	
/bin/chmod 777 /var/udhcpd.leases	
/sbin/udhcpd &
*/

	/* Touch leases file */
	if (!(fp = fopen("/var/udhcpd.leases", "a"))) {
		perror("/var/udhcpd.leases");
		return errno;
	}
	fclose(fp);


/*
EXAMPLE(/etc/udhcpd.conf):
start           192.168.2.20	
end             192.168.2.254	
interface       eth1	
lease_file      /var/udhcpd.leases	
opt     dns     192.168.1.1 192.168.1.100	
option  subnet  255.255.255.0	
opt     router  192.168.2.2	
opt     wins    192.168.2.10	
option  domain  local	
option  lease   864000
*/

	/* Write configuration file based on current information */
	if (!(fp = fopen("/var/udhcpd.conf", "w"))) {
		perror("/var/udhcpd.conf");
		return errno;
	}

	fprintf(fp, "pidfile /var/run/udhcpd.pid\n");
	fprintf(fp, "start %s\n", nvram_safe_get("dhcp_start"));
	fprintf(fp, "end %s\n", nvram_safe_get("dhcp_end"));
	fprintf(fp, "interface %s\n", "eth1");
	fprintf(fp, "remaining yes\n");
	fprintf(fp, "lease_file /var/udhcpd.leases\n");
	fprintf(fp, "option subnet %s\n", nvram_safe_get("lan_netmask"));
	fprintf(fp, "option router %s\n", nvram_safe_get("lan_ipaddr"));
	fprintf(fp, "option dns %s\n", nvram_safe_get("lan_ipaddr"));
	//modified by dvd.chen
	//fprintf(fp, "option lease %s\n", nvram_safe_get("lan_lease"));
	leaseTime = atoi( nvram_safe_get("dhcp_lease") );  //in half hour
	if( leaseTime <= 0 )
		leaseTime=1033377277;  // to be done, from now to year 2038 in seconds
	else
		leaseTime=leaseTime*30*60;
	snprintf(leaseTimeStr, 12, "%d", leaseTime);
	cprintf("%s::leasetime:%s\r\n", __FUNCTION__, leaseTimeStr);
	fprintf(fp, "option lease %s\n", leaseTimeStr);
	//end of dvd.chen
	snprintf(name, sizeof(name), "%s_wins", nvram_safe_get("dhcp_wins"));
	if (nvram_invmatch(name, ""))
		fprintf(fp, "option wins %s\n", nvram_get(name));
	snprintf(name, sizeof(name), "%s_domain", nvram_safe_get("dhcp_domain"));
	if (nvram_invmatch(name, ""))
		fprintf(fp, "option domain %s\n", nvram_get(name));
		
	for(i=0; i<2; i++) {
		sprintf(name, "manual %02X:%02X:%02X:%02X:%02X:%02X  %u.%u.%u.%u\n", 
				pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][0], pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][1],
				pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][2], pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][3],
				pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][4], pRomeCfgParam->dhcpsCfgParam.hardwareAddr[i][5],
				pRomeCfgParam->dhcpsCfgParam.manualIp[i]>>24, (pRomeCfgParam->dhcpsCfgParam.manualIp[i]&0x00ff0000)>>16,
				(pRomeCfgParam->dhcpsCfgParam.manualIp[i]&0x0000ff00)>>8, pRomeCfgParam->dhcpsCfgParam.manualIp[i]&0xff);
		fprintf(fp, name);
	}		
		
	fclose(fp);

	eval("udhcpd", "/var/udhcpd.conf");

	dprintf("done\n");
	return 0;
}
	
int
stop_dhcpd(void)
{
	char sigusr1[] = "-XX";
	int ret;

/*
* Process udhcpd handles two signals - SIGTERM and SIGUSR1
*
*  - SIGUSR1 saves all leases in /tmp/udhcpd.leases
*  - SIGTERM causes the process to be killed
*
* The SIGUSR1+SIGTERM behavior is what we like so that all current client
* leases will be honorred when the dhcpd restarts and all clients can extend
* their leases and continue their current IP addresses. Otherwise clients
* would get NAK'd when they try to extend/rebind their leases and they 
* would have to release current IP and to request a new one which causes 
* a no-IP gap in between.
*/
	sprintf(sigusr1, "-%d", SIGUSR1);
	eval("killall", sigusr1, "udhcpd");
	ret = eval("killall", "udhcpd");

	dprintf("done\n");
	return ret;
}


//start_dns(void)
int start_dns(int resetResolvConf)
{
	int ret;
	FILE *fp;
	
	if(resetResolvConf)
	{

		if (!(fp = fopen("/var/resolv.conf", "w"))) {
			perror("/var/resolv.conf");
			return errno;
		}
		
		/* add by lucas */
		fprintf(fp, "nameserver %s\n", "127.0.0.1");
		/* end by lucas */
		
		fclose(fp);
	}

	/* launch dns relay */
	ret = eval("dnsmasq", "-h", "-n",  //-n: NO POLL RESOLVE.CONF, OPT_NO_POLL
		   "-i", "eth1",
		   "-r", "/var/resolv.conf");

	cprintf("start_dns::done\r\n");
	return ret;
}	

int
stop_dns(void)
{
	int ret = eval("killall", "dnsmasq");
	
	/* Remove resolv.conf */
	unlink("/var/resolv.conf");

	dprintf("done\n");
	return ret;
}

int
start_httpd(void)
{
	int ret;

//	ret = eval("/usr/local/src/EasyAccess/bin/firebase","-d");
	if (task_is_running("httpd")!=1)
	{
		ret = eval("/usr/local/src/EasyAccess/bin/smm","-d");
		ret = eval("/usr/local/src/EasyAccess/bin/httpd","-d");
	}

	dprintf("done\n");
	return ret;
}

int
stop_httpd(void)
{
	int ret ; 
//	ret=eval("killall","-9" ,"httpd");
//	ret=eval("killall", "smm");
	dprintf("done\n");
	return ret;
}

int
stop_heartbeat(void)
{
	int ret;

	//unlink("/etc/ppp/link.ppp0");
	unlink(PPP_LINK_PATH"ppp0");
	unlink("/var/run/bpalogin.pid");
	
	ret = eval("killall", "bpalogin");  //send signal:TERM to bpalogin task, it will trigger the logout process
	sleep( 2 );
	ret += eval("killall", "-9", "bpalogin");

	//openlog("bpalogin", LOG_PID | LOG_NDELAY, LOG_DAEMON);
	//closelog();
	
	//log disconnect message in hb_disconnect_main
	//log_WAN_connect("WAN Telstra BigPond Disconnected");

	dprintf("done\n");
	
	return ret;
}
/*
* parse ifname to retrieve unit #
*/
int
ppp_ifunit(char *ifname)
{
	if (strncmp(ifname, "ppp", 3))
		return -1;
	if (!isdigit(ifname[3]))
		return -1;
	return atoi(&ifname[3]);
}


int
start_upnp(void)
{
	char *wan_ifname;
	int ret;

	if (!nvram_invmatch("upnp_enable", "0"))
		return 0;
	
	ret = eval("killall", "-SIGUSR1", "upnp");
	if (ret != 0) {
	    //modified by dvd.chen, when pppoe OR pptp ==> use ppp0 as wan_ifname
	    //wan_ifname = nvram_match(strcat_r(prefix, "proto", var), "pppoe") ? 
	    wan_ifname = ((nvram_match("wan_proto", "pppoe") 
	    				|| nvram_match("wan_proto", "pptp")) ? 
					"ppp0" : "eth0");

	    ret = eval("upnp", "-D",
		       "-L", "eth1",
		      "-W", wan_ifname, 
		      "-T", nvram_safe_get("upnp_advertime"));

	}
	dprintf("done\n");
	return ret;
}

int
stop_upnp(void)
{
	int ret = 0;

	//marked off by dvd.chen on 10/01/2004, kill upnp task to send ssdp_byebye 
	//whenever system is restarted
	//if (nvram_match("upnp_enable", "0"))
	    ret = eval("killall", "upnp");

	dprintf("done\n");
	return ret;
}

int do_kill(char *filename)
{
	int pid=0;
	FILE *f;
	f=fopen(filename,"r");
	if(f!=NULL)
	{
		fscanf(f,"%d",&pid);

		kill(pid,SIGTERM);
		kill(pid,SIGQUIT);
		sleep(1);
		kill(pid,SIGKILL);
		fclose(f);
		remove(filename);
		return 0;
	}
	return -1;	
}

int
start_ntpc(void)
{
#if 1
	char tempStr[100];
	char host1,host2;
	int	iTimeZone=atoi(nvram_get("time_zone"));
	FILE *fp;
	
	//sprintf(tempStr, "killall -15 ntpclient");	/* terminate */
	//system(tempStr);
	//sprintf(tempStr, "killall -9 ntpclient");
	//system(tempStr);
	do_kill("/var/run/ntpclient.pid");	
	eval("killall", "-9", "ntpclient");		
	sleep(1);
	iTimeZone= iTimeZone % 8;
	host1= 'a' + iTimeZone;
	iTimeZone= (iTimeZone+1) % 8;
	host2= 'a' + iTimeZone;
	sprintf(tempStr,"/bin/ntpclient -d -z %s -h time-%c.netgear.com,time-%c.netgear.com -i 3 -l -s &", nvram_get("time_zone"),host1,host2);	
	system(tempStr);

#else
	char ntp_servers[128];
	char ntp_server[3][48];
	char *token=NULL;
	char *tmp_ptr;
	int server_idx;
	int i=0;
	unsigned long wan_ip;

	if( task_is_running("ntpclient") == 1 )  //task is alive
	{
		cprintf("%s::ntpclient is running, will not start it again\n", __FUNCTION__);
		return 0;
	}
	
	cprintf("[T%d]%s::Enters\n",GetSysUpTime(),  __FUNCTION__);

	//randomize server_idx by using current_time and wan_ip
	inet_aton(nvram_safe_get("wan0_ipaddr"), (struct in_addr *)&(wan_ip));
	server_idx = (time(0) + ntohl(wan_ip)) % 3;
	
	//cprintf("%s::server_idx: %d\n", __FUNCTION__, server_idx);
	
	tmp_ptr = strdup(nvram_safe_get("ntp_server"));
	
	while ((token = strsep(&tmp_ptr, " ")) != NULL) 
	{
		sprintf(ntp_server[i], "%s", token);
		//cprintf("ntp_server[%d]: %s\n", i, ntp_server[i]);
		i++;
	}
	free(tmp_ptr);

	sprintf(ntp_servers, "%s %s %s", 
			ntp_server[server_idx], 
			ntp_server[(server_idx+1)%3], 
			ntp_server[(server_idx+2)%3]);
	//cprintf("ntp_servers: %s\n", ntp_servers);

	if (strlen(ntp_servers)) {
		char *ntp_argv[] = {"ntpclient", "-h", ntp_servers, "-i", "3", "-l", "-s", NULL};
		pid_t pid;

		_eval(ntp_argv, NULL, 0, &pid);
	}
	
	dprintf("done\n");
	return 0;
#endif
}

int
stop_ntpc(void)
{
	int ret = eval("killall", "ntpclient");

	dprintf("done\n");
	return ret;
}

int
start_services(void)
{

	// Commented by wayne.lu on 8/12/2005
	// Since httpd is better to be started after WAN is up, so the function is moved.
	//start_httpd();
	// End of modification.
	
	//start_dns();
	start_dns(1);  //reset resolv.conf
	start_dhcpd();
	start_upnp();

	dprintf("done\n");
	return 0;
}

int
stop_services(void)
{
	stop_upnp();
	stop_dhcpd();
	stop_dns();
	
	dprintf("done\n");
	return 0;
}

int
start_pptp(char **wan_ifname, char *wan_prefix)
{
	char username[80],passwd[80];
	char tmp[100];

	uint32 netmask = 0;
	uint8 strBuffer[128];
	int ret;
	FILE *f1, *f2;
	char papStr[20];
	char chapStr[20];	
	char pppOptionStr[20];	

	printf("Enter start_pptp\n");
		// protocol stack setting
		//upInterface("eth0");
		ifconfig( *wan_ifname, IFF_UP, "0.0.0.0", NULL );			
		
   		rtl8651_enableNaptAutoAdd(0);		
		
		netmask |= (pRomeCfgParam->pptpCfgParam.ipMask[0]) << 24;
		netmask |= (pRomeCfgParam->pptpCfgParam.ipMask[1]) << 16;
		netmask |= (pRomeCfgParam->pptpCfgParam.ipMask[2]) << 8;
		netmask |= pRomeCfgParam->pptpCfgParam.ipMask[3];

		ret = rtl8651_addNetworkIntf("eth0");

		ret = rtl8651_specifyNetworkIntfLinkLayerType("eth0", LL_PPTP, pRomeCfgParam->vlanCfgParam[0].vid);
		
		ret=rtl8651_addExtNetworkInterface((int8*)"eth0");		
		
		re865xIoctl("eth0", SET_VLAN_IP, *(uint32*)(&(pRomeCfgParam->pptpCfgParam.ipAddr)), *(uint32*)(&(pRomeCfgParam->pptpCfgParam.ipMask)),0);
	
		sprintf(strBuffer,"/sbin/ifconfig eth0 %d.%d.%d.%d netmask %d.%d.%d.%d",
			pRomeCfgParam->pptpCfgParam.ipAddr[0],
			pRomeCfgParam->pptpCfgParam.ipAddr[1],
			pRomeCfgParam->pptpCfgParam.ipAddr[2],
			pRomeCfgParam->pptpCfgParam.ipAddr[3],
			pRomeCfgParam->pptpCfgParam.ipMask[0],
			pRomeCfgParam->pptpCfgParam.ipMask[1],
			pRomeCfgParam->pptpCfgParam.ipMask[2],
			pRomeCfgParam->pptpCfgParam.ipMask[3]);

		system(strBuffer);

		ret = rtl8651_addSession(RTL8651_PPTP, 1, "eth0", pRomeCfgParam->vlanCfgParam[0].vid);
		
		ret = rtl8651_addRoute((ipaddr_t)0,(ipaddr_t)0,(int8*)"eth0",0);
		

	snprintf(tmp, sizeof(tmp), "-%d", SIGKILL);
	eval("killall", tmp, "pptp");	
	snprintf(tmp, sizeof(tmp), "-%d", SIGTERM);
	eval("killall", tmp, "pppd");		
	unlink("/var/run/pptp.pid");
	unlink("/var/run/pppd0.pid");

	f1 = fopen("/etc/pptp.options","r");
	if (!f1)
	{
		printf("pppd_start: fopen /etc/pptp.options failed\n");
		return;
	}

	
	sprintf(papStr,"/var/ppp/pap%u", 0);
	f2 = fopen(papStr,"w+");
	if (!f2)
	{
		printf("pppd_start: fopen %s failed\n",papStr);
		return;
	}
		
	sprintf(tmp,"%s * %s\n" ,pRomeCfgParam->pptpCfgParam.username,pRomeCfgParam->pptpCfgParam.password);
	fputs(tmp,f2);
	fclose(f2);

	sprintf(chapStr,"/var/ppp/chap%u", 0);
	unlink(chapStr);
		
	sprintf(chapStr,"/var/ppp/chap%u", 0);
	f2 = fopen(chapStr,"w+");
	if (!f2)
	{
		printf("pppd_start: fopen %s failed\n",chapStr);
		return;
	}
	sprintf(tmp,"NTDOMAIN %s %s\n" ,pRomeCfgParam->pptpCfgParam.username,pRomeCfgParam->pptpCfgParam.password);
	fputs(tmp,f2);
	sprintf(tmp,"%s NTDOMAIN %s\n" ,pRomeCfgParam->pptpCfgParam.username,pRomeCfgParam->pptpCfgParam.password);
	fputs(tmp,f2);			
	fclose(f2);

	sprintf(papStr,"/var/ppp/pap%u", 0);
	unlink(papStr);
		
	
	sprintf(pppOptionStr,"/var/ppp/options%u", 0);
	f2 = fopen(pppOptionStr,"w+");
	while (!feof(f1))
	{
		fgets(tmp,160,f1);
		fputs(tmp,f2);
	}			
	fclose(f1);
	
	sprintf(tmp,"mtu %d mru %d\n",pRomeCfgParam->pptpCfgParam.mtu,pRomeCfgParam->pptpCfgParam.mtu);
	fputs(tmp,f2);	
	sprintf(tmp,"name %s\n",pRomeCfgParam->pptpCfgParam.username);	
	fputs(tmp,f2);
	
	fputs("remotename NTDOMAIN\n",f2);
	
	/* generate idle option (auto disconnect) for pppd's reference */
	fprintf(f2,"idle %u\n",pRomeCfgParam->pptpCfgParam.silentTimeout);

	fprintf(f2,"defaultroute\n");
		
	fclose(f2);

	rtl8651_setNetMtu("eth0",pRomeCfgParam->pptpCfgParam.mtu);

	if (pRomeCfgParam->pptpCfgParam.demand==1)
	{
		pRomeCfgParam->pptpCfgParam.dialState=PPPOECFGPARAM_DIALSTATE_DIALED_WAITING;
		printf("Add Demand Route\n");
		AddDemandRoute(4,1); // 4: used for signal 44.		
	}
	else
	{
		pRomeCfgParam->pptpCfgParam.dialState=PPPOECFGPARAM_DIALSTATE_DIALED_TRYING;


		/* set auto redial options */
		f1 = fopen("/var/run/pptp_redial", "w");
		if (f1)
		{
			fprintf(f1, "%d", pRomeCfgParam->pptpCfgParam.autoReconnect?1:0);
			fclose(f1);
		}
	

		{
			sprintf(tmp,"/bin/pptp %u.%u.%u.%u &"
			,pRomeCfgParam->pptpCfgParam.svAddr[0]
			,pRomeCfgParam->pptpCfgParam.svAddr[1]
			,pRomeCfgParam->pptpCfgParam.svAddr[2]
			,pRomeCfgParam->pptpCfgParam.svAddr[3]);
			system(tmp);
		}
	}	

	
}

int
stop_pptp(void)
{
	int ret;

	//unlink("/etc/ppp/link.ppp0");
	unlink(PPP_LINK_PATH"ppp0");
	unlink("/etc/ppp/options");
	unlink("/etc/ppp/chap-secrets");
	unlink("/etc/ppp/pap-secrets");
	unlink("/etc/ppp/connect-errors");

	ret = eval("killall","-9","pppd");
	ret += eval("killall","-9","pptp");
	//ret += eval("killall","-9","listen");

	dprintf("done\n");
	return ret ;
}


void AddDemandRoute(int pppObjId,int defaultSession)
{
	rtl8651_tblDrvDemandRoute_t tmpRt;
	int32  i;
	uint32 ip_start,ip_end;
	uint16 port_start,port_end;
	
	rtl8651_flushDemandRoute(pppObjId);
	
	if (defaultSession){
		tmpRt.type=TRIGGER_DSTIP;
		tmpRt.ip_start=0;
		tmpRt.ip_end=0xffffffff;
		rtl8651_addDemandRoute(&tmpRt,pppObjId,0);
		return;
	}

}

