/*
 */


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>

#include <sys/klog.h>

#include <signal.h>

#include "../boa.h"
#include "httpd.h"

#include <board.h>
#include "../rtl865x/flashdrv.h"

extern romeCfgParam_t * pRomeCfgParam;

extern char *get_cgi(char *name);
extern int set_cgi(char *name, char *value);

//#define websWrite(wp, fmt, args...) (printf(fmt, ## args))
#define websError(wp, code, msg, args...) so_printf(wp, msg, ## args)
#define websHeader(wp) so_printf(wp, "<html lang=\"en\">")
#define websFooter(wp) so_printf(wp, "</html>")
#define websDone(wp, code)
#define websGetVar(wp, var, default) (get_cgi(var) ? : default)
#define websSetVar(wp, var, value) set_cgi(var, value)
#define websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query) ({ do_ej(path, wp); })
#define websBufferWrite websWrite



static const char * const apply_refresh_header =
"<head>"
"<meta http-equiv=\"Refresh\" content=\"%d; url=%s\">"
"</head>"
"<body>"
"</body>"
;


enum {
    NOTHING,
    REBOOT,
    RESTART,
    NOREPLY
};

#define EZC_FLAGS_READ      0x0001
#define EZC_FLAGS_WRITE     0x0002

#define EZC_FLAGS_NVRAM_NAME    0x0004

#define EZC_SUCCESS     0
#define EZC_ERR_NOT_ENABLED     1
#define EZC_ERR_INVALID_STATE   2
#define EZC_ERR_INVALID_DATA    3

#define ARRAYSIZE(a)        (sizeof(a)/sizeof(a[0]))

struct variable {
    char *name;
    char *longname;
    void (*validate)(webs_t wp, char *value, struct variable *v);
    char **argv;
    int nullok;
    int ezc_flags;
    char *nvram_name;
};

#define ARGV(args...) ((char *[]) { args, NULL })
#define XSTR(s) STR(s)
#define STR(s) #s

int variables_arraysize(void);


#define TRUE    1
#define FALSE   0

extern uint32 wan_pastType;
extern uint32 g_pastWanType;


#if 0
struct attached_devices _attached_devices[30];//ATTACHED_DEVICES
int atchedDvisCnt;//ATTACHED_DEVICES
struct dhcpc_rsvd_addr _dhcpc_rsvd_addr[30];//ATTACHED_DEVICES
int dhcpcRsvdAddrCnt;//ATTACHED_DEVICES
#endif


#define DirectoryIndex "index.html"

char return_page[64] = DirectoryIndex;
int refresh_time = 10;

int upgrading = 0;

char *none_backup_nvram[]=
{
    "wireless",
    "fw_pppoetype",
    "show_Product",
    "vs_count",
    "add_block_keyword_domain",
    "del_block_keyword_domain",
    "clr_block_keyword_domain",
    "fw_blockrulenumber",
    "fw_packrulenumber",
    "fw_inidays",
    "client_filter_ip",
    "client_filter_s_ip",
    "client_filter_e_ip",
    "logsubject",
    "alertsubject",
    "trigger_count",
    "dhcpc_res_count",
    //"http_passwd",
    "SmtWzdStep"
    "ddns_status",
    "dhcpc_lease_obtained",
    "dhcpc_lease_expired",

    "triggerporten",
    "sr_count",
    "wan_tel_manualHangup",
    "wizard_wan_spoofmac_detected",

    "qos_count",


    NULL
};

void httpc_info_save(void)
{
    copy_client_to_httpc_info((char *)&client_req);
    httpc_cfg_save();
}


void sys_reboot(void)
{
    upgrading = 1;
    httpc_info_save();
    kill(1, SIGQUIT);

}

void sys_restart(void)
{
    upgrading = 1;
    /* backup lease info */
    send_signal("/var/run/udhcpd.pid", SIGUSR1);
    httpc_info_save();
    kill(1, SIGHUP);
    //sysReinit(wan_pastType);

}

/******** Add by Timo Wu********/
#define NAME_CONFLICT       2
#define INBOUND_PORT_CONFLICT   3
#define TRIGGER_PORT_CONFLICT   4
#define FORWARD_PORT_CONFLICT   5

static int
check_porttrigger_entry(char *service_name, char *start_port, char *end_port, char *trigger , int current_index)
{
    int i, search_start;
    char nvram_name[20];
    char *ptr;
    char name[20];
    //for string pasing
    char *tmp;
    char *wan_port0, *wan_port1, *wan_proto, *lan_ipaddr, *lan_port0, *lan_port1, *proto, *enable, *used, *desc;
    int start ,end ,lan_start, lan_end;

    sprintf( name, "%s", service_name);

    search_start = 0;
    //for(i = search_start; i < current_index ;i++)
    for(i = search_start; i < MAX_SPECIAL_AP ;i++)
    {
        if (i == current_index)
            continue;

        sprintf(nvram_name, "trigger_port%d", i);
        ptr = nvram_safe_get(nvram_name);
    /*
     *  string parsing
     */

    tmp = ptr;

    /* Check for WAN PORT address specification */

    wan_proto = tmp;
    wan_port0 = strsep(&wan_proto, ",");
    if (!wan_proto)
        return TRUE;

    lan_ipaddr = wan_proto;
    wan_proto = strsep(&lan_ipaddr, ">");
    if (!lan_ipaddr)
        return TRUE;

    /* Check for LAN destination port specification */
    lan_port0 = lan_ipaddr;
    lan_ipaddr = strsep(&lan_port0, ":");
    if (!lan_port0)
        return TRUE;

    /* Check for protocol specification */
    proto = lan_port0;
    lan_port0 = strsep(&proto, ":,");
    if (!proto)
        return TRUE;

    /* Check for enable specification */
    enable = proto;
    proto = strsep(&enable, ":,");
    if (!enable)
        return TRUE;

    /* Check for used specification (optional) */
    used = enable;
    enable = strsep(&used, ":,");
    if (!used)
        return TRUE;

    /* Check for description specification (optional) */
    desc = used;
    used = strsep(&desc, ":,");
    if (!desc)
        return TRUE;

    /* Check for WAN destination port range (optional) */
    wan_port1 = wan_port0;
    wan_port0 = strsep(&wan_port1, "-");
    if (!wan_port1)
        wan_port1 = wan_port0;

    /* Check for LAN destination port range (optional) */
    lan_port1 = lan_port0;
    lan_port0 = strsep(&lan_port1, "-");
    if (!lan_port1)
        lan_port1 = lan_port0;

        lan_start = atoi(lan_port0);
        lan_end = atoi(lan_port1);
        start = atoi(start_port);
        end = atoi(end_port);
        //printf("service_name %s , tigger %s, start_port %s, end_port %s\n", desc, wan_port0, lan_port0 ,lan_port1);
        if (strcmp(name, desc)==0)
        {

            return NAME_CONFLICT; //1 means find previous one conflict
        }
        //else if( (strcmp(start_port, lan_port0)==0) && (strcmp(end_port, lan_port1)==0) )
        else if( ((start <= lan_start)&&( lan_start <= end)) || ((start <= lan_end)&&( lan_end <= end)) || ((lan_start <= start)&&( start <= lan_end)) || ((lan_start <= end)&&( end <= lan_end)))
        {

            return INBOUND_PORT_CONFLICT;
        }
        else if ( strcmp(trigger , wan_port0)==0 )
        {

            return TRIGGER_PORT_CONFLICT;
        }


    }

#if 0
    search_start = current_index + 1;
    for(i = search_start; i < MAX_SPECIAL_AP ;i++)
    {
        sprintf(nvram_name, "trigger_port%d", i);
        ptr = nvram_safe_get(nvram_name);

    /*
     *  string parsing
     */

    tmp = ptr;

    /* Check for WAN PORT address specification */

    wan_proto = tmp;
    wan_port0 = strsep(&wan_proto, ",");
    if (!wan_proto)
        return TRUE;

    lan_ipaddr = wan_proto;
    wan_proto = strsep(&lan_ipaddr, ">");
    if (!lan_ipaddr)
        return TRUE;

    /* Check for LAN destination port specification */
    lan_port0 = lan_ipaddr;
    lan_ipaddr = strsep(&lan_port0, ":");
    if (!lan_port0)
        return TRUE;

    /* Check for protocol specification */
    proto = lan_port0;
    lan_port0 = strsep(&proto, ":,");
    if (!proto)
        return TRUE;

    /* Check for enable specification */
    enable = proto;
    proto = strsep(&enable, ":,");
    if (!enable)
        return TRUE;

    /* Check for used specification (optional) */
    used = enable;
    enable = strsep(&used, ":,");
    if (!used)
        return TRUE;

    /* Check for description specification (optional) */
    desc = used;
    used = strsep(&desc, ":,");
    if (!desc)
        return TRUE;

    /* Check for WAN destination port range (optional) */
    wan_port1 = wan_port0;
    wan_port0 = strsep(&wan_port1, "-");
    if (!wan_port1)
        wan_port1 = wan_port0;

    /* Check for LAN destination port range (optional) */
    lan_port1 = lan_port0;
    lan_port0 = strsep(&lan_port1, "-");
    if (!lan_port1)
        lan_port1 = lan_port0;

        lan_start = atoi(lan_port0);
        lan_end = atoi(lan_port1);
        start = atoi(start_port);
        end = atoi(end_port);
        //printf("service_name %s , tigger %s, start_port %s, end_port %s\n", desc, wan_port0, lan_port0 ,lan_port1);
        if (strcmp(name, desc)==0)
        {
            return NAME_CONFLICT; //1 means find previous one conflict
        }
        //else if( (strcmp(start_port, lan_port0)==0) && (strcmp(end_port, lan_port1)==0) )
        else if( ((start <= lan_start)&&( lan_start <= end)) || ((start <= lan_end)&&( lan_end <= end)) || ((lan_start <= start)&&( start <= lan_end)) || ((lan_start <= end)&&( end <= lan_end)))
        {

            return INBOUND_PORT_CONFLICT;
        }
        else if ( strcmp(trigger , wan_port0)==0 )
        {

            return TRIGGER_PORT_CONFLICT;
        }

    }
#endif

    return TRUE; //not finding any one conflict
}
/*****End*****/

#define MAX_CGI_SIZE    1024
//#define MAX_CGI_SIZE  4096
static char cgi_buf[MAX_CGI_SIZE];

void
init_form_cgi(webs_t wp, char **query)
{
    struct stat statbuf;
    char *buf;

    fstat(wp->post_data_fd, &statbuf);
    //buf=(char *)malloc(statbuf.st_size+1);
    if (statbuf.st_size >= MAX_CGI_SIZE) {
        init_dni_cgi(NULL);
        return;
    }
    buf = cgi_buf;
    if (buf) {
        lseek(wp->post_data_fd, SEEK_SET, 0);
        read(wp->post_data_fd, buf, statbuf.st_size);
        init_dni_cgi(NULL);
        buf[statbuf.st_size] = 0;
        init_dni_cgi(buf);
        //*query = buf;
    }
}

static int
validate_dhcpc_res_ip_adding(char *newIP)
{
    unsigned char *ptr, str[20], *startDhcp, myStartDhcp[4], myIP[10];
    unsigned char path[60], *endDhcp, myEndDhcp[4];
    int ret = 0;
    webs_t wp;

    strcpy(str, newIP);
    ret = inet_aton(str, myIP);

    if (!ret)
        printf("error !\n");

    //printf("myIP[] = %d %d %d %d\n", myIP[0], myIP[1], myIP[2], myIP[3]);
    startDhcp = nvram_get("dhcp_start");
    //printf("startDhcp[] = %s\n", startDhcp);
    inet_aton(startDhcp, myStartDhcp);
    //printf("myStartDhcp = %d %d %d %d\n", myStartDhcp[0], myStartDhcp[1], myStartDhcp[2], myStartDhcp[3]);

    endDhcp = nvram_get("dhcp_end");
    //printf("endDhcp[] = %s\n", endDhcp);
    inet_aton(endDhcp, myEndDhcp);
    //printf("myEndDhcp = %d %d %d %d\n", myEndDhcp[0], myEndDhcp[1], myEndDhcp[2], myEndDhcp[3]);

#if 0
    myStartDhcp = strsep(&startDhcp, ".");
    printf("myStartDhcp0 = %s\n",myStartDhcp);
    myStartDhcp = strsep(&startDhcp, ".");
    printf("myStartDhcp1 = %s\n",myStartDhcp);
    myStartDhcp = strsep(&startDhcp, ".");
    printf("myStartDhcp2 = %s\n",myStartDhcp);
    printf("myStartDhcp3 = %s\n",startDhcp);

    myEndDhcp = strsep(&endDhcp, ".");
    printf("myEndDhcp0 = %s\n",myEndDhcp);
    myEndDhcp = strsep(&endDhcp, ".");
    printf("myEndDhcp1 = %s\n",myEndDhcp);
    myEndDhcp = strsep(&endDhcp, ".");
    printf("myEndDhcp2 = %s\n",myEndDhcp);
    printf("myEndDhcp3 = %s\n",endDhcp);
#endif

    //printf("myIP[3] = %d, myStartDhcp[3] = %d, myEndDhcp[3] = %d\n", myIP[3], myStartDhcp[3], myEndDhcp[3]);
    if ((myIP[3] > myEndDhcp[3]) || (myIP[3] < myStartDhcp[3]) || (myIP[0] != myStartDhcp[0])  || (myIP[1] != myStartDhcp[1])  || (myIP[2] != myStartDhcp[2])) {
        //printf("Exceed DHCP range\n");
        return FALSE;
    }

    return TRUE;

}

int action = NOTHING;
int noread = 0;
#ifdef CONFIG_WAN_LAN_MATCH_DETECTION
void Set_Flag_WAN_LAN_MATCH_DET(int flag);
#endif

void
apply_cgi(webs_t wp, char_t *path, char_t **query)
{
    char *value;
    int delay;

        char strURL[128];
        char *path;
        char *ptr, *newIP; //ATTACHED_DEVICES
    int ret; //ATTACHED_DEVICES

        wan_pastType = pRomeCfgParam->ifCfgParam[0].connType;

    action = NOTHING;
    init_form_cgi(wp, query);

#if defined(CONFIG_WAN_LAN_MATCH_DETECTION)

    value = websGetVar(wp,"SubnetMatch",NULL);
    if(value !=NULL) {
        ret=atoi(value);
        delay = atoi(nvram_get("WL_ip_match"));
        printf("%s, ret is %d, delay is %d, ",__FUNCTION__,ret,delay);
        if(ret != delay) {
            printf("SetWanLanFlag ");
            Set_Flag_WAN_LAN_MATCH_DET(ret);
            return;
        }
        printf("\n");
    }
#endif
    value = websGetVar(wp, "action", "");

    /* Apply values */
    if (!strcmp(value, "Apply")) {
      value = websGetVar(wp, "Button", "");

      if (strstr(value, "Apply"))
      {

        if(strcmp(websGetVar(wp, "reboot", ""), "1")==0)
        {
            action = REBOOT;
            delay = 30;
        }
        else
        {
            //action = RESTART;
                //delay = 10;
        }

        //if (!strcmp(websGetVar(wp, "Load", ""), "W"))
        //  refresh_time = 8;
        //else
        //  refresh_time = 2;
        //websHeader(wp);
#if 1
        if (!strcmp(websGetVar(wp, "Load", ""), "H")) {
            newIP = websGetVar(wp, "subip", NULL);
            if (!newIP) {
                newIP = websGetVar(wp, "editiperr", NULL);
            }
            ret = validate_dhcpc_res_ip_adding(newIP);
            if (ret == FALSE) {
                char path[64] = "/www/edit_exceed_err.html";
                ptr = websGetVar(wp, "page", "");
                //printf("ptr = %s\n", ptr);
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else{
                            sprintf(return_page, "/%s", ptr);
                    //strcpy(path, "/www/");
                    //strcat(path, "edit_exceed_err.html");
                        }
                noread = 1;
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                return;
            }
        }
#endif

        if (!strcmp(websGetVar(wp, "Load", ""), "I1")) {
            if( validate_fw_service_cmp(wp) == 1){
                char path[64] = "/www/error_portexist.html";
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else{
                            sprintf(return_page, "/%s", ptr);
                        }
                        noread = 1;
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                return;
            }
        }

            {
                    char *pstrLanIP;
                    char path[64] = "/www/test_conn1.html";


                    value = websGetVar(wp, "page", NULL);
                    if (!value)
                        sprintf(return_page, "/%s", DirectoryIndex);
                    else
                        sprintf(return_page, "/%s", value);

                    pstrLanIP= websGetVar(wp, "lan_ipaddr", NULL);
                    if (pstrLanIP == NULL)
                pstrLanIP= websGetVar(wp, "LANIPAddr", NULL);

            if ((pstrLanIP != NULL) && (strcmp(nvram_safe_get("lan_ipaddr"), pstrLanIP) != 0)) {

                sprintf(strURL, "http://%s/%s", pstrLanIP, DirectoryIndex);

            }
                    else if (pstrLanIP != NULL)
                        sprintf(strURL, "http://%s/%s", pstrLanIP, websGetVar(wp, "page", DirectoryIndex));
                    else
                        sprintf(strURL, "/%s", websGetVar(wp, "page", DirectoryIndex));

                    //websWrite(wp, (char_t *) apply_refresh_header, delay, strURL);
            //strcat(path, value);
            //websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);


            }

        validate_cgi(wp);
        nvram_set("is_modified", "1");
        nvram_set("is_default", "0");
        //nvram_commit();

        //websFooter(wp);
        //websDone(wp, 200);
#if 1
        if ((action != RESTART) && (action != REBOOT) && (action != NOREPLY) ) //Add by Timo Wu
            {
                    char *pstrLanIP;
                    char path[64] = "/www/";
                    value = websGetVar(wp, "page", NULL);
            strcat(path, value);
            websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
            }
            else if(action != NOREPLY) //Add by Timo Wu
            {
#if 0           //Kevin:05Feb15:TestButton
                    char path[64] = "/www/test_conn1.html"; //Update Settings
            websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
#else
                    char path[64] = "/www/test_conn1.html"; //Update Settings
                    char path1[64] = "/www/Test_update_proc.html"; //Update Settings
            char * DoTest = websGetVar(wp, "DoTest", "");

            if(!strcmp(DoTest,"yes")) {
                websDefaultHandler(wp, urlPrefix, webDir, arg, url, path1, query);
            } else {
                websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
            }
#endif
        }

#endif


      }
      else if (strstr(value, "Add") || strstr(value, "Edit") || strstr(value, "Modify") || strstr(value, "Del") || strstr(value, "Clear")) {
            validate_cgi(wp);
      }
      else if (!strcmp(value, "Reset_POT")) {
                  char path[64] = "/www/";
        //printf("####restore button\n");

              //POT_factoryDefault(POT_GetResolution()+1, POT_GetRepackThreshold()+1);
              POT_factoryDefault(1, 30);


              value = websGetVar(wp, "page", NULL);
          strcat(path, value);
          websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
          }

    }
    else if (!strncmp(value, "Logout", 6)) {

        value = websGetVar(wp, "page", NULL);
        if (!value) {
            sprintf(strURL, "/www/%s", DirectoryIndex);
        }
        else
            sprintf(strURL, "/www/%s", value);
        path = strURL;
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    }
    /* Save backup */
    else if (!strncmp(value, "Backup", 6)) {
        value = nvram_show_all(1, 0, none_backup_nvram);
                  /* 1 -> from first */
        while (value) {
            if (*value) {
                //ret += websWrite(wp, "%s\n", value);
                websWrite(wp, "%s\n", value);
            }
            value = nvram_show_all(0, 0, none_backup_nvram);
                      /* 0 -> continue */
        }
    }

    else if (!strncmp(value, "Restore2Default", 15)) {

        value = websGetVar(wp, "page", NULL);
        if ((!value) || (*value == 0)) {
            sprintf(strURL, "/www/%s", DirectoryIndex);
        }
        else
            sprintf(strURL, "/www/%s", value);

extern ifCfgParam_t ifCfgParamDefault[];
        char old_lan_ipaddr[20];
        strcpy(old_lan_ipaddr, nvram_get("lan_ipaddr"));
        nvram_set("lan_ipaddr", inet_string((uint32)&(ifCfgParamDefault[1].ipAddr)));
        path = strURL;
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
        stop_service();
        if (vfork() == 0) {
            close(wp->fd);
            nvram_set("restore_defaults", "1");
            cfgmgr_factoryDefault(0);
            //action = REBOOT;
            if (strcmp(old_lan_ipaddr, nvram_get("lan_ipaddr"))) {
                int i;
                for(i=1;i<5;i++) {
                    rtl8651_setEthernetPortPowerDown(i,1);
                }
                sleep(3);
            }
            sys_reboot();
            exit(1);
        }
        //wp->status = 10/*REBOOT*/;
        //Sys_Restart(wp);
    }
    else if (!strncmp(value, "Restore", 7)) {
        //nvram_show(wp);
    }

    else if (!strncmp(value, "Scanning", 8)) {

        value = websGetVar(wp, "page", NULL);
        if ((!value) || (*value == 0)) {
            sprintf(strURL, "/www/%s", DirectoryIndex);
        }
        else
            sprintf(strURL, "/www/%s", value);

        /* scaning LAN devices via ARP, for example,192.168.1.2-192.168.1.254 */
        path = strURL;
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    }

    /* Log action */
    else if (!strcmp(value, "Log")) {
        validate_log_action_setting(wp, NULL, NULL);

    }
    /* Connection Action */
    else if (!strncmp(value, "connection", 6)) {
        validate_connection_action(wp, NULL, NULL);

    }
    /* Reboot router */
    else if (!strcmp(value, "Reboot")) {
        char path[64] = "/var/www/ui_restart_succ.html";
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
        action = REBOOT;
    }
//Yili 2006/1/10
#if defined(DNI)
    /* Menu login */
    else if (!strcmp(value, "MNU_Login")) {
        validate_mnu_login_action(wp, NULL, NULL);

        char path[64] = "/www/MNU_login.html";
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
    }
    else if (!strcmp(value, "VER_sel")) {
        validate_mnu_login_action(wp, NULL, NULL);

        char path[64] = "/www/basicform1.html";
        websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
    }
#endif
//end



    /* Invalid action */
    else {
        websWrite(wp, "Invalid action %s<br>", value);
    }

    if (action == REBOOT) {
        //send_signal("/var/run/checkwan.pid", SIGHUP);
        sys_reboot();
    }
    else if (action == RESTART) {
#if defined(DHCPPPTP)
#define DHCPROUTEING_INFO     "/var/dhcproute.conf"
    remove(DHCPROUTEING_INFO);
#endif
        sys_restart();
    }

    return;
}

//#define DPRINTF       printf
#define DPRINTF

#define MIN(a, b)   ((a <= b) ? a : b)

#if 1
void
confretrieve_cgi(webs_t wp, char_t *path, char_t **query)
{

    char buf[1024];
    int len, length;
    FILE *conn_fp;
    char path[64] = "/www/";
    char *nvram_name, *value;
    struct stat statbuf;
    char old_lan_ipaddr[20];

    strcpy(old_lan_ipaddr, nvram_get("lan_ipaddr"));

        wan_pastType = pRomeCfgParam->ifCfgParam[0].connType;

    fstat(wp->post_data_fd, &statbuf);
    lseek(wp->post_data_fd, SEEK_SET, 0);
    DPRINTF("\n\nReading restored data from HTTP.....%d\n", statbuf.st_size);

    if (!(conn_fp = fopen(wp->post_file_name, "rb"))) {
        perror("fopen");
        goto exit_1;
    }

    len = statbuf.st_size;
    length = statbuf.st_size;
    /* Look for our part */
    while (len > 0) {
        /* read instead of */
        if (!fgets(buf, MIN(len + 1, sizeof(buf)), conn_fp)) {
            DPRINTF("%s 1\n", __FUNCTION__);
            goto exit_1;
        }
        len -= strlen(buf);
        if (!strncasecmp(buf, "Content-Disposition:", 20) &&
            strstr(buf, "name=\"file\""))
            break;
    }

    /* Skip boundary and headers */
    while (len > 0) {
        if (!fgets(buf, MIN(len + 1, sizeof(buf)), conn_fp)) {
            DPRINTF("%s 2\n", __FUNCTION__);
            goto exit_1;
        }
        len -= strlen(buf);
        if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
            break;
    }
    DPRINTF( "%s:: Before len [%d]\n", __FUNCTION__, len );

    char *chksum, /**cfg_chksum_str = "cfg_chksum", */cfg_value[100], *ptr;
    int i = 0, val = 0, config_len = 0, totalLen = 0, cfg_len = 0, right_config_file = 0, chksum_flag = 0;
    int hw_version_ok = 0, model_ok = 0;

    config_len = len;
    length -= config_len;
    while (len > 0) {
        if (!fgets(buf, MIN(len + 1, sizeof(buf)), conn_fp)) {
            DPRINTF("%s 3\n", __FUNCTION__);
            goto exit_1;
        }

        DPRINTF("%s", buf);

        len -= strlen(buf);

        if (strstr(buf, "-----"))
            break;

        value = buf;

        memset(cfg_value, 0, 100);
        strncpy(cfg_value, value, 100);
        buf[strlen(buf)-1] = 0; /* skip last 0x0a */
        nvram_name = strsep(&value, "=");
        //printf("%d nvram_name = %s; value = %s\n", __LINE__, nvram_name, value);

        //if (strncmp (nvram_name, cfg_chksum_str, strlen(cfg_chksum_str)) == NULL) {
        if (strncmp (nvram_name, "end", strlen("end")) == NULL) {
            chksum_flag = 1;
        } else if (chksum_flag == 1) {
            //printf("%d cfg_value = %s\n", __LINE__, cfg_value);
            val = atoi(cfg_value);
            right_config_file = 1;
            break;
        } else {
            //printf("%d = %s\n", __LINE__, cfg_value);
            cfg_len = strlen(cfg_value)-1;

            for (i = 0; i < cfg_len; i++)
                totalLen += cfg_value[i];
        }

        //if (strncmp (nvram_name, "end", strlen("end")) == NULL)
            //chksum_flag = 1;
        if (!hw_version_ok && nvram_name && value) {
            if ((strcmp(nvram_name, "hw_version") == 0) &&
                (strcmp(value, "V4H1") == 0))
                hw_version_ok = 1;
            else if ((strcmp(nvram_name, "model") == 0) &&
                     (strcmp(value, "RP614v4") == 0))
                model_ok = 1;
        }
    }

    chksum_flag = 0;
    printf("%d =  totalLen = %d : val = %d\n", __LINE__, totalLen, val);

    if ( (right_config_file == 0) || (hw_version_ok == 0) || (model_ok == 0))
            goto exit_1;

    //if (conn_fp)
        //fclose(conn_fp);

    if (val  == totalLen) {
        if (!(conn_fp = freopen(wp->post_file_name, "rb", conn_fp))) {
            perror("fopen");
            goto exit_1;
        }

        fseek(conn_fp, length, SEEK_SET);
        while (config_len > 0) {
            if (!fgets(buf, MIN(config_len + 1, sizeof(buf)), conn_fp)) {
                DPRINTF("%s 3\n", __FUNCTION__);
                goto exit_1;
            }

            //DPRINTF("%s", buf);

            len -= strlen(buf);
            if (strstr(buf, "-----"))
                break;

            value = buf;
            buf[strlen(buf)-1] = 0; /* skip last 0x0a */
            nvram_name = strsep(&value, "=");
            //if (strncmp (nvram_name, cfg_chksum_str, strlen(cfg_chksum_str)) == NULL)
            if (strncmp (nvram_name, "end", strlen("end")) == NULL)
                chksum_flag = 1;
            else if (chksum_flag == 1)
                break;
            else
                nvram_set(nvram_name, value);

            //if (strncmp (nvram_name, "end", strlen("end")) == NULL)
                //chksum_flag = 1;
        }

    } else {
exit_1:
            //char path[64] = "/www/config_chksum_err.html";
            strcat(path, "config_chksum_err.html");
            ptr = websGetVar(wp, "page", "");
/*pamela            printf("ptr = %s\n", ptr);
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                    else{
                        sprintf(return_page, "/%s", ptr);
                //strcpy(path, "/www/");
                //strcat(path, "edit_exceed_err.html");
    }
*/
            //value = websGetVar(wp, "page", NULL);
            //if (!value)
            //  sprintf(return_page, "/%s", "backup.html");
            //else
            //  sprintf(return_page, "/%s", value);

            /* multipart can not provide cgi-variable */
            /* return page always is "backup.html */
            sprintf(return_page, "/%s", "backup.html");

            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            return;
    }
//Yili 2006/1/16
    char *pstrLanIP;
    char strURL[128];
    //strcat(path, "test_conn1.html");
    strcat(path, "erase_proc.html");
    //strcat(path, "upgrade_proc.html");

    //value = websGetVar(wp, "page", NULL);
    //if (!value)
    //  sprintf(return_page, "/%s", "backup.html");
    //else
    //  sprintf(return_page, "/%s", value);

    /* multipart can not provide cgi-variable */
    /* return page always is "backup.html */
    sprintf(return_page, "/%s", "backup.html");

    pstrLanIP= websGetVar(wp, "lan_ipaddr", NULL);
    if (pstrLanIP == NULL)
        pstrLanIP= websGetVar(wp, "LANIPAddr", NULL);

    if ((pstrLanIP != NULL) && (strcmp(nvram_safe_get("lan_ipaddr"), pstrLanIP) != 0)) {

        sprintf(strURL, "http://%s/%s", pstrLanIP, DirectoryIndex);

    }
    else if (pstrLanIP != NULL)
        sprintf(strURL, "http://%s/%s", pstrLanIP, "backup.html"/*websGetVar(wp, "page", DirectoryIndex)*/);
    else
        sprintf(strURL, "/%s", "backup.html"/*websGetVar(wp, "page", DirectoryIndex)*/);

    //strcat(path, "restoresucc.html");
//Yili end

    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    if (conn_fp)
        fclose(conn_fp);
    stop_service();
    if (vfork() == 0) {
        char *ptr;

        close(wp->fd);
        //memset((char *)&(pRomeCfgParam->httpclientCfgParam), 0x0, sizeof(pRomeCfgParam->httpclientCfgParam));
        nvram_commit();
        ptr = nvram_get("lan_ipaddr");
        if (strcmp(old_lan_ipaddr, ptr)) {
            int i;
            clr_http_client();
            for(i=1;i<5;i++) {
                rtl8651_setEthernetPortPowerDown(i,1);
            }
            sleep(3);
        }
        sys_reboot();
        exit(1);
    }
    //sys_restart();
    //sysReinit(wan_pastType);
    return;

exit_2:
    strcat(path, "restoreerr.html");
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    if (conn_fp)
        fclose(conn_fp);

    return;
}
#else

/*
 * Reads data into the buffer, storing up to buf_size-1 bytes, or until a
 * (CR)LF has been read, or until eof.
 * Returns -1 on error, 0 on eof, or number of bytes read.
 */
int getline(unsigned char *buffer, int buf_size, unsigned char *start, int remain)
{
    int i;
    unsigned char ch;

    for (i=0; i < (buf_size-1) && remain >= 0;) {
        ch = *(start++);
        remain--;
        if (ch == '\n') {
            buffer[i++] = ch;
            break;
        }
        else if (ch == '\r') {
            buffer[i++] = ch;
        }
        else
            buffer[i++] = ch;
    }
    buffer[i] = '\0';
    return i;
}

void
confretrieve_cgi(webs_t wp, char_t *path, char_t **query)
{

    char buf[1024];
    int len;
    char path[64] = "/www/restoresucc.html";

    struct stat statbuf;
    char *ptr;
    char *buf1=NULL;


    fstat(wp->post_data_fd, &statbuf);

    DPRINTF("\n\nReading restored data from HTTP.....%d\n", statbuf.st_size);

    len = statbuf.st_size;

    buf1=(char *)malloc(statbuf.st_size+1);
    lseek(wp->post_data_fd, SEEK_SET, 0);
    read(wp->post_data_fd, buf1, statbuf.st_size);
    buf1[statbuf.st_size]=0;

    len = statbuf.st_size;
    ptr = buf1;

    /* Look for our part */
    while (len > 0) {
        /* read instead of */
        if (getline(buf, 1024, ptr, len) == 0) {
            goto exit_1;
        }
        len -= strlen(buf);
        ptr += strlen(buf);
        if (!strncasecmp(buf, "Content-Disposition:", 20) &&
            strstr(buf, "name=\"file\""))
            break;
    }

    /* Skip boundary and headers */
    while (len > 0) {
        if (getline(buf, 1024, ptr, len) == 0) {
            goto exit_1;
        }
        ptr += strlen(buf);
        len -= strlen(buf);
        if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
            break;
    }
    DPRINTF( "%s:: Before len [%d]\n", __FUNCTION__, len );
    while (len > 0) {
        if (getline(buf, 1024, ptr, len) == 0) {
            goto exit_1;
        }

        DPRINTF("%s", buf);
        len -= strlen(buf);
        ptr += strlen(buf);
    }
exit_1:
    if (buf1)
        free(buf1);
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);

    return;
}
#endif

#if 0

static char upgrade_file_name[50];
static unsigned int startPos,endPos;
void
upgrade_cgi(webs_t wp, char_t *path, char_t **query)
{

    int len;
    char path[64] = "/www/";

    FILE    *fp=NULL;
    struct stat statbuf;
    unsigned char c;

    /* find the start and end positive of run time image */
    printf("\nTry to get file size of new firmware\n");

    fstat(wp->post_data_fd, &statbuf);
    lseek(wp->post_data_fd, SEEK_SET, 0);
    printf("file size=%d\n",statbuf.st_size);
    strcpy(upgrade_file_name, wp->post_file_name);
    fp=fopen(wp->post_file_name,"rb");
    if(fp==NULL) {
        goto upgrade_fail;
    }
    /* Look for our part */
    do
    {
        if(feof(fp))
        {
            printf("Cannot find start of file\n");
            goto upgrade_fail;
        }
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        break;
    }while(1);

    do
    {
        if(feof(fp))
        {
            printf("Cannot find start of file\n");
            goto upgrade_fail;
        }
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        break;
    }while(1);

    startPos=ftell(fp);

    if(fseek(fp,statbuf.st_size-0x100,SEEK_SET)<0)
        goto upgrade_fail;
    do
    {
        if(feof(fp))
        {
            printf("Cannot find end of file\n");
            goto upgrade_fail;
        }
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!='-')
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!='-')
            continue;
        break;
    }while(1);
    endPos=ftell(fp);

    /* check header and checksum of this image */
    printf("endPos=%u startPos=%u\n",endPos,startPos);
    if (_checkImage(fp, startPos, endPos-startPos-4) != 0)
        goto upgrade_fail;

    free(wp->post_file_name);
    close(wp->post_data_fd);
    wp->post_file_name=NULL;
    wp->post_data_fd=NULL;
    strcat(path, "fwupstart.html");
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    return;

upgrade_fail:
    strcat(path, "fwuperr.html");
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);

    return;
}

#else

static char upgrade_file_name[50];
static unsigned int startPos,endPos;
void
upgrade_cgi(webs_t wp, char_t *path, char_t **query)
{

    int len;
    char buf[512];
    char path[64] = "/www/";

    FILE    *fp=NULL;
    struct stat statbuf;
    unsigned char c;


    /* find the start and end positive of run time image */
    printf("\nTry to get file size of new firmware\n");

    fstat(wp->post_data_fd, &statbuf);
    lseek(wp->post_data_fd, SEEK_SET, 0);
    printf("file size=%d\n",statbuf.st_size);
    strcpy(upgrade_file_name, wp->post_file_name);
    fp=fopen(wp->post_file_name,"rb");
    if(fp==NULL) {
        goto upgrade_fail;
    }
    /* Look for our part */
    /* Look for our part */
    len = statbuf.st_size;
    while (len > 0) {
        /* read instead of */
        if (!fgets(buf, MIN(len + 1, sizeof(buf)), fp)) {
            DPRINTF("%s 1\n", __FUNCTION__);
            goto upgrade_fail;
        }
        len -= strlen(buf);
        if (!strncasecmp(buf, "Content-Disposition:", 20) &&
            strstr(buf, "name=\"file\""))
            break;
    }

    /* Skip boundary and headers */
    while (len > 0) {
        if (!fgets(buf, MIN(len + 1, sizeof(buf)), fp)) {
            DPRINTF("%s 2\n", __FUNCTION__);
            goto upgrade_fail;
        }
        len -= strlen(buf);
        if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
            break;
    }
    DPRINTF( "%s:: Before len [%d]\n", __FUNCTION__, len );


    startPos=ftell(fp);

    if(fseek(fp,statbuf.st_size-0x100,SEEK_SET)<0)
        goto upgrade_fail;
    do
    {
        if(feof(fp))
        {
            printf("Cannot find end of file\n");
            goto upgrade_fail;
        }
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xd)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!=0xa)
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!='-')
            continue;
        c= fgetc(fp);
        printf("%c", c);
        if (c!='-')
            continue;
        break;
    }while(1);
    endPos=ftell(fp);

    /* check header and checksum of this image */
    printf("endPos=%u startPos=%u\n",endPos,startPos);
    if (_checkImage(fp, startPos, endPos-startPos-4) != 0)
        goto upgrade_fail;

    if (fp)
      fclose(fp);
    //strcat(path, "fwuperr1.html"); //Yili 2006/1/16
    //websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    doupgrade_cgi(wp, path, query);

    return;

upgrade_fail:
    if (fp)
      fclose(fp);
    strcat(path, "fwuperr1.html");
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    sys_restart();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

    return;
}
#endif

void
doupgrade_cgi(webs_t wp, char_t *path, char_t **query)
{

    int len;
    char path[64] = "/www/";

    FILE    *fp=NULL;

    printf("upgrade_file_name = %s\n", upgrade_file_name);
    fp = fopen(upgrade_file_name, "rb");
    if (!fp)
    {
        printf("[Assertion fail!] fp=%d\n", fp);
        goto up_fail;
    }
    //do_kill("/var/run/checkwan.pid");

    httpc_info_save();

    //strcpy(path, "/var/tmp/erase_proc.html");
    strcpy(path, "/www/upgrade_proc.html");
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    close(wp->fd);

    /* update firmware */
    printf("Update Firmsize %d \n",endPos-4-startPos);

    fseek(fp,startPos,SEEK_SET);

    setLedLight(0/*TEST_LED*/, 0/*LED_ON*/);
    printf("Start Flash Writing . . .\n");

    flashdrv_filewrite(fp,endPos-4-startPos,(void*)FLASH_MAP_RUN_IMAGE_ADDR);
    //fclose(fp);
    //printf("Flash write OK, reboot to activate new firmware.\n");

    //unlink(wp->post_file_name);
    //free(wp->post_file_name);
    //wp->post_file_name=NULL;

    //close(wp->post_data_fd);
    //wp->post_data_fd = -1;

    //strcpy(path, "/var/tmp/erase_proc.html");
    //strcpy(path, "/www/erase_proc.html");
    //websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    //close(wp->fd);

    //wp->status = 10/*REBOOT*/;
    //sys_reboot();
    reboot_();
    return;

up_fail:
    //strcat(path, "fwuperr.html");
    strcat(path, "fwuperr1.html"); //Yili 2006/1/16
    websDefaultHandler(wp, urlPrefix, webDir, arg, url, path , query);
    sys_restart();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

    return;
}

#define ETHER_ADDR_LEN      6
/*
 * Convert Ethernet address string representation to binary data
 * @param   a   string in xx:xx:xx:xx:xx:xx notation
 * @param   e   binary data
 * @return  TRUE if conversion was successful and FALSE otherwise
 */
int
ether_atoe(const char *a, unsigned char *e)
{
    char *c = (char *) a;
    int i = 0;

    memset(e, 0, ETHER_ADDR_LEN);
    for (;;) {
        e[i++] = (unsigned char) strtoul(c, &c, 16);
        if (!*c++ || i == ETHER_ADDR_LEN)
            break;
    }
    return (i == ETHER_ADDR_LEN);
}


/*
 * Convert Ethernet address binary data to string representation
 * @param   e   binary data
 * @param   a   string in xx:xx:xx:xx:xx:xx notation
 * @return  a
 */
char *
ether_etoa(const unsigned char *e, char *a)
{
    char *c = a;
    int i;

    for (i = 0; i < ETHER_ADDR_LEN; i++) {
        if (i)
            *c++ = ':';
        c += sprintf(c, "%02X", e[i] & 0xff);
    }
    return a;
}

char * trim(char *s)
{
    /* trim trailing whitespace */
    while (isspace(s[strlen(s)-1]))
        s[strlen(s)-1]='\0';

    /* trim leading whitespace */
    memmove(s, &s[strspn(s, " \n\r\t\v")], strlen(s));
    return s;
}


//#define MAX_DHCPC 4//ATTACHED_DEVICES
/* #defined in board.h */
#define MAX_DHCPC   MAX_RESERVED_FIXED  /* 30 */
int dhcpc_res_cur_index = MAX_DHCPC;

static void
validate_lan_setting(webs_t wp, char *value, struct variable *v)
{
    struct variable fields[] = {
        { name: "LANIPAddr", longname: "LAN IP Address" },
        { name: "LANSubnetMask", longname: "LAN Subnet Mask" },
    };

    struct variable rip_fields[] = {
        { name: "RIPVer", longname: "RIP Ver" },
        { name: "RIPDir", longname: "RIP Dir" },
    };

    struct variable dhcp_fields[] = {
        { name: "DHCPServer", longname: "DHCP Server" },
    };

    char *lan_ipaddr, *lan_netmask;
    struct in_addr ipaddr, netmask, netaddr, broadaddr;
    // dmz ip address need not be changed.
    // char *lan_ipaddrs[] = { "dhcp_start", "dhcp_end", "dmz_ipaddr" };
    char *lan_ipaddrs[] = { "dhcp_start", "dhcp_end" };
    int i;
    char *ptr, *ptr1, *ptr2, *ptr3;
    char tmp[20];
    int need_adjust_related_LAN_IP = 0;

    ptr = websGetVar(wp, "Button", NULL);
    if (!ptr)
        return;
    if (strstr(ptr, "Add")) {
        /* add dhcpc res */
        char path[64] = "/www/linkip.html";

        //dhcpc_res_cur_index = MAX_DHCPC;//ATTACHED_DEVICES
        dhcpc_res_cur_index = atoi(ptr);    //ATTACHED_DEVICES
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        noread = 2;
        return;

    }
    else if (strstr(ptr, "Edit") || strstr(ptr, "Modify")) {
        /* edit dhcpc res */
        char path[64] = "/www/addfixed.html";

        ptr = websGetVar(wp, "R1", NULL);//0104
        //dhcpc_res_cur_index = atoi(ptr);
        printf("ptr = %s\n", ptr);
        //if (ptr != '-1') {
        if(!strstr(ptr, "-1")){
            dhcpc_res_cur_index = atoi(ptr);
            printf("dhcpc_res_cur_index = %d\n", dhcpc_res_cur_index);
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            noread = 2;
        }
        else {
            ptr = websGetVar(wp, "page", "");
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                    else
                        sprintf(return_page, "/%s", ptr);
            strcpy(path, "/www/");
            strcat(path, "edit_fixed_err.html");
            noread = 1;
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }
        return;
    }
    else if (strstr(ptr, "Del")) {
        /* delete dhcpc res */
        char path[64] = "/www/lanform.html";

        ptr = websGetVar(wp, "R1", NULL);//0104

        //if (ptr) {
        if(!strstr(ptr, "-1")){

            dhcpc_res_cur_index = atoi(ptr);

            //sprintf(tmp, "dhcpc_res%d", dhcpc_res_cur_index);//ATTACHED_DEVICES
            //nvram_set(tmp, "");//ATTACHED_DEVICES
            validate_dhcpc_res_deleting(dhcpc_res_cur_index);//ATTACHED_DEVICES
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }
        else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "del_fixed_err.html");
                noread = 1;
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }
        return;

    }
    else {      /* apply */
    /* Basic validation */
    if (!(lan_ipaddr = websGetVar(wp, fields[0].name, NULL)) ||
        !(lan_netmask = websGetVar(wp, fields[1].name, NULL))) /* || */
        //!valid_ipaddr(wp, lan_ipaddr, &fields[0]) ||
        //!valid_ipaddr(wp, lan_netmask, &fields[1]))
        return;

    /* Check for broadcast or network address */
    (void) inet_aton(lan_ipaddr, &ipaddr);
    (void) inet_aton(lan_netmask, &netmask);
    netaddr.s_addr = ipaddr.s_addr & netmask.s_addr;
    broadaddr.s_addr = netaddr.s_addr | ~netmask.s_addr;
    if (ipaddr.s_addr == netaddr.s_addr) {
        websBufferWrite(wp, "Invalid <b>%s</b> %s: cannot be the network address<br>",
              fields[0].longname, lan_ipaddr);
        return;
    }
    if (ipaddr.s_addr == broadaddr.s_addr) {
        websBufferWrite(wp, "Invalid <b>%s</b> %s: cannot be the broadcast address<br>",
              fields[0].longname, lan_ipaddr);
        return;
    }

    /* LAN ipaddr is changed
       clear httpc info */
    if ((unsigned int)(ipaddr.s_addr) != *((unsigned int *)&(pRomeCfgParam->ifCfgParam[1].ipAddr))) {
        clr_http_client();
    }

    /* netmask changed */
    if ((unsigned int)(ipaddr.s_addr & netmask.s_addr) !=
        ((*((unsigned int *)&(pRomeCfgParam->ifCfgParam[1].ipAddr))) & (*((unsigned int *)&(pRomeCfgParam->ifCfgParam[1].ipMask)))) ) {
        need_adjust_related_LAN_IP = 1;
    }
    /* netmask changed */
    if ((unsigned int)(ipaddr.s_addr & 0x0ffffff00L) !=
        ((*((unsigned int *)&(pRomeCfgParam->ifCfgParam[1].ipAddr))) & (0x0ffffff00L)) ) {
        need_adjust_related_LAN_IP = 1;
    }

    nvram_set("lan_ipaddr", lan_ipaddr);
    nvram_set("lan_netmask", lan_netmask);

    if (need_adjust_related_LAN_IP)
        adjust_related_LAN_IP();

    /* Fix up LAN IP addresses */
    for (i = 0; i < ARRAYSIZE(lan_ipaddrs); i++) {
        value = nvram_get(lan_ipaddrs[i]);
        if (value && *value) {
            (void) inet_aton(value, &ipaddr);
            ipaddr.s_addr &= ~netmask.s_addr;
            ipaddr.s_addr |= netaddr.s_addr;
            nvram_set(lan_ipaddrs[i], inet_ntoa(ipaddr));
        }
    }
    ptr = websGetVar(wp, rip_fields[0].name, NULL);
    if (ptr != NULL)
        nvram_set("rip_ver", ptr);
    ptr = websGetVar(wp, rip_fields[1].name, NULL);
    if (ptr != NULL)
        nvram_set("rip_dir", ptr);

    ptr = websGetVar(wp, dhcp_fields[0].name, "0");
    nvram_set("lan_proto", (*ptr == '1') ? "dhcp" : "static");

    ptr = websGetVar(wp, "PStart1", NULL);
    ptr1 = websGetVar(wp, "PStart2", NULL);
    ptr2 = websGetVar(wp, "PStart3", NULL);
    ptr3 = websGetVar(wp, "PStart4", NULL);

    if ((ptr == NULL) || (ptr1 == NULL) || (ptr2 == NULL) || (ptr3 == NULL)) {
        websBufferWrite(wp, "Invalid <b>%s</b> %s: cannot be the network address<br>",
              "DHCP Start", NULL);
        return;
    }

    sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
    nvram_set("dhcp_start", tmp);

    ptr = websGetVar(wp, "PEnd1", NULL);
    ptr1 = websGetVar(wp, "PEnd2", NULL);
    ptr2 = websGetVar(wp, "PEnd3", NULL);
    ptr3 = websGetVar(wp, "PEnd4", NULL);

    if ((ptr == NULL) || (ptr1 == NULL) || (ptr2 == NULL) || (ptr3 == NULL)) {
        websBufferWrite(wp, "Invalid <b>%s</b> %s: cannot be the network address<br>",
              "DHCP End", NULL);
        return;
    }

    sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
    nvram_set("dhcp_end", tmp);

    lan_cfg_save();

    //dhcps_start();
    /* kill old udhcpd if it exists */
    /* backup lease info */
    send_signal("/var/run/udhcpd.pid", SIGUSR1);
    //sleep(1);
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
    //sys_restart();
    if (need_adjust_related_LAN_IP) {
        //Power Down phy, then turn on whne call f#httpc_init
extern uint16 g_lan_ip_changed;
        g_lan_ip_changed = 1;
    }
    action = RESTART;

    }
}



//ATTACHED_DEVICED
static void
validate_dhcpc_res_deleting(number)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6;
    int i, j = 0, index = 1;
    char tmp[100], nvram_name[20];
    char *c;
    unsigned char mac[20], cnt = 0;
    int ret = 0, NoOfDhcpcRsvdAddrFound = 0, dhcpcRsvdAddrCk = 0, dhcpcAvailable = -1;

    cnt = dhcpcRsvdAddrCnt;
    if ( cnt == 0) {
        return;
    }
    else
    {
        for (i = 0; i < cnt; i++) {
            if (number == i) {
                    printf("Delete!\n");
                    j = j + 1;
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_host, _dhcpc_rsvd_addr[j].rsvd_host, sizeof(_dhcpc_rsvd_addr[++j].rsvd_host));
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_mac, _dhcpc_rsvd_addr[j].rsvd_mac, sizeof(_dhcpc_rsvd_addr[++j].rsvd_mac));
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_ip, _dhcpc_rsvd_addr[j].rsvd_ip, sizeof(_dhcpc_rsvd_addr[++j].rsvd_ip));

                    dhcpcRsvdAddrCnt--;
                    j++;
            }
            else {
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_host, _dhcpc_rsvd_addr[j].rsvd_host, sizeof(_dhcpc_rsvd_addr[j].rsvd_host));
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_mac, _dhcpc_rsvd_addr[j].rsvd_mac, sizeof(_dhcpc_rsvd_addr[j].rsvd_mac));
                    memcpy(_dhcpc_rsvd_addr[i].rsvd_ip, _dhcpc_rsvd_addr[j].rsvd_ip, sizeof(_dhcpc_rsvd_addr[j].rsvd_ip));
                    j++;
            }
            }
    }

    //printf("dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    for ( i = 0; i < dhcpcRsvdAddrCnt; i++) {
        sprintf(tmp, "%s,%s,%s",
        _dhcpc_rsvd_addr[i].rsvd_ip,
        _dhcpc_rsvd_addr[i].rsvd_mac,
        _dhcpc_rsvd_addr[i].rsvd_host );
        if (i >= MAX_DHCPC) {
            dhcpcAvailable = get_free_dhcpc_res_entry();
            if ((dhcpcAvailable < 0) || (dhcpcAvailable >= MAX_DHCPC)) {
                printf("no free dhcpc res entry\n");
                return;
            }
        }
        sprintf(nvram_name, "dhcpc_res%d", i);
        nvram_set(nvram_name, tmp);

    }

    while ( i < MAX_DHCPC) {
        //printf("i = %d\n", i);
        sprintf(nvram_name, "dhcpc_res%d", i++);
        nvram_set(nvram_name, "");
    }

    //dhcps_cfg_save();
    //dhcps_start();


}

extern uint GetCurWanConnType(void);
extern void WanConfigSaveToFlash(uint connType);
int wizard_wan_spoofmac_detected = 0;

static void
validate_wan_setting(webs_t wp, char *value, struct variable *v)
{
    char *encap;
    char *ptr, *ptr1, *ptr2, *ptr3;
    char *Test;
    int i;
    char tmp[61];

    struct variable fields[] = {

        { name: "SysName", longname: "WAN hostname", nvram_name: "wan_hostname" },
        //{ name: "Domain", longname: "WAN domain", nvram_name: "wan_domain" },
//Yili 2006/1/6
#if defined(DNI)
        { name: "Domain", longname: "WAN domain", nvram_name: "lan_domain" },
        //{ name: "LAN_Domain", longname: "LAN domain", nvram_name: "lan_domain" },
#endif

        { name: "PoeUserName", longname: "PPPOE username", nvram_name: "wan_pppoe_username" },
        { name: "PoePasswd", longname: "PPPOE password", nvram_name: "wan_pppoe_passwd" },
        { name: "PoeServName", longname: "PPPOE service name", nvram_name: "wan_pppoe_service" },
        { name: "PoeIdleTime", longname: "PPPOE idle timeout", nvram_name: "wan_pppoe_idletime" },
        { name: "PoeDemand", longname: "PPPOE dial on demand", nvram_name: "wan_pppoe_demand" },
        //{ name: "PoeFixedIPAssign", longname: "PPPOE assign fixed ip", nvram_name: "wan_pppoe_dynamicassign" },
        { name: "PoeFixedIP", longname: "PPPOE fixed ip", nvram_name: "wan_pppoe_fixedipAddr" },

        { name: "PPTPUserName", longname: "PPTP username", nvram_name: "wan_pptp_username" },
        { name: "PPTPPasswd", longname: "PPTP password", nvram_name: "wan_pptp_password" },
        { name: "PPTPConnectName", longname: "PPTP Connect ID", nvram_name: "wan_pptp_conn_id" },
        { name: "PPTPIdleTime", longname: "PPTP idle timeout", nvram_name: "wan_pptp_idle_interval" },
        { name: "PPTPIdleDisconn", longname: "PPTP dial on demand", nvram_name: "wan_pptp_idle_disconnect" },

        { name: "L2TPUserName", longname: "L2TP username", nvram_name: "wan_l2tp_username" },
        { name: "L2TPPasswd", longname: "L2TP password", nvram_name: "wan_l2tp_password" },
        { name: "L2TPConnectName", longname: "L2TP Connect ID", nvram_name: "wan_l2tp_conn_id" },
        { name: "L2TPIdleTime", longname: "L2TP idle timeout", nvram_name: "wan_l2tp_idle_interval" },
        { name: "L2TPIdleDisconn", longname: "L2TP dial on demand", nvram_name: "wan_l2tp_idle_disconnect" },


        { name: "TelUserName", longname: "Telstra username", nvram_name: "wan_tel_username" },
        { name: "TelPasswd", longname: "Telstra password", nvram_name: "wan_tel_password" },
        { name: "TelServName", longname: "Telstra service name", nvram_name: "wan_tel_service" },
        { name: "TelIdleTime", longname: "Telstra idle timeout", nvram_name: "wan_tel_idle_interval" },
        { name: "TelDemand", longname: "Telstra dial on demand", nvram_name: "wan_tel_demand" },

        { name: "DNSAssign", longname: "DNS auto", nvram_name: "wan_autodns" },
        { name: "DNSAddr1", longname: "DNS Promary", nvram_name: "wan_dns1" },
        { name: "DNSAddr2", longname: "DNS Secondary", nvram_name: "wan_dns2" },

        { name: "passwd", longname: "Set password", nvram_name: "http_passwd" },
        
        { name: "wan_pppoe_conn_mode", longname: "Set connection mode", nvram_name: "wan_pppoe_conn_mode" },
        { name: "wan_pptp_conn_mode", longname: "Set connection mode", nvram_name: "wan_pptp_conn_mode" },
        { name: "wan_tel_conn_mode", longname: "Set connection mode", nvram_name: "wan_tel_conn_mode" },

    };
    int cur_connType;


    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
          if (strcmp(fields[i].nvram_name, "wan_autodns") == 0) {
            if (strcmp(ptr, "Dynamic") == 0)
                nvram_set(fields[i].nvram_name, "1");
            else
                nvram_set(fields[i].nvram_name, "0");
          }
          else
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

	ptr = websGetVar(wp, "wan_pppoe_conn_mode", NULL);
	if(ptr)
		nvram_set("wan_pppoe_conn_mode", ptr);
	ptr = websGetVar(wp, "wan_pptp_conn_mode", NULL);
	if(ptr)
		nvram_set("wan_pptp_conn_mode", ptr);
	ptr = websGetVar(wp, "wan_tel_conn_mode", NULL);
	if(ptr)
		nvram_set("wan_tel_conn_mode", ptr);

    ptr = websGetVar(wp, "InternetAssign", NULL);
    if (ptr) {
        if (strcmp(ptr, "Dynamic") == 0)
            nvram_set("wan_pppoe_dynamicassign", "1");
        else
        {
            nvram_set("wan_pppoe_dynamicassign", "0");

            ptr = websGetVar(wp, "InAddr", NULL);
            nvram_set("wan_pppoe_fixedipAddr", ptr);
        }
    }

    ptr = websGetVar(wp, "PoeFixedIPAssign", NULL);
    if (ptr) {
        if (strcmp(ptr, "1") == 0)
            nvram_set("wan_pppoe_dynamicassign", "0");
        else
        {
            nvram_set("wan_pppoe_dynamicassign", "1");
        }
    }


    ptr = websGetVar(wp, "PPTPMyAddr1", NULL);
    ptr1 = websGetVar(wp, "PPTPMyAddr2", NULL);
    ptr2 = websGetVar(wp, "PPTPMyAddr3", NULL);
    ptr3 = websGetVar(wp, "PPTPMyAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_pptp_ipaddr", tmp);
        nvram_set("wan_pptp_netmask", "255.255.255.0");
    }
#if defined(DHCPPPTP)
        ptr = websGetVar(wp, "wan_pptp_server_ipaddr", NULL);
	if (!ptr) { /* for old smart and CA PPTP pages */	
		ptr = websGetVar(wp, "PPTPAddr1", NULL);
	        ptr1 = websGetVar(wp, "PPTPAddr2", NULL);
        	ptr2 = websGetVar(wp, "PPTPAddr3", NULL);
	        ptr3 = websGetVar(wp, "PPTPAddr4", NULL);
    		if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        		sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
			ptr = tmp;	
		}
	}
		
        if (ptr) {
        	nvram_set("wan_pptp_server_ipaddr", ptr);

        //struct in_addr svaddr;
        //if(inet_aton(ptr, &svaddr) == 0)

#else
    ptr = websGetVar(wp, "PPTPAddr1", NULL);
    ptr1 = websGetVar(wp, "PPTPAddr2", NULL);
    ptr2 = websGetVar(wp, "PPTPAddr3", NULL);
    ptr3 = websGetVar(wp, "PPTPAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_pptp_server_ipaddr", tmp);
#endif

        /* check pptp netmask */
        {

#if 1
            int myip4;
            ptr = websGetVar(wp, "PPTPMyAddr1", "");
            myip4 = atoi(ptr);
            if (myip4 < 0x80) /* MSB 0X */
                nvram_set("wan_pptp_netmask", "255.0.0.0"); /* class A MSB 0X*/
            else if (myip4 < 0xc0) /* 10X */
                nvram_set("wan_pptp_netmask", "255.255.0.0"); /* class B 10X*/
            else if (myip4 < 0xdf) /* 110X */
                nvram_set("wan_pptp_netmask", "255.255.255.0"); /* class C 110X*/

#else
            char *lan_ipaddr, *lan_gwaddr;
            struct in_addr ipaddr, netmask, gwaddr;

            lan_ipaddr = nvram_get("wan_pptp_ipaddr");
            (void) inet_aton(lan_ipaddr, &ipaddr);
            lan_gwaddr = nvram_get("wan_pptp_server_ipaddr");
            (void) inet_aton(lan_gwaddr, &gwaddr);

            (void) inet_aton("255.255.255.0", &netmask);
            if ((ipaddr.s_addr & netmask.s_addr) == (gwaddr.s_addr & netmask.s_addr)) {
                nvram_set("wan_pptp_netmask", "255.255.255.0");
            } else {
                (void) inet_aton("255.255.0.0", &netmask);
                if ((ipaddr.s_addr & netmask.s_addr) == (gwaddr.s_addr & netmask.s_addr)) {
                    nvram_set("wan_pptp_netmask", "255.255.0.0");
                }
                else
                    nvram_set("wan_pptp_netmask", "255.0.0.0");
            }
#endif
        }
    }

    ptr = websGetVar(wp, "PPTPGAddr1", NULL);
    ptr1 = websGetVar(wp, "PPTPGAddr2", NULL);
    ptr2 = websGetVar(wp, "PPTPGAddr3", NULL);
    ptr3 = websGetVar(wp, "PPTPGAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_pptp_gateway_ipaddr", tmp);
    }

    ptr = websGetVar(wp, "L2TPMyAddr1", NULL);
    ptr1 = websGetVar(wp, "L2TPMyAddr2", NULL);
    ptr2 = websGetVar(wp, "L2TPMyAddr3", NULL);
    ptr3 = websGetVar(wp, "L2TPMyAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_l2tp_ipaddr", tmp);
        nvram_set("wan_l2tp_netmask", "255.255.255.0");
    }

    ptr = websGetVar(wp, "L2TPAddr1", NULL);
    ptr1 = websGetVar(wp, "L2TPAddr2", NULL);
    ptr2 = websGetVar(wp, "L2TPAddr3", NULL);
    ptr3 = websGetVar(wp, "L2TPAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_l2tp_server_ipaddr", tmp);

        /* check l2tp netmask */
        {
            char *lan_ipaddr, *lan_gwaddr;
            struct in_addr ipaddr, netmask, gwaddr;

            lan_ipaddr = nvram_get("wan_l2tp_ipaddr");
            (void) inet_aton(lan_ipaddr, &ipaddr);
            lan_gwaddr = nvram_get("wan_l2tp_server_ipaddr");
            (void) inet_aton(lan_gwaddr, &gwaddr);

            (void) inet_aton("255.255.255.0", &netmask);
            if ((ipaddr.s_addr & netmask.s_addr) == (gwaddr.s_addr & netmask.s_addr)) {
                nvram_set("wan_l2tp_netmask", "255.255.255.0");
            } else {
                (void) inet_aton("255.255.0.0", &netmask);
                if ((ipaddr.s_addr & netmask.s_addr) == (gwaddr.s_addr & netmask.s_addr)) {
                    nvram_set("wan_l2tp_netmask", "255.255.0.0");
                }
                else
                    nvram_set("wan_l2tp_netmask", "255.0.0.0");
            }
        }
    }

    ptr = websGetVar(wp, "L2TPGAddr1", NULL);
    ptr1 = websGetVar(wp, "L2TPGAddr2", NULL);
    ptr2 = websGetVar(wp, "L2TPGAddr3", NULL);
    ptr3 = websGetVar(wp, "L2TPGAddr4", NULL);

    if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
        sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
        nvram_set("wan_l2tp_gateway_ipaddr", tmp);
    }

    ptr = websGetVar(wp, "MACAssign", NULL);
    if (ptr) {
        if (strcmp(ptr, "default") == 0) {
            /*if  (atoi(nvram_get("wizard_wan_spoofmac_detected")) && atoi(nvram_get("wan_mac_mode")) == 2) {
                printf("888888888888888888888888888888888888888888888888888\n");
                eth0_with_wan_mac_up();
                nvram_set("wan_mac_mode", "0");
                nvram_set("wizard_wan_spoofmac_detected", "0");
            }
            else*/
                nvram_set("wan_mac_mode", "0");
        }
        else if (strcmp(ptr, "thisPC") == 0)
            nvram_set("wan_mac_mode", "1");
        else if (strcmp(ptr, "spoofmac") == 0)
            nvram_set("wan_mac_mode", "2");
    }
    //printf("wizard_wan_spoofmac_detected = %d\n", wizard_wan_spoofmac_detected);
    if (wizard_wan_spoofmac_detected == 1) {
            nvram_set("wan_mac_mode", "2");
    }

    ptr = nvram_safe_get("wan_mac_mode");
    if (strcmp(ptr, "1") == 0) {
        char str_clientip[20];
        char *str_mac;
        struct sockaddr_in peer;
        int    peer_len=sizeof(peer);
        int sockfd;

        sockfd = wp->fd;
        if (getpeername(sockfd,(struct sockaddr*)&peer,&peer_len ) >= 0)
        {
            strcpy(str_clientip, inet_ntoa(peer.sin_addr));
            str_mac = get_http_client_hwaddr(str_clientip);
            nvram_set("wan_spoofmac", str_mac);
            nvram_set("wan_mac_mode", "2");
        }

    }
    else if (strcmp(ptr, "2") == 0) {
        if (wizard_wan_spoofmac_detected == 0) {
        ptr = websGetVar(wp, "Spoofmac", NULL);
            if (ptr) {
            nvram_set("wan_spoofmac", ptr);
    }
        }
        wizard_wan_spoofmac_detected = 0;
    }
#if 0
    //Jenny 2006/01/05
    ptr1 = websGetVar(wp, "Resolution", NULL);
    ptr2 = websGetVar(wp, "Threshold", NULL);

    //printf("####### Resolution=%d Threshold=%d\r\n",atoi(ptr1),atoi(ptr2));

    if (ptr1 && ptr2)
    {

        if((atoi(ptr1)== POT_GetResolution()) && (atoi(ptr2) == POT_GetRepackThreshold()))
        {
         ;
        }
        else
        {
        //printf("####### POT_factoryDefault\r\n");
        POT_factoryDefault(atoi(ptr1), atoi(ptr2));
        }
    }
        //end
#endif


    encap = websGetVar(wp, "Encap", NULL);
    if (encap == NULL)
        return;
    switch (atoi(encap)) {
        case 0 :
            ptr = websGetVar(wp, "WANAssign", NULL);
            if (ptr) {
                if (strcmp(ptr, "Dynamic") == 0){
                    nvram_set("wan_proto", "dhcp");
                    ptr = websGetVar(wp, "WANIPethr", NULL);
                    ptr1 = websGetVar(wp, "WANMask", NULL);
                    ptr2 = websGetVar(wp, "WANGateway", NULL);
                    if ((ptr) && (ptr1) && (ptr2)) {
                        nvram_set("wan_ipaddr", ptr);
                        nvram_set("wan_netmask", ptr1);
                        nvram_set("wan_gateway", ptr2);
                    }
                }
                else {
                    ptr = websGetVar(wp, "WANIPethr", NULL);
                    ptr1 = websGetVar(wp, "WANMask", NULL);
                    ptr2 = websGetVar(wp, "WANGateway", NULL);
                    if ((ptr) && (ptr1) && (ptr2)) {
                        nvram_set("wan_proto", "static");
                        nvram_set("wan_ipaddr", ptr);
                        nvram_set("wan_netmask", ptr1);
                        nvram_set("wan_gateway", ptr2);
                    }
                }
            }
            break;
        case 1:
            nvram_set("wan_proto", "pppoe");
            break;
        case 2:
            nvram_set("wan_proto", "pptp");
            break;
//Yili 2006/1/6/
#if defined(DNI)
        case 3:
            nvram_set("wan_proto", "telstra");
            break;
        case 4:
            nvram_set("wan_proto", "l2tp");
            break;
#endif
    }

#if 1   //Kevin:06Feb17:TestButton
    Test = websGetVar(wp, "DoTest", "");

    if(!strcmp(Test,"yes")) {
        char* proto;
        proto=nvram_get("wan_proto");
        if(!strcmp(proto,"pppoe")) { nvram_set("wan_pppoe_demand", "0"); }
        if(!strcmp(proto,"pptp")) { nvram_set("wan_pptp_idle_disconnect", "0"); }
        if(!strcmp(proto,"telstra")) { nvram_set("wan_tel_demand", "0"); }
        if(!strcmp(proto,"l2tp")) { nvram_set("wan_l2tp_idle_disconnect", "0"); }
    }
#endif
    if (atoi(nvram_get("wan_autodns")) == 0) {
        ptr = websGetVar(wp, "DNSAddr1", NULL);
        if ((ptr == NULL) || (strcmp(ptr, "...") == 0)) {
            nvram_set("wan_dns1", "0.0.0.0");
        }
        ptr = websGetVar(wp, "DNSAddr2", NULL);
        if ((ptr == NULL) || (strcmp(ptr, "...") == 0)) {
            nvram_set("wan_dns2", "0.0.0.0");
        }
    }
    dhcpc_cfg_save();
    cur_connType=GetCurWanConnType();
    WanConfigSaveToFlash(cur_connType);
    action = RESTART;
    //sysReinit(past_connType);

}

static void
validate_fw_site_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

    struct variable fields[] = {

        { name: "KeywordEnable", longname: "KeywordEnable", nvram_name: "fw_keywordenable" },
        { name: "TrustIP", longname: "TrustIP", nvram_name: "fw_trustip" },

    };

    ptr = websGetVar(wp, "Button", NULL);
    if (ptr) {
        if (strstr(ptr, "Apply")) {
            //urlfilter_tableDriverdel();
            for (i = 0; i < ARRAYSIZE(fields); i++) {
                ptr = websGetVar(wp, fields[i].name, NULL);
                if (ptr) {
                    nvram_set(fields[i].nvram_name, ptr);
                }
            }
            ptr = websGetVar(wp, "TrustIpEnble", "0");
            nvram_set("fw_trustipenable", ptr);
            fw_blocksitet_cfg_save();
            schedule_cfg_save();
            trusted_cfg_save();

            urlfilter_tableDriverAccess();
            acl_tableDriverAccess();
            remoteMgmt_tableDriverAccess();
            email_tableDriverAccess();
            rtl8651_flushUrlFilterTrustedUser();
#if 0
            for(i=0;i<URL_FILTER_MAX_TRUSTED_USERS;i++)
            {
                if((pRomeCfgParam->urlfilterTrustedUserCfgParam[i].trustedIP!=0) && (pRomeCfgParam->urlfilterTrustedUserCfgParam[i].enable==1))
                    rtl8651_addUrlFilterTrustedUser(pRomeCfgParam->urlfilterTrustedUserCfgParam[i].trustedIP);
            }
#endif
            if((pRomeCfgParam->urlfilterTrustedUserCfgParam[0].trustedIP!=0) && (pRomeCfgParam->urlfilterTrustedUserCfgParam[0].enable==1))
                rtl8651_addUrlFilterTrustedUser(pRomeCfgParam->urlfilterTrustedUserCfgParam[0].trustedIP);
            //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

            //Timo 0215
            if(atoi(nvram_get("fw_keywordenable"))==0)
            {//get after restore
                FILE *fp;

                fp=fopen("/var/url.log", "r");
                if(fp)
                    unlink("/var/url.log");
            }

            return;
        }
        else if (strstr(ptr, "Add")) {
            /* add keyword */
            ptr = websGetVar(wp, "AddKeyword", NULL);
            if (ptr) {
              nvram_set("add_block_keyword_domain", ptr);
              /* need-insert */
            }
        }
        else if (strstr(ptr, "Del")) {
            /* del keyword */
            ptr = websGetVar(wp, "AddKeyword", NULL);
            if (ptr) {
              nvram_set("del_block_keyword_domain", ptr);
            }
        }
        else if (strstr(ptr, "Clear")) {
            /* clr all keyword */
            nvram_set("clr_block_keyword_domain", NULL);
        }
    }
    {
        char path[64] = "/www/";

        ptr = websGetVar(wp, "page", "");
        strcat(path, ptr);
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
    }

}

//#define MAX_ACL       8
int filter_client_cur_index = MAX_ACL;
static void
validate_fw_service_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

    struct variable fields[] = {

        { name: "Packet", longname: "Packet", nvram_name: "fw_packenable" },
        //{ name: "countp", longname: "countp", nvram_name: "fw_packrulenumber" },

    };

    ptr = websGetVar(wp, "Button", NULL);
    if (!ptr)
        return;
    if (strstr(ptr, "Apply")) {

        for (i = 0; i < ARRAYSIZE(fields); i++) {
            ptr = websGetVar(wp, fields[i].name, NULL);
            if (ptr) {
                nvram_set(fields[i].nvram_name, ptr);
            }
        }
        fw_filter_client_cfg_save();
        /* access table driver, return value not checked ? */
        urlfilter_tableDriverAccess();
        acl_tableDriverAccess();
        remoteMgmt_tableDriverAccess();
        email_tableDriverAccess();//Timo Wu 2005/1/12
        //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

    }
    else if (strstr(ptr, "Add")) {
        /* add service block */
        char path[64] = "/www/addpacketfilter.html";

        filter_client_cur_index = MAX_ACL;
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html
        noread = 2;

    }
    else if (strstr(ptr, "Edit") || strstr(ptr, "Modify")) {
        /* edit service */
        char path[64] = "/www/addpacketport.html";
        //char path[64] = "/www/addpacketfilter.html";

        ptr = websGetVar(wp, "X1", NULL);

        if (ptr) {

            filter_client_cur_index = atoi(ptr);
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html
            noread = 2;
        }
        else {
            ptr = websGetVar(wp, "page", "");
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                    else
                        sprintf(return_page, "/%s", ptr);
            strcpy(path, "/www/");
            strcat(path, "edit_client_err.html");
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }

    }
    else if (strstr(ptr, "Del")) {
        char path[64] = "/www/";

        /* delete this service */
        ptr = websGetVar(wp, "X1", NULL);

        if (ptr) {
            i = atoi(ptr);
            sprintf(tmp, "filter_client%d", i);
            nvram_set(tmp, "");
            ptr = websGetVar(wp, "page", "");
            strcat(path, ptr);
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            //fw_filter_client_cfg_save();
            /* access table driver, return value not checked ? */
            //urlfilter_tableDriverAccess();
            //acl_tableDriverAccess();
            //email_tableDriverAccess();//Timo Wu 2005/1/12
            //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
        }
        else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "del_client_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }

    }
    else {
        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
                else
                    sprintf(return_page, "/%s", ptr);
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, return_page , NULL);
    }

}

/* %s-%s:%d-%d,%s,%d,%d,%d,%d,%d,%d,%d:%d-%d,%s,%s */
static void
validate_fw_service_adding(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4;
    char *c;
    int i;
    char tmp[200];
    char nvram_name[20], apply_proto[10];
    char *lan_ipaddr0, *lan_ipaddr1, *lan_port0, *lan_port1, *proto;
    char *day_start, *day_end, *sec_start, *sec_end;
    char *enable, *desc;
    char filter_ipaddr1[20], filter_ipaddr2[20];

#if 0
    if( validate_fw_service_cmp(wp) == 1){
        char path[64] = "/www/error_portexist.html";
        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            }
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        return;
    }
#endif

    ptr = websGetVar(wp, "UserDefined", NULL);
    if (!ptr)
        ptr = websGetVar(wp, "TYPE", NULL);
    ptr1 = websGetVar(wp, "Protocol", NULL);
    if (!ptr || !ptr1)
        return;
    //ptr2 = websGetVar(wp, "Portstart", NULL);
    //ptr3 = websGetVar(wp, "Portend", NULL);
    ptr2 = websGetVar(wp, "verify", NULL);      /* start port */
    ptr3 = websGetVar(wp, "verend", NULL);      /* end port */

    ptr4 = websGetVar(wp, "filteriptype", NULL);

    if (strcmp(ptr1, "TCP/UDP") == 0)
        strcpy(apply_proto, "both");
    else if (strcmp(ptr1, "UDP") == 0)
        strcpy(apply_proto, "udp");
    else
        strcpy(apply_proto, "tcp");

    if (filter_client_cur_index >= MAX_ACL) {   /* add */
        if (strcmp(ptr4, "0") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_PCIP1", "0"),
                    websGetVar(wp, "F_PCIP2", "0"),
                    websGetVar(wp, "F_PCIP3", "0"),
                    websGetVar(wp, "F_PCIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = lan_ipaddr0;
        }
        else if (strcmp(ptr4, "1") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_StartIP1", "0"),
                    websGetVar(wp, "F_StartIP2", "0"),
                    websGetVar(wp, "F_StartIP3", "0"),
                    websGetVar(wp, "F_StartIP4", "0"));
            sprintf(filter_ipaddr2, "%s.%s.%s.%s",
                    websGetVar(wp, "F_EndIP1", "0"),
                    websGetVar(wp, "F_EndIP2", "0"),
                    websGetVar(wp, "F_EndIP3", "0"),
                    websGetVar(wp, "F_EndIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
        else {
            strcpy(filter_ipaddr1, "0.0.0.0");
            strcpy(filter_ipaddr2, "255.255.255.255");
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
        sprintf(tmp, "%s-%s:%s-%s,%s,%s:%s-%s,%s,%s",
            lan_ipaddr0,
            lan_ipaddr1,
            ptr2,       /* start port */
            ptr3,       /* end port */
            apply_proto,//proto,
            "0,0,0,0,0,0,0",
            "0",
            "0",
            "1",
            //desc
            ptr
            );


    }
    else {          /* edit */
        sprintf(nvram_name, "filter_client%d",  filter_client_cur_index);
        c = nvram_safe_get(nvram_name);

        /* Check for port specification */
        lan_port0 = c;
        lan_ipaddr0 = strsep(&lan_port0, ":");
        if (!lan_port0)
            return 0;

        /* Check for protocol specification */
        proto = lan_port0;
        lan_port0 = strsep(&proto, ",");
        if (!proto)
            return 0;

        /* Check for day specification */
        day_start = proto;
        proto = strsep(&day_start, ",");
        if (!day_start)
            return 0;

        /* Check for time specification */
        sec_start = day_start;
        day_start = strsep(&sec_start, ":");
        if (!sec_start)
            return 0;

        /* Check for enable specification */
        enable = sec_start;
        sec_start = strsep(&enable, ",");
        if (!enable)
            return 0;

        /* Check for description specification (optional) */
        desc = enable;
        enable = strsep(&desc, ",");

        /* Check for address range (optional) */
        lan_ipaddr1 = lan_ipaddr0;
        lan_ipaddr0 = strsep(&lan_ipaddr1, "-");
        if (!lan_ipaddr1)
            lan_ipaddr1 = lan_ipaddr0;

        /* Check for port range (optional) */
        lan_port1 = lan_port0;
        lan_port0 = strsep(&lan_port1, "-");
        if (!lan_port1)
            lan_port1 = lan_port0;

        /* Check for day range (optional) */
        day_end = day_start;
        day_start = strsep(&day_end, "-");
        if (!day_end)
            day_end = day_start;

        /* Check for time range (optional) */
        sec_end = sec_start;
        sec_start = strsep(&sec_end, "-");
        if (!sec_end)
            sec_end = sec_start;

        //if (filter_client_cur_index >= MAX_ACL) {
        if (strcmp(ptr4, "0") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                websGetVar(wp, "F_PCIP1", "0"),
                websGetVar(wp, "F_PCIP2", "0"),
                websGetVar(wp, "F_PCIP3", "0"),
                websGetVar(wp, "F_PCIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = lan_ipaddr0;
        }
        else if (strcmp(ptr4, "1") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_StartIP1", "0"),
                    websGetVar(wp, "F_StartIP2", "0"),
                    websGetVar(wp, "F_StartIP3", "0"),
                    websGetVar(wp, "F_StartIP4", "0"));
            sprintf(filter_ipaddr2, "%s.%s.%s.%s",
                    websGetVar(wp, "F_EndIP1", "0"),
                    websGetVar(wp, "F_EndIP2", "0"),
                    websGetVar(wp, "F_EndIP3", "0"),
                    websGetVar(wp, "F_EndIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
        else {
            strcpy(filter_ipaddr1, "0.0.0.0");
            strcpy(filter_ipaddr2, "255.255.255.255");
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }

        sprintf(tmp, "%s-%s:%s-%s,%s,%s:%s-%s,%s,%s",
            lan_ipaddr0,
            lan_ipaddr1,
            ptr2,       /* start port */
            ptr3,       /* end port */
            apply_proto,//proto,
            day_start,
            sec_start,
            sec_end,
            enable,
            //desc
            ptr
            );
    }
    if (filter_client_cur_index >= MAX_ACL) {
        i = get_free_filter_client_entry();
        if ((i < 0) || (i >= MAX_ACL)) {
            printf("no free filter client entry\n");
            return;
        }
        sprintf(nvram_name, "filter_client%d", i);
    }

    nvram_set(nvram_name, tmp);
    //fw_filter_client_cfg_save();
    /* access table driver, return value not checked ? */
    //urlfilter_tableDriverAccess();
    //acl_tableDriverAccess();
    //email_tableDriverAccess();//Timo Wu 2005/1/12
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}
static int
validate_fw_service_cmp(webs_t wp)
{
    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4;
    char *c;
    int i;
    char tmp[200];
    char nvram_name[20], apply_proto[10];
    char *lan_ipaddr0, *lan_ipaddr1, *lan_port0, *lan_port1, *proto;
    char *day_start, *day_end, *sec_start, *sec_end;
    char *enable, *desc;
    char filter_ipaddr1[20], filter_ipaddr2[20],nv_name[20];

    ptr = websGetVar(wp, "UserDefined", NULL);
    if (!ptr)
        ptr = websGetVar(wp, "TYPE", NULL);
    ptr1 = websGetVar(wp, "Protocol", NULL);
    if (!ptr || !ptr1)
        return;
    //ptr2 = websGetVar(wp, "Portstart", NULL);
    //ptr3 = websGetVar(wp, "Portend", NULL);
    ptr2 = websGetVar(wp, "verify", NULL);      /* start port */
    ptr3 = websGetVar(wp, "verend", NULL);      /* end port */

    ptr4 = websGetVar(wp, "filteriptype", NULL);

    if (strcmp(ptr1, "TCP/UDP") == 0)
        strcpy(apply_proto, "both");
    else if (strcmp(ptr1, "UDP") == 0)
        strcpy(apply_proto, "udp");
    else
        strcpy(apply_proto, "tcp");

    if (filter_client_cur_index >= MAX_ACL) {   /* add */
        if (strcmp(ptr4, "0") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_PCIP1", "0"),
                    websGetVar(wp, "F_PCIP2", "0"),
                    websGetVar(wp, "F_PCIP3", "0"),
                    websGetVar(wp, "F_PCIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = lan_ipaddr0;
        }
        else if (strcmp(ptr4, "1") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_StartIP1", "0"),
                    websGetVar(wp, "F_StartIP2", "0"),
                    websGetVar(wp, "F_StartIP3", "0"),
                    websGetVar(wp, "F_StartIP4", "0"));
            sprintf(filter_ipaddr2, "%s.%s.%s.%s",
                    websGetVar(wp, "F_EndIP1", "0"),
                    websGetVar(wp, "F_EndIP2", "0"),
                    websGetVar(wp, "F_EndIP3", "0"),
                    websGetVar(wp, "F_EndIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
        else {
            strcpy(filter_ipaddr1, "0.0.0.0");
            strcpy(filter_ipaddr2, "255.255.255.255");
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
    }
    else {          /* edit */
        sprintf(nvram_name, "filter_client%d",  filter_client_cur_index);
        c = nvram_safe_get(nvram_name);
        strcpy(tmp, c);
        c = tmp;

        /* Check for port specification */
        lan_port0 = c;
        lan_ipaddr0 = strsep(&lan_port0, ":");
        if (!lan_port0)
            return 0;

        /* Check for protocol specification */
        proto = lan_port0;
        lan_port0 = strsep(&proto, ",");
        if (!proto)
            return 0;

        /* Check for day specification */
        day_start = proto;
        proto = strsep(&day_start, ",");
        if (!day_start)
            return 0;

        /* Check for time specification */
        sec_start = day_start;
        day_start = strsep(&sec_start, ":");
        if (!sec_start)
            return 0;

        /* Check for enable specification */
        enable = sec_start;
        sec_start = strsep(&enable, ",");
        if (!enable)
            return 0;

        /* Check for description specification (optional) */
        desc = enable;
        enable = strsep(&desc, ",");

        /* Check for address range (optional) */
        lan_ipaddr1 = lan_ipaddr0;
        lan_ipaddr0 = strsep(&lan_ipaddr1, "-");
        if (!lan_ipaddr1)
            lan_ipaddr1 = lan_ipaddr0;

        /* Check for port range (optional) */
        lan_port1 = lan_port0;
        lan_port0 = strsep(&lan_port1, "-");
        if (!lan_port1)
            lan_port1 = lan_port0;

        /* Check for day range (optional) */
        day_end = day_start;
        day_start = strsep(&day_end, "-");
        if (!day_end)
            day_end = day_start;

        /* Check for time range (optional) */
        sec_end = sec_start;
        sec_start = strsep(&sec_end, "-");
        if (!sec_end)
            sec_end = sec_start;

        //if (filter_client_cur_index >= MAX_ACL) {
        if (strcmp(ptr4, "0") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                websGetVar(wp, "F_PCIP1", "0"),
                websGetVar(wp, "F_PCIP2", "0"),
                websGetVar(wp, "F_PCIP3", "0"),
                websGetVar(wp, "F_PCIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = lan_ipaddr0;
        }
        else if (strcmp(ptr4, "1") == 0) {
            sprintf(filter_ipaddr1, "%s.%s.%s.%s",
                    websGetVar(wp, "F_StartIP1", "0"),
                    websGetVar(wp, "F_StartIP2", "0"),
                    websGetVar(wp, "F_StartIP3", "0"),
                    websGetVar(wp, "F_StartIP4", "0"));
            sprintf(filter_ipaddr2, "%s.%s.%s.%s",
                    websGetVar(wp, "F_EndIP1", "0"),
                    websGetVar(wp, "F_EndIP2", "0"),
                    websGetVar(wp, "F_EndIP3", "0"),
                    websGetVar(wp, "F_EndIP4", "0"));
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
        else {
            strcpy(filter_ipaddr1, "0.0.0.0");
            strcpy(filter_ipaddr2, "255.255.255.255");
            lan_ipaddr0 = filter_ipaddr1;
            lan_ipaddr1 = filter_ipaddr2;
        }
    }

    char *o_lan_ipaddr0, *o_lan_ipaddr1, *o_lan_port0, *o_lan_port1, *o_proto;
    char *o_day_start, *o_sec_start, *o_enable, *o_desc;
    for (i = 0;  i < MAX_ACL; i++) {
        if (i == filter_client_cur_index)
            continue;
        sprintf(nv_name, "filter_client%d", i);

        c = nvram_safe_get(nv_name);

        if (c[0] == '\0')
            return 0;
        /* Check for port specification */
        o_lan_port0 = c;
        o_lan_ipaddr0 = strsep(&o_lan_port0, ":");
        if (!o_lan_port0){
            return 0;
        }

        /* Check for protocol specification */
        o_proto = o_lan_port0;
        o_lan_port0 = strsep(&o_proto, ",");

        /* Check for day specification */
        o_day_start = o_proto;
        o_proto = strsep(&o_day_start, ",");

        /* Check for time specification */
        o_sec_start = o_day_start;
        o_day_start = strsep(&o_sec_start, ":");

        /* Check for enable specification */
        o_enable = o_sec_start;
        o_sec_start = strsep(&o_enable, ",");

        /* Check for description specification (optional) */
        o_desc = o_enable;
        o_enable = strsep(&o_desc, ",");

        /* Check for address range (optional) */
        o_lan_ipaddr1 = o_lan_ipaddr0;
        o_lan_ipaddr0 = strsep(&o_lan_ipaddr1, "-");

        /* Check for port range (optional) */
        o_lan_port1 = o_lan_port0;
        o_lan_port0 = strsep(&o_lan_port1, "-");

        if((!strcmp(o_lan_ipaddr0,lan_ipaddr0)) && (!strcmp(o_lan_ipaddr1,lan_ipaddr1)) &&
            ( !strcmp(o_lan_port0,ptr2) ) && (!strcmp(o_lan_port1,ptr3)) &&
            (!strcmp(o_desc,ptr)) && (!strcmp(o_proto,apply_proto))){
            return 1;
        }
    }
    return 0;
}


static void
validate_fw_schedule_setting(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20]= "0,0,0,0,0,0,0";

    struct variable fields[] = {

        { name: "StartHour", longname: "StartHour", nvram_name: "fw_starthour" },
        { name: "EndHour", longname: "EndHour", nvram_name: "fw_endhour" },
        { name: "StartMinute", longname: "StartMinute", nvram_name: "fw_startmin" },
        { name: "EndMinute", longname: "EndMinute", nvram_name: "fw_endmin" },

    };

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    ptr = websGetVar(wp, "BEveryday", NULL);
    nvram_set("fw_everyday", (ptr) ? "1" : "0");

    if (!ptr) {
        ptr = websGetVar(wp, "BlockSun", NULL);
        tmp[0] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockMon", NULL);
        tmp[2] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockTue", NULL);
        tmp[4] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockWed", NULL);
        tmp[6] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockThu", NULL);
        tmp[8] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockFri", NULL);
        tmp[10] = ((ptr) ? '1' : '0');
        ptr = websGetVar(wp, "BlockSat", NULL);
        tmp[12] = ((ptr) ? '1' : '0');
        nvram_set("fw_blockdays", tmp);
    }

    ptr = websGetVar(wp, "BlockAllday", NULL);
    nvram_set("fw_blockallday", (ptr) ? "1" : "0");

    //urlfilter_tableDriverdel(); //clear old urlfilter rule entry first 2006/01/04  Timo
    fw_schedule_cfg_save();
    /* access table driver, return value not checked ? */
    urlfilter_tableDriverAccess();
    acl_tableDriverAccess();
    remoteMgmt_tableDriverAccess();
    email_tableDriverAccess();//Timo Wu 2005/1/12
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}

//#if defined(DNI)
//#if 0
//  #define MAX_SPECIAL_AP      20 //pamela20060106
//#else
//#define MAX_SPECIAL_AP        8
//#endif
int trigger_port_cur_index = MAX_SPECIAL_AP;
static void
validate_fw_porttrigger_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];
    char path[64];

    struct variable fields[] = {

        { name: "triggertime", longname: "triggertime", nvram_name: "fw_triggertime" },

    };

    ptr = websGetVar(wp, "Button", NULL);
    if (!ptr)
        return;
    if (strstr(ptr, "Delete")) {
        ptr = websGetVar(wp, "delonly", "0");
        if (strcmp(ptr, "1") == 0) {
            ptr = websGetVar(wp, "R1", NULL);
            if (ptr) {
                sprintf(tmp, "trigger_port%d", atoi(ptr));
                nvram_set(tmp, "");
                ptr = websGetVar(wp, "page", "");
                strcpy(path, "/www/");
                if (!ptr)
                        strcat(path, DirectoryIndex);
                        else
                            strcat(path, ptr);
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                //porttrig_cfg_save();
                // the packets of special application will trap to CPU by ACL
                //urlfilter_tableDriverAccess();
                //acl_tableDriverAccess();
                //email_tableDriverAccess();//Timo Wu 2005/1/12
            }
            else {

                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "del_trigger_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

            }
        }
    }
    else if (strstr(ptr, "Apply")) {
        for (i = 0; i < ARRAYSIZE(fields); i++) {
            ptr = websGetVar(wp, fields[i].name, NULL);
            if (ptr) {
                nvram_set(fields[i].nvram_name, ptr);
            }
        }

        ptr = websGetVar(wp, "PortTriEnable", NULL);
        nvram_set("fw_porttrigenable", (ptr) ? "0" : "1");

        ptr = websGetVar(wp, "triggerporten", NULL); //pamela20060106
        nvram_set("triggerporten", ptr);

        porttrig_cfg_save();
        // the packets of special application will trap to CPU by ACL

        /* avoid already opens port still work (HNAPT) */
        urlfilter_tableDriverAccess();
        acl_tableDriverAccess();
        remoteMgmt_tableDriverAccess();
        email_tableDriverAccess();//Timo Wu 2005/1/12
        //action = RESTART;
        //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

    }
    else if (strstr(ptr, "Add")) {

        /* add porttriger */
        char path[64] = "/www/addporttrigger.html";

        trigger_port_cur_index = MAX_SPECIAL_AP;
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        noread = 2;
    }
    else if (strstr(ptr, "Modify")) {
        /* edit port trigger */
        char path[64] = "/www/addporttrigger.html";

        ptr = websGetVar(wp, "R1", NULL);

        if (ptr) {
            trigger_port_cur_index = atoi(ptr);
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            noread = 2;
        }
        else {
            ptr = websGetVar(wp, "page", "");
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                    else
                        sprintf(return_page, "/%s", ptr);
            strcpy(path, "/www/");
            strcat(path, "edit_trigger_err.html");
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }
    }
    else if (strstr(ptr, "Del")) {
        char path[64] = "/www/";

        /* delete this port trigger */
        ptr = websGetVar(wp, "R1", NULL);

        if (ptr) {
            i = atoi(ptr);
            sprintf(tmp, "trigger_port%d", i);
            nvram_set(tmp, "");

            ptr = websGetVar(wp, "page", "");
            strcat(path, ptr);
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            //porttrig_cfg_save();
            // the packets of special application will trap to CPU by ACL
            //urlfilter_tableDriverAccess();
            //acl_tableDriverAccess();
            //email_tableDriverAccess();//Timo Wu 2005/1/12
        }
        else {
            ptr = websGetVar(wp, "page", "");
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                    else
                        sprintf(return_page, "/%s", ptr);
            strcpy(path, "/www/");
            strcat(path, "del_trigger_err.html");
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        }
    }

}

static int
get_free_trigger_port_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_SPECIAL_AP; i++) {
        sprintf(nvram_name, "trigger_port%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;

}

static int
get_free_dhcpc_res_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_DHCPC; i++) {
        sprintf(nvram_name, "dhcpc_res%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;
}


/* d-d,wan_proto>ip:d-d,proto,on/off,used,desc */
static void
validate_fw_porttrigger_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6, *ptr7;
    int i;
    char tmp[100], nvram_name[20];
    char ipaddr[20];
    char on_off[10];

    char path[64]; //Add by Timo Wu

    int ret; //Add by Timo Wu

    sprintf(nvram_name, "trigger_port%d",   trigger_port_cur_index);

    ptr = websGetVar(wp, "TriggerEn", NULL);
    //ptr = websGetVar(wp, "src_ip_type", NULL);
    ptr3 = websGetVar(wp, "pt_type", NULL);
    ptr1 = websGetVar(wp, "pt_name", NULL);
    ptr2 = websGetVar(wp, "pt_port", NULL);     //pamela
    //ptr3 = websGetVar(wp, "OutEndPort", NULL);
    ptr4 = websGetVar(wp, "in_port_start", NULL);//pamela
    ptr5 = websGetVar(wp, "in_port_end", NULL); //pamela
    ptr6 = websGetVar(wp, "in_port_type", NULL);

    ptr7 = websGetVar(wp, "src_ip_type", NULL);

    if (!ptr || !ptr1 || !ptr2 || !ptr3 || !ptr4 || !ptr5 || !ptr6 || !ptr7)
        return;

    //Add by Timo wu
    ret = check_porttrigger_entry( ptr1, ptr4, ptr5, ptr2, trigger_port_cur_index);
    if (ret != TRUE)
    {

        switch (ret)
        {
            case NAME_CONFLICT:

                sprintf(path ,"/www/error_port_servicename_exist.html");
                break;
            case INBOUND_PORT_CONFLICT:
                sprintf(path ,"/www/error_triggerinbound_exist.html");
                break;
            case TRIGGER_PORT_CONFLICT:
                sprintf(path ,"/www/error_triggerport_exist.html");
                break;
        }

        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            }
            noread = 1;
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

        action = NOREPLY; //only reply error page, not reply previous page in apply_cgi

        return;
    }

#if 1
    if (strcmp(ptr7, "2") == 0) {
        sprintf(ipaddr, "%s.%s.%s.%s", websGetVar(wp, "src_ip1", "0"),
                        websGetVar(wp, "src_ip2", "0"),
                        websGetVar(wp, "src_ip3", "0"),
                        websGetVar(wp, "src_ip4", "0"));
    }
    else {
        strcpy(ipaddr, "0.0.0.0");
    }
#else
    strcpy(ipaddr, "0.0.0.0");
#endif

    if (trigger_port_cur_index >= MAX_SPECIAL_AP)
        strcpy(on_off, "on");   /* new add always on */
    else if (pRomeCfgParam->specialapCfgParam[0][trigger_port_cur_index].enable)
        strcpy(on_off, "on");   /* keep setting */
    else
        strcpy(on_off, "off");


    sprintf(tmp, "%s-%s,%s>%s:%s-%s,%s,%s,%s,%s",
            ptr2,
            ptr2,//ptr3,
            ((atoi(ptr3) == 6) ? "tcp" : "udp"),
            //"192", "168", "1", "100",
            ipaddr,
            ptr4,
            ptr5,
            //"tcp",
            ((atoi(ptr6) == 6) ? "tcp" : ((atoi(ptr6) == 17) ? "udp" : "both")),
            //((strcmp(ptr, "on") == 0) ? "on" : "off"),
            //"on",
            on_off,
            "on",
            ((ptr1) ? ptr1 : "")
            );
    if (trigger_port_cur_index >= MAX_SPECIAL_AP) {
        i = get_free_trigger_port_entry();
        if ((i < 0) || (i >= MAX_SPECIAL_AP)) {
            printf("no free trigger port entry\n");
            return;
        }
        sprintf(nvram_name, "trigger_port%d", i);
    }

    nvram_set(nvram_name, tmp);
    //porttrig_cfg_save();
    // the packets of special application will trap to CPU by ACL
    //urlfilter_tableDriverAccess();
    //acl_tableDriverAccess();
    //email_tableDriverAccess();//Timo Wu 2005/1/12
}


static void
validate_email_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

    struct variable fields[] = {

        { name: "LogSubject", longname: "LogSubject", nvram_name: "logsubject" },
        { name: "AlertSub", longname: "AlertSub", nvram_name: "alertsubject" },
        { name: "mailServer", longname: "mailServer", nvram_name: "mailserver" },
        { name: "toMail", longname: "toMail", nvram_name: "tomail" },
        { name: "AlertSelect", longname: "AlertSelect", nvram_name: "alerttype" },
        { name: "AlertTime", longname: "AlertTime", nvram_name: "alertday" },
        { name: "AlertHour", longname: "AlertHour", nvram_name: "alerthour" },
        { name: "AlertGMT", longname: "AlertGMT", nvram_name: "time_zone" },

        { name: "AuthName", longname: "AuthName", nvram_name: "email_authname" },
        { name: "AuthPassWd", longname: "AuthPassWd", nvram_name: "email_authpasswd" },


    };

    int old_zone = atoi(nvram_get("time_zone"));
    int old_dayloght = atoi(nvram_get("daylight"));

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    ptr = websGetVar(wp, "emailEnable", NULL);
    nvram_set("emailenable", (ptr) ? "1" : "0");

    ptr = websGetVar(wp, "ServerAuth", NULL);
    nvram_set("email_serverauth", (ptr) ? "1" : "0");

    ptr = websGetVar(wp, "AlertImm", NULL);
    nvram_set("alertimm", (ptr) ? "1" : "0");

    ptr = websGetVar(wp, "AlertAMPM", NULL);
    //nvram_set("alertampm", (ptr) ? "1" : "0");
    if (ptr) {
        nvram_set("alertampm", ptr);
    }

    ptr = websGetVar(wp, "Daylight", NULL);
    nvram_set("daylight", (ptr) ? "1" : "0");

    email_cfg_save();

    urlfilter_tableDriverAccess();//Timo Wu 2006/1/12
    acl_tableDriverAccess();//Timo Wu 2006/1/12
    remoteMgmt_tableDriverAccess();
    email_tableDriverAccess();//Timo Wu 2006/1/12
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
    if ((old_zone != pRomeCfgParam->emailCfgParam.time_zone) ||
        (old_dayloght != pRomeCfgParam->emailCfgParam.daylight)) {
            time_start();
    }

}

static void
validate_wan_other_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];
    ipaddr_t    naptIp=0;

        struct variable fields[] = {
        { name: "autoconnect", longname: "autoconnect", nvram_name: "auto_connect" },   //pamela20051228
        { name: "dmzenable", longname: "dmzenable", nvram_name: "dmz_enable" }, //pamela20051228
        { name: "wanping", longname: "wanping", nvram_name: "wan_ping" },   //pamela20051228
        { name: "fwdisable", longname: "fwdisable", nvram_name: "fw_disable" }, //pamela20051228
        { name: "DmzAddr", longname: "DmzAddr", nvram_name: "dmz_ipaddr" },
        { name: "MTU", longname: "MTU", nvram_name: "fw_wan_mtu" },

    };

    int wan_mtu = atoi(nvram_safe_get("fw_wan_mtu"));
    int auto_connect = atoi(nvram_safe_get("auto_connect"));
    int fwdisable = atoi(nvram_safe_get("fw_disable"));


    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }


    //pamela20051228
    //ptr = websGetVar(wp, "WanPing", NULL);
    //nvram_set("wan_ping", (ptr) ? "1" : "0");

    //ptr = websGetVar(wp, "SPILog", NULL);
    //nvram_set("fw_disable", (ptr) ? "1" : "0");

    dmz_cfg_save();         /* dmz */
    memcpy(&naptIp,pRomeCfgParam->ifCfgParam[0].ipAddr,4);
    rtl8651a_delDmzHost(0,(ipaddr_t)naptIp);
    if (pRomeCfgParam->dmzCfgParam[0].enable) {
        //ipaddr_t  naptIp=0;

        //memcpy(&naptIp,pRomeCfgParam->ifCfgParam[0].ipAddr,4);
        //rtl8651a_delDmzHost(0,(ipaddr_t)naptIp);
        rtl8651a_addDmzHost(0,(ipaddr_t)naptIp, (ipaddr_t)pRomeCfgParam->dmzCfgParam[0].dmzIp);
    }
    mgmt_cfg_save();        /* wan_ping */
    remoteMgmt_tableDriverAccess();
    firewall_starting();

    if (fwdisable != atoi(nvram_safe_get("fw_disable"))) {
        rtl8651a_setDosStatus(0, ((pRomeCfgParam->dosCfgParam[0].enable)?(pRomeCfgParam->dosCfgParam[0].enableItem):0));
        dos_cfg_save();
    }

    if (wan_mtu != atoi(nvram_safe_get("fw_wan_mtu"))) {
        wan_security_cfg_save();    /* mtu, ... */
        action = RESTART;
        //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
    }

    if (auto_connect != atoi(nvram_safe_get("auto_connect"))) {
        //wan_cfg_save(wan_pastType);
        wan_security_cfg_save();
        action = RESTART;
        //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
    }

}

static void
validate_password_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];
    char path[64] = "/www/error_password.html";


    ptr = websGetVar(wp, "OldPwd", NULL);
    if ((ptr == NULL) || (strcmp(ptr, (char *)nvram_safe_get("http_passwd")) != 0)) {

        //websBufferWrite(wp, "Invalid <b>%s</b> %s: old password <br>",
        //    ptr, NULL);
        //return;
        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            }
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

        action = NOREPLY; //only reply error page, not reply previous page in apply_cgi

        return;
    }

    ptr = websGetVar(wp, "NewPwd", NULL);
    ptr1 = websGetVar(wp, "CnfPwd", NULL);
    if ((ptr == NULL) || (ptr1 == NULL) || (strcmp(ptr, ptr1) != 0)) {

        //websBufferWrite(wp, "Invalid <b>%s</b> %s: new password <br>",
        //    ptr, ptr1);
        //
        //return;
        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            }
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

        action = NOREPLY; //only reply error page, not reply previous page in apply_cgi

        return;
    }

    nvram_set("http_passwd", ptr);
    mgmt_cfg_save();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);


}

static void
validate_management_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

        struct variable fields[] = {

        { name: "RemoteIpPort", longname: "RemoteIpPort", nvram_name: "http_wanport" },

    };

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    ptr = websGetVar(wp, "remoteen", NULL);
    nvram_set("fw_remoteenable", (ptr) ? "1" : "0");

    ptr = websGetVar(wp, "iptypehidden", NULL);
    if (ptr) {
        int i = atoi(ptr);
        if (i == 0) {
            nvram_set("remote_config_ip_start", "0.0.0.0");
            nvram_set("remote_config_ip_end", "255.255.255.255");
        }
        else if (i == 1) {
            ptr = websGetVar(wp, "StartIP1", "0");
            ptr1 = websGetVar(wp, "StartIP2", "0");
            ptr2 = websGetVar(wp, "StartIP3", "0");
            ptr3 = websGetVar(wp, "StartIP4", "0");

            if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
                sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
                if (strcmp(tmp, "...") == 0)
                    strcpy(tmp, "0.0.0.0");
                nvram_set("remote_config_ip_start", tmp);
            }
            ptr = websGetVar(wp, "EndIP1", "0");
            ptr1 = websGetVar(wp, "EndIP2", "0");
            ptr2 = websGetVar(wp, "EndIP3", "0");
            ptr3 = websGetVar(wp, "EndIP4", "0");

            if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
                sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
                if (strcmp(tmp, "...") == 0)
                    strcpy(tmp, "0.0.0.0");
                nvram_set("remote_config_ip_end", tmp);
            }
        }
        else if (i == 2) {

            ptr = websGetVar(wp, "PCIP1", "0");
            ptr1 = websGetVar(wp, "PCIP2", "0");
            ptr2 = websGetVar(wp, "PCIP3", "0");
            ptr3 = websGetVar(wp, "PCIP4", "0");

            if ((ptr) && (ptr1) && (ptr2) && (ptr3)) {
                sprintf(tmp, "%s.%s.%s.%s", ptr, ptr1, ptr2, ptr3);
                if (strcmp(tmp, "0.0.0.0") == 0)
                    strcpy(tmp, "0.0.0.0");
                nvram_set("remote_config_ip_start", tmp);
                nvram_set("remote_config_ip_end", tmp);
            }
        }
    }

    mgmt_cfg_save();
    remoteMgmt_tableDriverAccess();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
    firewall_starting();


}


typedef struct {
    char *name;
    char protocol;
    char ports;
    char *ports_ptr;
} vs_variable;

#define _TCP    6
#define _UDP    17

vs_variable app_ports[] = {

{ "AIM", _TCP, 1, "5190-5190" }, /* AIM */
{ "Age-Of-Empire", _UDP, 1, "47624-47624" }, /* Age-of-Empire   */
{ "FTP", _TCP, 1, "20-21" }, /* FTP */
{ "HTTP", _TCP, 1, "80-80" }, /* HTTP */
{ "ICUII", _TCP, 1, "23566-23566" }, /* ICUII */
{ "IP_Phone", _TCP, 1, "6670-6670" }, /* IP_Phone */
//{ "NetMeeting", _TCP, 1, "1720-1720,1503-1503" }, /* NetMeeting */
{ "NetMeeting", _TCP, 1, "1720-1720" }, /* NetMeeting */
{ "News", _TCP, 1, "119-119" }, /* News */
{ "PPTP", (_TCP+_UDP), 1, "1723-1723" }, /* PPTP */
{ "QuakeII/III", (_TCP +_UDP), 1, "27960-27960" }, /* QuakeII/III */
{ "Real-Audio", (_TCP+_UDP), 1, "6970-7170" }, /* Real-Audio */
{ "Telnet", _TCP, 1, "23-23" }, /* TELNET */

{NULL, 0, NULL}

};

/*********Add by Timo Wu*****/
static int
check_forwardport_entry(char *service_name, char *start_port, char *end_port ,int current_index, char *status)
{
    int i, search_start;
    char nvram_name[20];
    char *ptr;
    char tmp[200];
    //char name[20];
    char *wan_port0, *wan_port1, *lan_ipaddr, *lan_port0, *lan_port1, *proto, *enable, *desc;
    int start ,end ,lan_start, lan_end;

    //sprintf( name, "on,%s", service_name);

    search_start = 0;
    //for(i = search_start; i < current_index ;i++)
    for(i = search_start; i < MAX_SERVER_PORT ;i++)
    {
        //printf("search_start is 0\n");
        if (i == current_index)
            continue;

        sprintf(nvram_name, "forward_port%d", i);
        ptr = nvram_safe_get(nvram_name);

        strcpy(tmp, ptr);
        /* Check for wan income port specification */
        lan_ipaddr = tmp;
        wan_port0 = strsep(&lan_ipaddr, ">");
        if (!lan_ipaddr)
            return TRUE;

        /* Check for LAN destination port specification */
        lan_port0 = lan_ipaddr;
        lan_ipaddr = strsep(&lan_port0, ":");
        if (!lan_port0)
            return TRUE;

        /* Check for protocol specification */
        proto = lan_port0;
        lan_port0 = strsep(&proto, ":,");
        if (!proto)
            return TRUE;

        /* Check for enable specification */
        enable = proto;
        proto = strsep(&enable, ":,");
        if (!enable)
            return TRUE;

        /* Check for description specification (optional) */
        desc = enable;
        enable = strsep(&desc, ":,");

        /* Check for WAN destination port range (optional) */
        wan_port1 = wan_port0;
        wan_port0 = strsep(&wan_port1, "-");
        if (!wan_port1)
            wan_port1 = wan_port0;

        /* Check for LAN destination port range (optional) */
        lan_port1 = lan_port0;
        lan_port0 = strsep(&lan_port1, "-");
        if (!lan_port1)
            lan_port1 = lan_port0;
        //printf("lan_port0 %s lan_port1 %s start_port %s end_port %s status %s\n", lan_port0, lan_port1, start_port, end_port , status);
        lan_start = atoi(lan_port0);
        lan_end = atoi(lan_port1);

        //printf("name %s desc %s\n", name, desc);
        if( ( (strstr(ptr, start_port)) && (strcmp(service_name, desc)==0)) ) //for Port forwarding Add Match 80-80 (This means service name conflict
        {
            //printf("name conflict\n");
            return NAME_CONFLICT;
        }


        if( strcmp( status, "edit")==0)
        {
            start = atoi(start_port);
            end = atoi(end_port);

          if( ((start <= lan_start)&&( lan_start <= end)) || ((start <= lan_end)&&( lan_end <= end)) || ((lan_start <= start)&&( start <= lan_end)) || ((lan_start <= end)&&( end <= lan_end)))
          { //This means service port conflict,
            //printf("forward port conflict\n");
            return FORWARD_PORT_CONFLICT;
          }
        }
    }

#if 0
    search_start= current_index+1;
    for( i = search_start; i < MAX_SERVER_PORT; i++)
    {
        //printf("search_start = %d\n", search_start);
        sprintf(nvram_name, "forward_port%d", i);
        ptr = nvram_safe_get(nvram_name);

        strcpy(tmp, ptr);
        /* Check for wan income port specification */
        lan_ipaddr = tmp;
        wan_port0 = strsep(&lan_ipaddr, ">");
        if (!lan_ipaddr)
            return TRUE;

        /* Check for LAN destination port specification */
        lan_port0 = lan_ipaddr;
        lan_ipaddr = strsep(&lan_port0, ":");
        if (!lan_port0)
            return TRUE;

        /* Check for protocol specification */
        proto = lan_port0;
        lan_port0 = strsep(&proto, ":,");
        if (!proto)
            return TRUE;

        /* Check for enable specification */
        enable = proto;
        proto = strsep(&enable, ":,");
        if (!enable)
            return TRUE;

        /* Check for description specification (optional) */
        desc = enable;
        enable = strsep(&desc, ":,");

        /* Check for WAN destination port range (optional) */
        wan_port1 = wan_port0;
        wan_port0 = strsep(&wan_port1, "-");
        if (!wan_port1)
            wan_port1 = wan_port0;

        /* Check for LAN destination port range (optional) */
        lan_port1 = lan_port0;
        lan_port0 = strsep(&lan_port1, "-");
        if (!lan_port1)
            lan_port1 = lan_port0;
        //printf("lan_port0 %s lan_port1 %s start_port %s end_port %s status %s\n", lan_port0, lan_port1, start_port, end_port , status);
        lan_start = atoi(lan_port0);
        lan_end = atoi(lan_port1);


        if( ( (strstr(ptr, start_port)) && (strcmp(service_name, desc)==0)) ) //for Port forwarding Add Match 80-80 (This means service name conflict
        {
            return NAME_CONFLICT;
        }


        if( strcmp( status, "edit")==0)
        {
            start = atoi(start_port);
            end = atoi(end_port);
          if( ((start <= lan_start)&&( lan_start <= end)) || ((start <= lan_end)&&( lan_end <= end)) || ((lan_start <= start)&&( start <= lan_end)) || ((lan_start <= end)&&( end <= lan_end)))
          { //This means service port conflict,
            return FORWARD_PORT_CONFLICT;
          }
        }
    }
#endif

    return TRUE; //not finding any one conflict
}

static int
check_qos_entry(char *service_name, char *start_port, char *end_port ,int current_index, char *status)
{
    int i, search_start;
    char nvram_name[20];
    char *ptr;
    char tmp[200];
    //char name[20];
    char *port0, *port1, *proto, *enable, *priority, *desc;
    int start ,end ,lan_start, lan_end;

    search_start = 0;
    for(i = search_start; i < MAX_QOS ;i++)
    {
        if (i == current_index)
            continue;

        sprintf(nvram_name, "qos_port%d", i);
        ptr = nvram_safe_get(nvram_name);

        strcpy(tmp, ptr);

        /* Check for port specification */
        proto = tmp;
        port0 = strsep(&proto, ":,");
        if (!proto)
            return TRUE;

        /* Check for proto specification */
        enable = proto;
        proto = strsep(&enable, ":,");
        if (!enable)
            return TRUE;

        /* Check for priority specification (optional) */
        priority = enable;
        enable = strsep(&priority, ":,");
        if (!priority)
            return TRUE;

        /* Check for description specification (optional) */
        desc = priority;
        priority = strsep(&desc, ":,");

        /* Check for port range (optional) */
        port1 = port0;
        port0 = strsep(&port1, "-");
        if (!port1)
            port1 = port0;

        lan_start = atoi(port0);
        lan_end = atoi(port1);

        //printf("name %s desc %s\n", name, desc);
        if( ( (strstr(ptr, start_port)) && (strcmp(service_name, desc)==0)) ) //for Port forwarding Add Match 80-80 (This means service name conflict
        {
            //printf("name conflict\n");
            return NAME_CONFLICT;
        }


        if( strcmp( status, "edit")==0)
        {
            start = atoi(start_port);
            end = atoi(end_port);

          if( ((start <= lan_start)&&( lan_start <= end)) || ((start <= lan_end)&&( lan_end <= end)) || ((lan_start <= start)&&( start <= lan_end)) || ((lan_start <= end)&&( end <= lan_end)))
          { //This means service port conflict,
            //printf("forward port conflict\n");
            return FORWARD_PORT_CONFLICT;
          }
        }
    }


    return TRUE; //not finding any one conflict
}

/*********End******/

//#define MAX_SERVER_PORT       8
static int
get_free_forwardport_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_SERVER_PORT; i++) {
        sprintf(nvram_name, "forward_port%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;

}

static int
get_free_filter_client_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_ACL; i++) {
        sprintf(nvram_name, "filter_client%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;

}

static int
get_free_qos_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_QOS; i++) {
        sprintf(nvram_name, "qos_port%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;

}

//static void
static int //Modified by Timo Wu
validate_service_port_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i,ret;
    char tmp[200], nvram_name[20];
    vs_variable *vs_ptr=NULL;


    ptr = websGetVar(wp, "vsService", NULL);
    ptr1 = websGetVar(wp, "vsIPAddr", NULL);
    if ((ptr == NULL) || (ptr1 == NULL))
        return TRUE;
    i = atoi(ptr);
    if (i > ARRAYSIZE(app_ports))
        return TRUE;

    vs_ptr = &(app_ports[i-1]);

    ptr = strtok(vs_ptr->ports_ptr, ",:");

    //Add by Timo wu
    ret = check_forwardport_entry( vs_ptr->name, ptr , "BLANK" , MAX_SERVER_PORT , "BLANK");
    switch(ret)
    {
        case FORWARD_PORT_CONFLICT:
        case NAME_CONFLICT:
            return NAME_CONFLICT;
        default:
            break;
    }

    i = 0;
    while (ptr) {
        i = get_free_forwardport_entry();
        if ((i < 0) || (i >= MAX_SERVER_PORT)) {
            printf("no free server entry\n");
            return TRUE;
        }
        sprintf(nvram_name, "forward_port%d", i);
        sprintf(tmp, "%s>%s:%s,%s,on,%s",
                ptr, ptr1, ptr,
                ((vs_ptr->protocol == _TCP+_UDP) ? "both" :
                    ((vs_ptr->protocol == _TCP) ? "tcp" : "udp")),
                vs_ptr->name);
        nvram_set(nvram_name, tmp);
        ptr = strtok(NULL, ",:");
        i++;
    }
    forwardport_cfg_save();
    serverp_tableDriverAccess(0);

    return TRUE; //Add by Timo
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);
}


extern int forward_port_cur_index;

/* d-d>ip:d-d,proto,on/off,desc */

static void
validate_service_port_editing(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6, *ptr7;
    int i, ret;
    char tmp[100], nvram_name[20];

    sprintf(nvram_name, "forward_port%d",   forward_port_cur_index);

    ptr = websGetVar(wp, "GlobalPort", NULL);
    ptr1 = websGetVar(wp, "LocalPort", NULL);
    ptr2 = websGetVar(wp, "IP1", NULL);
    ptr3 = websGetVar(wp, "IP2", NULL);
    ptr4 = websGetVar(wp, "IP3", NULL);
    ptr5 = websGetVar(wp, "IP4", NULL);
    ptr6 = websGetVar(wp, "VirSerAct", NULL);
    ptr7 = websGetVar(wp, "ServiceType", NULL);

    //printf("%s\n", __FUNCTION__);
    if (!ptr || !ptr1 || !ptr2 || !ptr3 || !ptr4 || !ptr5 || !ptr7)
        return;

    //Add by Timo wu

    ret = check_forwardport_entry( "BLANK", ptr, ptr1, forward_port_cur_index  ,"edit");
    switch(ret)

    {
        case NAME_CONFLICT:
        case FORWARD_PORT_CONFLICT:
    {
        char path[64] = "/www/error_port_serviceport_exist.html";

        ptr = websGetVar(wp, "page", "");
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            }
            noread = 1;
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        action = NOREPLY; //only reply error page, not reply previous page in apply_cgi
        return;
            }
        default:
            break;
    }


    sprintf(tmp, "%s-%s>%s.%s.%s.%s:%s-%s,%s,%s,%s",
            ptr,
            ptr1,
            ptr2,ptr3,ptr4,ptr5,
            ptr,
            ptr1,
            ((atoi(ptr7) == 6) ? "tcp" : ((atoi(ptr7) == 17) ? "udp" : "both")),
            //( (ptr6) ? "on" : "off"),
            "on",               /* always on */
            websGetVar(wp, "Remark", "")
            );
    if (forward_port_cur_index >= MAX_SERVER_PORT) {
        i = get_free_forwardport_entry();
        if ((i < 0) || (i >= MAX_SERVER_PORT)) {
            printf("no free server entry\n");
            return;
        }
        sprintf(nvram_name, "forward_port%d", i);
    }

    nvram_set(nvram_name, tmp);
    forwardport_cfg_save();
    serverp_tableDriverAccess(0);
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}




int forward_port_cur_index = MAX_SERVER_PORT;

static void
validate_service_port_setting(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[200], nvram_name[20];
    vs_variable *vs_ptr=NULL;


    ptr = websGetVar(wp, "Button", NULL);
    if (ptr == NULL)
        return;
    if (ptr) {
        //if (strstr(ptr, "Add_Vs")) {
        if (!strcmp(ptr, "Add_Vs")) {

            //Add by Timo Wu
            if( validate_service_port_adding(wp, value, v)== NAME_CONFLICT)
            {
                //printf("validate_service_port_adding is FALSE\n");
                char path[64] = "/www/error_port_servicename_exist.html";
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                    else{
                        sprintf(return_page, "/%s", ptr);
                    }
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                return;
            }
            else
            {
                //printf("validate_service_port_adding is TRUE\n");
            ptr = websGetVar(wp, "page", NULL);
            if (!ptr) {
                sprintf(tmp, "/www/%s", DirectoryIndex);
            }
            else
                sprintf(tmp, "/www/%s", ptr);

            websDefaultHandler(wp, urlPrefix, webDir, arg, url, tmp , query);
            }
        }
        //else if (strstr(ptr, "Add")) {
        else if (!strcmp(ptr, "Add")) {
            /* add service block */
            char path[64] = "/www/ports_man.html";

            forward_port_cur_index = MAX_SERVER_PORT;

            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html

        }
        //else if (strstr(ptr, "Modify")) {
        else if (!strcmp(ptr, "Modify")) {
            /* edit service */
            char path[64] = "/www/ports_man.html";

            ptr = websGetVar(wp, "R1", NULL);
            if (ptr) {
                forward_port_cur_index = atoi(ptr);

                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html
            }
            else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "edit_forward_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            }

        }
        //else if (strstr(ptr, "Delete")) {
        else if (!strcmp(ptr, "Delete")) {
            char path[64] = "/www/ports_man.html";

            ptr = websGetVar(wp, "onlydel", "0");
            if (strcmp(ptr, "1") == 0) {
                /* delete this service */
                ptr = websGetVar(wp, "R1", NULL);
                if (ptr) {
                    i = atoi(ptr);

                    sprintf(nvram_name, "forward_port%d", i);
                    nvram_set(nvram_name, "");
                    ptr = websGetVar(wp, "page", "");
                    strcpy(path, "/www/");
                    if (!ptr)
                            strcat(path, DirectoryIndex);
                            else
                                strcat(path, ptr);
                            //strcpy(return_page, path);
                    if (!ptr)
                            sprintf(return_page, "/%s", DirectoryIndex);
                            else
                                sprintf(return_page, "/%s", ptr);

                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                    //websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/test_conn1.html" , NULL);

                    forwardport_cfg_save();
                    serverp_tableDriverAccess(0);
                        /* avoid already opens port still work (HNAPT) */
                    //action = RESTART;
                    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

                }
                else {

                    ptr = websGetVar(wp, "page", "");
                    if (!ptr)
                            sprintf(return_page, "/%s", DirectoryIndex);
                            else
                                sprintf(return_page, "/%s", ptr);
                    strcpy(path, "/www/");
                    strcat(path, "del_forward_err.html");
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

                }

            }
        }
    }


}

//#define MAX_ROUTE 8

int static_route_cur_index = MAX_ROUTING;

static int
get_free_static_route_entry(void)
{
    int i;
    char nvram_name[20];
    char *ptr;

    for (i = 0;  i < MAX_ROUTING; i++) {
        sprintf(nvram_name, "static_route%d", i);
        ptr = nvram_safe_get(nvram_name);
        if (strcmp(ptr, "") == 0)
            return i;
    }

    return -1;

}


static void
validate_static_route_setting(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3,flag=0;
    int i;
    char tmp[200], nvram_name[20];
    vs_variable *vs_ptr=NULL;


    ptr = websGetVar(wp, "Button", NULL);
    if (ptr == NULL)
        return;
    if (ptr) {
        if (strstr(ptr, "Add")) {
            /* add static route */
            char path[64] = "/www/routeentry.html";

            static_route_cur_index = MAX_ROUTING;

            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            noread = 2;

        }
        else if (strstr(ptr, "Modify")) {
            /* edit static route */
            char path[64] = "/www/routeentry.html";

            ptr = websGetVar(wp, "R1", NULL);
            if (ptr) {
                static_route_cur_index = atoi(ptr);

                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                noread = 2;

            }
            else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "edit_route_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            }

        }
        else if (strstr(ptr, "Delete")) {
            char path[64] = "/www/";
            /* delete this static route */

            printf("ptr = %s\n", ptr);
            ptr = websGetVar(wp, "R1", NULL);
            if (ptr) {
                routing_tableDriverdel();
                i = atoi(ptr);

                sprintf(nvram_name, "static_route%d", i);
                nvram_set(nvram_name, "");
                ptr = websGetVar(wp, "page", "");
                strcat(path, ptr);
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                routing_tableDriveradd();
                flag=1;
            }
            else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "del_route_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            }

        }
    }
    if (flag)
    {
        routing_cfg_save();

#if 0
        //stop and re-start ripd
        char dir[2]={0,0},ver[2]={0,0};
        strcpy(dir,nvram_get("rip_dir"));
        strcpy(ver,nvram_get("rip_ver"));

        if (*dir==0||*ver==0)
            return; // get rip direction or rip version error

        if (atoi(dir)>0&&atoi(ver)>0)
        {
            do_exec("/bin/killall","ripd");
            unlink("/var/run/ripd.pid");
            //wait killall to kill ripd process
            sleep(1);
            do_exec("/bin/ripd", NULL);
        }
#endif
        ripd_start();
    }
}


/* ip,netmask,gateway,on/off,desc */

static void
validate_static_route_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6;
    int i;
    char tmp[100], nvram_name[20];
    int lan_routing = 0;

    sprintf(nvram_name, "static_route%d",   static_route_cur_index);

    ptr = websGetVar(wp, "RoutName", NULL);
    ptr1 = websGetVar(wp, "RtDest", NULL);
    ptr2 = websGetVar(wp, "RoutMask", NULL);
    ptr3 = websGetVar(wp, "RoutGate", NULL);
    ptr4 = websGetVar(wp, "RoutAct", NULL);
    ptr5 = websGetVar(wp, "RtPriv", NULL);
    ptr6 = websGetVar(wp, "RtMetric", NULL);


    if (!ptr || !ptr1 || !ptr2 || !ptr3)
        return;

    /* check RoutGate is same lan netmask */
    {
        char *lan_ipaddr, *lan_netmask;
        struct in_addr ipaddr, netmask, gwaddr;

        lan_ipaddr = nvram_get("lan_ipaddr");
        (void) inet_aton(lan_ipaddr, &ipaddr);
        lan_netmask = nvram_get("lan_netmask");
        (void) inet_aton(lan_netmask, &netmask);
        (void) inet_aton(ptr3, &gwaddr);


        ipaddr.s_addr = ipaddr.s_addr & netmask.s_addr;
        gwaddr.s_addr = gwaddr.s_addr & netmask.s_addr;

        if (ipaddr.s_addr == gwaddr.s_addr) {
            lan_routing = 1;
        }
        else
            lan_routing = 0;
    }

    sprintf(tmp, "%s,%s,%s,%s,%s,%s,%s,%s",
            ptr1,
            ptr2,
            ptr3,
            ptr6,
            ((lan_routing == 1) ? "1" : "0"),       /* 1->LAN, 0->WAN */
            ((ptr4) ? "on" : "off"),
            ((ptr5) ? "1" : "0"),
            ptr
            );
    if (static_route_cur_index >= MAX_ROUTING) {
        i = get_free_static_route_entry();
        if ((i < 0) || (i >= MAX_ROUTING)) {
            printf("no free static route entry\n");
            return;
        }
        sprintf(nvram_name, "static_route%d", i);
    }

    routing_tableDriverdel();

    nvram_set(nvram_name, tmp);
    routing_cfg_save();

    routing_tableDriveradd();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

#if 0
    //stop and re-start ripd
    char dir[2]={0,0},ver[2]={0,0};
    strcpy(dir,nvram_get("rip_dir"));
    strcpy(ver,nvram_get("rip_ver"));

    if (*dir==0||*ver==0)
        return; // get rip direction or rip version error

    if (atoi(dir)>0&&atoi(ver)>0)
    {
        do_exec("/bin/killall","ripd");
        unlink("/var/run/ripd.pid");
        //wait killall to kill ripd process
        sleep(1);
        do_exec("/bin/ripd", NULL);
    }
#endif
    ripd_start();

}

/* ip,macaddr,desc */

#if 0
static void
validate_dhcpc_res_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6;
    int i;
    char tmp[100], nvram_name[20];
    char *c;

    ptr = websGetVar(wp, "subip", NULL);
    if (!ptr) {
        ptr = websGetVar(wp, "editiperr", NULL);
    }
    ptr1 = websGetVar(wp, "FixMacAdr", NULL);
    ptr2 = websGetVar(wp, "FixedRemark", NULL);


    if (!ptr || !ptr1 || !ptr2)
        return;

    c=nvram_name;

    for (i = 0; i < 6; i++) {
        *c++ = ptr1[i*2];
        *c++ = ptr1[i*2+1];
        if (i < 5)
           *c++ = ':';
    }
    *c = 0;
    sprintf(tmp, "%s,%s,%s",
            ptr,
            nvram_name,
            ptr2
            );
    i = dhcpc_res_cur_index;
    if (dhcpc_res_cur_index >= MAX_DHCPC) {
        i = get_free_dhcpc_res_entry();
        if ((i < 0) || (i >= MAX_DHCPC)) {
            printf("no free dhcpc res entry\n");
            return;
        }
    }
    sprintf(nvram_name, "dhcpc_res%d", i);
    nvram_set(nvram_name, tmp);

    dhcps_cfg_save();
    dhcps_start();

}
#endif

//ATTACHED_DEVICES
static void
validate_dhcpc_res_adding(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6, *newIP;
    int i, j = 0, index = 1;
    char tmp[100], nvram_name[20];
    char *c, ipa[20]="", host[34]="";
    unsigned char mac[20];
    int ret = 0;
    int len = 0, dhcpcAvailable = -1;
    unsigned char str[20], *startDhcp, myStartDhcp[4], myIP[10];
    unsigned char path[60], *endDhcp, myEndDhcp[4], value = 0;

    ptr = websGetVar(wp, "subip", NULL);
    if (!ptr) {
        ptr = websGetVar(wp, "editiperr", NULL);
    }
#if 0 //Do Not open following funtions, the funtions were already written in apply_cgi()
    ret = validate_dhcpc_res_ip_adding(ptr);
    if (ret == FALSE) {
        char path[64] = "/www/edit_exceed_err.html";
        ptr = websGetVar(wp, "page", "");
        //printf("ptr = %s\n", ptr);
        if (!ptr)
                sprintf(return_page, "/%s", DirectoryIndex);
            else{
                sprintf(return_page, "/%s", ptr);
            //strcpy(path, "/www/");
            //strcat(path, "edit_exceed_err.html");
            }
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
        return;
    }
#endif

    memset(ipa, '\0', 20);
    strcpy(ipa, ptr);
    ptr1 = websGetVar(wp, "FixMacAdr", NULL);
    ptr2 = websGetVar(wp, "FixedRemark", NULL);
    memset(mac, '\0', sizeof(mac));
    //printf("%s %d ptr1 = %s\n", __FUNCTION__, __LINE__, ptr1);

#if 1 //make a judgement for mac with ':' or without ':'
    if(memchr(ptr1, ':', 20) != NULL){
        //printf("%d ptr1 = %s\n", __LINE__, ptr1);
        memcpy(mac, ptr1, strlen(ptr1));
        //printf("%s %d mac = %s\n", __FUNCTION__, __LINE__, mac);
    } else {
        for (i = 0; i < 12; i++) {
            //printf("ptr1[%d] = %c\n", i, ptr1[i]);
            //printf("ptr1[%d] = %d\n", i, ptr1[i]);
            mac[j++] = ptr1[i];
            if( (i == 1) ||( i == 3) || ( i == 5) ||( i == 7) || ( i == 9) )
                mac[j++]= ':';
            //printf("%s %d mac = %s\n", __FUNCTION__, __LINE__, mac);
        }
    }
#endif

    mac[17] = NULL;

    if (!ptr || !ptr1 || !ptr2)
        return;

    //printf("dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    //if ( dhcpcRsvdAddrCnt == 0) {
        //dhcpcRsvdAddrCk = 0;
        //_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].addrBound = 0;
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host, ptr2, strlen(ptr2));
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac, mac, sizeof(mac));
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip, ptr, strlen(ptr));
        //dhcpcRsvdAddrCnt++;
    //}
    //else
    //{
        for (i = 0; i < dhcpcRsvdAddrCnt; i++) {
            if (strlen(ipa) > strlen(_dhcpc_rsvd_addr[i].rsvd_ip)) {
                len = strlen(ipa);
                //printf("strlen(ipa) = %d\n", len);
            }
            else {
                len = strlen(_dhcpc_rsvd_addr[i].rsvd_ip);
                //printf("strlen(_dhcpc_rsvd_addr[i].rsvd_ip = %d\n", strlen(_dhcpc_rsvd_addr[i].rsvd_ip));
            }
            //printf("strlen(_dhcpc_rsvd_addr[i].rsvd_mac) = %d\n", strlen(_dhcpc_rsvd_addr[i].rsvd_mac));
            //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_host);
            //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_mac);
            //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_ip);
            if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ipa, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) == NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) == NULL) {
                    printf("Found ip and mac matched with attached devices %d\n", i+1);
                    _dhcpc_rsvd_addr[i].addrBound=1;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ipa, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) == NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) != NULL) {
                    printf("Found ip matched attached devices %d\n", i+1);
                    _dhcpc_rsvd_addr[i].addrBound=1;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ipa, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) != NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) == NULL) {
                    printf("Found mac matched attached devices %d\n", i+1);
                    _dhcpc_rsvd_addr[i].addrBound=1;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ipa,  len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) != NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) != NULL) {
                    printf("Found NO ip and mac matched in attached devices !\n");
                    _dhcpc_rsvd_addr[i].addrBound=0;
            }
            else {
                printf("%s : unknow status\n", __FUNCTION__);
            }
        }
    //}

    //printf("dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    j = 0;
    for ( i = 0; i < dhcpcRsvdAddrCnt; i++) {
            //printf("i = %d; j = %d\n", i, j);
            //printf("_dhcpc_rsvd_addr[%d].addrBound = %d\n", i, _dhcpc_rsvd_addr[i].addrBound);
            if ( _dhcpc_rsvd_addr[i].addrBound == 0) {
                _dhcpc_rsvd_addr[j].addrBound = 0;
                memcpy(_dhcpc_rsvd_addr[j].rsvd_host, _dhcpc_rsvd_addr[i].rsvd_host, sizeof(_dhcpc_rsvd_addr[i].rsvd_host));
                memcpy(_dhcpc_rsvd_addr[j].rsvd_mac, _dhcpc_rsvd_addr[i].rsvd_mac, sizeof(_dhcpc_rsvd_addr[i].rsvd_mac));
                memcpy(_dhcpc_rsvd_addr[j].rsvd_ip, _dhcpc_rsvd_addr[i].rsvd_ip, sizeof(_dhcpc_rsvd_addr[i].rsvd_ip));
                //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_host);
                //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_mac);
                //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_ip);
                j++;
            } else {
                _dhcpc_rsvd_addr[i].addrBound = 0;
                memset(_dhcpc_rsvd_addr[i].rsvd_host, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_host));
                memset(_dhcpc_rsvd_addr[i].rsvd_mac, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_mac));
                memset(_dhcpc_rsvd_addr[i].rsvd_ip, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_ip));
            }
    }

    dhcpcRsvdAddrCnt = j;

    printf("%s = dhcpcRsvdAddrCnt = %d\n", __FUNCTION__, dhcpcRsvdAddrCnt);
    if (dhcpcRsvdAddrCnt >= MAX_DHCPC) {
        printf("no free dhcpc res entry\n");
        return;
    }

    _dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].addrBound = 0;
    memset(host, '\0', 34);
    strncpy(host, ptr2, strlen(ptr2));
    memset(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host, '\0', sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host));
    memset(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac, '\0', sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac));
    memset(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip, '\0', sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip));
    memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host, host/*ptr2*/, strlen(ptr2));
    memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac, mac, sizeof(mac));
    memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip, ipa, strlen(ptr));
    dhcpcRsvdAddrCnt++;


    for ( i = 0; i < dhcpcRsvdAddrCnt; i++) {
        sprintf(tmp, "%s,%s,%s",
        _dhcpc_rsvd_addr[i].rsvd_ip,
        _dhcpc_rsvd_addr[i].rsvd_mac,
        _dhcpc_rsvd_addr[i].rsvd_host );
#if 0
        if (i >= MAX_DHCPC) {
            dhcpcAvailable = get_free_dhcpc_res_entry();
            if (( dhcpcAvailable < 0) || (dhcpcAvailable >= MAX_DHCPC)) {
                printf("no free dhcpc res entry\n");
                return;
            }
        }
#endif
        sprintf(nvram_name, "dhcpc_res%d", i);
        nvram_set(nvram_name, tmp);

    }

    while ( i < MAX_DHCPC) {
        sprintf(nvram_name, "dhcpc_res%d", i++);
        nvram_set(nvram_name, "");
    }


    //dhcps_cfg_save();
    //dhcps_start();

}

#if 0
//ATTACHED_DEVICES
static void
validate_dhcpc_res_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6;
    int i, j = 0, index = 1;
    char tmp[100], nvram_name[20];
    char *c;
    unsigned char mac[20];
    int ret = 0, NoOfDhcpcRsvdAddrFound = 0, dhcpcRsvdAddrCk = 0;
    int len = 0;

    ptr = websGetVar(wp, "subip", NULL);
    if (!ptr) {
        ptr = websGetVar(wp, "editiperr", NULL);
    }
    ptr1 = websGetVar(wp, "FixMacAdr", NULL);
    ptr2 = websGetVar(wp, "FixedRemark", NULL);
    memset(mac, '\0', sizeof(mac));
    //printf("ptr   = %s\n", ptr);
    //printf("ptr1 = %s\n", ptr1);

    for (i = 0; i < 12; i++) {
        //printf("ptr1[i] = %c\n", ptr1[i]);
        //printf("ptr1[i] = %d\n", ptr1[i]);
        mac[j++] = ptr1[i];
        if( (i == 1) ||( i == 3) || ( i == 5) ||( i == 7) || ( i == 9) )
            mac[j++]= ':';
    }
    mac[j] = NULL;

        //printf("mac = %s\n", mac);
    //printf("ptr2 = %s\n", ptr2);

    if (!ptr || !ptr1 || !ptr2)
        return;

    //printf("dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    //if ( dhcpcRsvdAddrCnt == 0) {
        //dhcpcRsvdAddrCk = 0;
        //_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].addrBound = 0;
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host, ptr2, strlen(ptr2));
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac, mac, sizeof(mac));
        //memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip, ptr, strlen(ptr));
        //dhcpcRsvdAddrCnt++;
    //}
    //else
    //{
        for (i = 0; i < dhcpcRsvdAddrCnt; i++) {
            if (strlen(ptr) > strlen(_dhcpc_rsvd_addr[i].rsvd_ip)) {
                len = strlen(ptr);
                printf("strlen(ptr) = %d\n", len);
            }
            else {
                len = strlen(_dhcpc_rsvd_addr[i].rsvd_ip);
                printf("strlen(_dhcpc_rsvd_addr[i].rsvd_ip = %d\n", strlen(_dhcpc_rsvd_addr[i].rsvd_ip));
            }
            //printf("strlen(_dhcpc_rsvd_addr[i].rsvd_mac) = %d\n", strlen(_dhcpc_rsvd_addr[i].rsvd_mac));
            //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_host);
            //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_mac);
            //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", i, _dhcpc_rsvd_addr[i].rsvd_ip);
            if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ptr, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) == NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) == NULL) {
                    printf("Found ip and mac matched with attached devices\n");
                    dhcpcRsvdAddrCk++;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ptr, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) == NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) != NULL) {
                    printf("Found ip matched attached devices %d\n", i);
                    dhcpcRsvdAddrCk++;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ptr, len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) != NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) == NULL) {
                    printf("Found mac matched attached devices %d\n", i);
                    dhcpcRsvdAddrCk++;
            }
            else if (memcmp(_dhcpc_rsvd_addr[i].rsvd_ip, ptr,  len/*sizeof(_dhcpc_rsvd_addr[i].rsvd_ip)*/) != NULL && memcmp(_dhcpc_rsvd_addr[i].rsvd_mac, mac, 17/*sizeof(_dhcpc_rsvd_addr[i].rsvd_mac)*/) != NULL) {
                    printf("Found no ip and mac matched in attached devices !\n");
                    //dhcpcRsvdAddrCk = 0;
            }
            else {
                printf("%s : unknow status\n", __FUNCTION__);
            }

            printf("dhcpcRsvdAddrCk = %d\n", dhcpcRsvdAddrCk);
            if ( dhcpcRsvdAddrCk == 1 ) {
                j = 0;
                j++;
                if (j < 1){
                    _dhcpc_rsvd_addr[i].addrBound = 0;
                    NoOfDhcpcRsvdAddrFound = i;
                    printf("NoOfDhcpcRsvdAddrFound = %d\n", NoOfDhcpcRsvdAddrFound);
                }
            } else if ( dhcpcRsvdAddrCk > 1 ) {
                _dhcpc_rsvd_addr[i].addrBound = 1;
                _dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].addrBound = 1;
                NoOfDhcpcRsvdAddrFound = -1;
            }

        }
    //}

    if (dhcpcRsvdAddrCk == 0 && dhcpcRsvdAddrCnt < 4) {
        _dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].addrBound = 0;
        memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host, ptr2, sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host));
        memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac, mac, sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac));
        memcpy(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip, ptr, sizeof(_dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip));
        //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", dhcpcRsvdAddrCnt, _dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_host);
        //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", dhcpcRsvdAddrCnt, _dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_mac);
        //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", dhcpcRsvdAddrCnt, _dhcpc_rsvd_addr[dhcpcRsvdAddrCnt].rsvd_ip);
        dhcpcRsvdAddrCnt++;
        if ( dhcpcRsvdAddrCnt >= MAX_DHCPC)
            printf("no free dhcpc res entry\n");

    }else if ( dhcpcRsvdAddrCk == 1 ) {
        printf("NoOfDhcpcRsvdAddrFound = %d\n", NoOfDhcpcRsvdAddrFound);
        _dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].addrBound = 0;
        memcpy(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_host, ptr2, sizeof(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_host));
        memcpy(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_mac, mac, sizeof(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_mac));
        memcpy(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_ip, ptr, sizeof(_dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_ip));
        //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", NoOfDhcpcRsvdAddrFound, _dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_host);
        //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", NoOfDhcpcRsvdAddrFound, _dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_mac);
        //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", NoOfDhcpcRsvdAddrFound, _dhcpc_rsvd_addr[NoOfDhcpcRsvdAddrFound].rsvd_ip);
    }
    //else
        //printf("%s : unknow status\n", __FUNCTION__);

    printf("dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    j = 0;
    for ( i = 0; i < dhcpcRsvdAddrCnt; i++) {
            printf("i = %d; j = %d\n", i, j);
            printf("_dhcpc_rsvd_addr[%d].addrBound = %d\n", i, _dhcpc_rsvd_addr[i].addrBound);
            if ( _dhcpc_rsvd_addr[i].addrBound == 0) {
                _dhcpc_rsvd_addr[j].addrBound = 0;
                memcpy(_dhcpc_rsvd_addr[j].rsvd_host, _dhcpc_rsvd_addr[i].rsvd_host, sizeof(_dhcpc_rsvd_addr[i].rsvd_host));
                memcpy(_dhcpc_rsvd_addr[j].rsvd_mac, _dhcpc_rsvd_addr[i].rsvd_mac, sizeof(_dhcpc_rsvd_addr[i].rsvd_mac));
                memcpy(_dhcpc_rsvd_addr[j].rsvd_ip, _dhcpc_rsvd_addr[i].rsvd_ip, sizeof(_dhcpc_rsvd_addr[i].rsvd_ip));
                //printf("_dhcpc_rsvd_addr[%d].rsvd_host = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_host);
                //printf("_dhcpc_rsvd_addr[%d].rsvd_mac  = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_mac);
                //printf("_dhcpc_rsvd_addr[%d].rsvd_ip   = %s\n", j, _dhcpc_rsvd_addr[j].rsvd_ip);
                j++;
            } else {
                _dhcpc_rsvd_addr[i].addrBound = 0;
                memset(_dhcpc_rsvd_addr[i].rsvd_host, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_host));
                memset(_dhcpc_rsvd_addr[i].rsvd_mac, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_mac));
                memset(_dhcpc_rsvd_addr[i].rsvd_ip, '\0', sizeof(_dhcpc_rsvd_addr[i].rsvd_ip));
            }
    }

    dhcpcRsvdAddrCnt = j;
    printf("******************* dhcpcRsvdAddrCnt = %d\n", dhcpcRsvdAddrCnt);
    for ( i = 0; i < dhcpcRsvdAddrCnt; i++) {
        /*sprintf(tmp, "%s,%s,%s",
        ptr,
        nvram_name,
        ptr2 );*/
        sprintf(tmp, "%s,%s,%s",
        _dhcpc_rsvd_addr[i].rsvd_ip,
        _dhcpc_rsvd_addr[i].rsvd_mac,
        _dhcpc_rsvd_addr[i].rsvd_host );
        //printf("dhcpc_res_cur_index = %d\n", dhcpc_res_cur_index);
        printf("tmp = %s\n", tmp);
        if (i >= MAX_DHCPC) {
             i = get_free_dhcpc_res_entry();
            if (( i < 0) || ( i >= MAX_DHCPC)) {
                printf("no free dhcpc res entry\n");
                return;
            }
        }
        sprintf(nvram_name, "dhcpc_res%d", i);
        nvram_set(nvram_name, tmp);

        dhcps_cfg_save();
        dhcps_start();
}


}
#endif

static void
validate_ddns_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

        struct variable fields[] = {

        { name: "DDNSHostName", longname: "DDNSHostName", nvram_name: "ddns_hostname" },
        { name: "DDNSLoginName", longname: "DDNSLoginName", nvram_name: "ddns_login" },
        { name: "DDNSLoginPassword", longname: "DDNSLoginPassword", nvram_name: "ddns_passwd" },

    };

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    ptr = websGetVar(wp, "DynamicDNS", "0");
    nvram_set("ddns_enable", (*ptr == '1') ? "1" : "0");

    ptr = websGetVar(wp, "wildcard", NULL);
    nvram_set("ddns_wildcard", (ptr) ? "1" : "0");

    ddns_cfg_save();        /* ddns setting */
    ddns_start();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}

static void
validate_upnp_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[20];

        struct variable fields[] = {

        { name: "AdverTime", longname: "AdverTime", nvram_name: "upnp_advertime" },
        { name: "TimeToLive", longname: "TimeToLive", nvram_name: "upnp_timetolive" },

    };

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    ptr = websGetVar(wp, "UPnP", "0");

    //nvram_set("upnp_enable", (*ptr == '1') ? "1" : "0");
    nvram_set("upnp_enable", (!strcmp(ptr, "on")) ? "1" : "0");

    upnp_cfg_save();        /* upnp setting */
    if( !strcmp(ptr, "on"))
        upnp_start();
    else
        upnp_stop();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}

static void
validate_log_action_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char path[64] = "/www/";

    ptr = websGetVar(wp, "logAction", "0");
    if (strcmp(ptr, "3") == 0) {    /* clear log */
        clear_syslog();
        log_clear_message(); //Add by Timo Wu 2006/01/14
        unlink("/var/url.log");
        ptr = websGetVar(wp, "page", "");
        strcat(path, ptr);
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
    }
        else if (strcmp(ptr, "5") == 0) {   /* send log */
        email_log();
        unlink("/var/url.log");
        strcat(path, "sendlog.html");
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
    }
    else {
        ptr = websGetVar(wp, "page", "");
        strcat(path, ptr);
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
    }

}

static void
validate_pot_setting(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char path[64] = "/www/";

    //Jenny 2006/01/05
    ptr1 = websGetVar(wp, "Resolution", NULL);
    ptr2 = websGetVar(wp, "Threshold", NULL);

    //printf("####### Resolution=%d Threshold=%d\r\n",atoi(ptr1),atoi(ptr2));

    if (ptr1 && ptr2)
    {

        if((atoi(ptr1)== POT_GetResolution()) && (atoi(ptr2) == POT_GetRepackThreshold()))
        {
         ;
        }
        else
        {
        //printf("####### POT_factoryDefault\r\n");
        POT_factoryDefault(atoi(ptr1), atoi(ptr2));
        }
    }
        //end
}

#if defined(_TRAFFIC_METER_)
extern int meter_last_sec;
#endif

extern int udhcpc_need_restart;
static void
validate_connection_action(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char path[64] = "/";

    ptr = websGetVar(wp, "pppoeAct", NULL);
    if (ptr) {
        if (strcmp(ptr, "1") == 0) {    /* pppoe connect */
            printf("pppoe connect\n");
#if defined(_TRAFFIC_METER_)
            clear_meter_related();
#endif
            pRomeCfgParam->pppoeCfgParam[0].uptime = 0;
            pRomeCfgParam->pppoeCfgParam[0].manualHangup = 0;
            pRomeCfgParam->pppoeCfgParam[0].manualStart = 1;
                        if(pRomeCfgParam->pppoeCfgParam[0].dialState == PPPOECFGPARAM_DIALSTATE_DIALED_SUCCESS) {
                                pppd_stop(0);
                sleep(3);
            }
            pppd_start(0);
        }
            else if (strcmp(ptr, "2") == 0) {   /* pppoe disconnect */
            printf("pppoe disconnect\n");
#if defined(_TRAFFIC_METER_)
            clear_meter_related();
#endif
   
            pRomeCfgParam->pppoeCfgParam[0].uptime = 0;
            if (pRomeCfgParam->pppoeCfgParam[0].dialState == PPPOECFGPARAM_DIALSTATE_DIALED_SUCCESS) {
#if defined(EXTENSION_RU) 
		pRomeCfgParam->pppoeCfgParam[0].manualHangup = 1;
		pRomeCfgParam->pppoeCfgParam[0].manualStart = 0;	    
#endif            	
                pppd_stop(0);           /* manualHanup = 1 */
                //if ((pRomeCfgParam->pppoeCfgParam[0].demand==1)) {
                //  pRomeCfgParam->pppoeCfgParam[0].dialState=PPPOECFGPARAM_DIALSTATE_DIALED_WAITING;
                //  AddDemandRoute(0,1);
                //}
            }
        }

        else if (strcmp(ptr, "3") == 0) {   /* dhcpc release */
               char *proto;

               proto = nvram_safe_get("wan_proto");
               if (strcmp(proto, "static") != 0) {
            printf("dhcpc release\n");
            send_signal("/var/run/udhcpc.pid", SIGUSR2);
            nvram_set("dhcpc_lease_obtained", "0");
            nvram_set("dhcpc_lease_expired", "0");

            //i=0;
            //memcpy(pRomeCfgParam->ifCfgParam[0].ipAddr,&i,4);
            //memcpy(pRomeCfgParam->ifCfgParam[0].ipMask,&i,4);
            //memcpy(pRomeCfgParam->ifCfgParam[0].gwAddr,&i,4);
            //memcpy(pRomeCfgParam->ifCfgParam[0].dnsPrimaryAddr,&i,4);
            //memcpy(pRomeCfgParam->ifCfgParam[0].dnsSecondaryAddr,&i,4);
               }

        }
            else if (strcmp(ptr, "4") == 0) {   /* dhcpc renew */
               char *proto;

               proto = nvram_safe_get("wan_proto");
               if (strcmp(proto, "static") != 0) {
            printf("dhcpc renew\n");
            //send_signal("/var/run/udhcpc.pid", SIGUSR1);
            udhcpc_renew();
               }

        }
            else if (strcmp(ptr, "5") == 0) {   /* bpa connect */
                FILE *fp;

            printf("bpa connect\n");
            pRomeCfgParam->telCfgParam.manualHangup = 0;
            pRomeCfgParam->telCfgParam.manualStart = 1;

            fp = fopen("/var/run/bpalogin.pid", "r");
            if (fp == NULL) {
                bpa_start();
            }
            else {
                fclose(fp);
                send_signal("/var/run/bpalogin.pid", SIGUSR1);
                sleep(2);
            }
        }
            else if (strcmp(ptr, "6") == 0) {   /* bpa disconnect */
            printf("bpa disconnect\n");
            //pRomeCfgParam->telCfgParam.manualHangup = 1;
            send_signal("/var/run/bpalogin.pid", SIGTERM);  //send signal:TERM to bpalogin task, it will trigger the logout process
            sleep(2);
        }
    }
    ptr = websGetVar(wp, "pptpAct", NULL);
    if (ptr) {
        if (strcmp(ptr, "1") == 0) {    /* pptp connect */
            printf("pptp connect\n");
#if defined(_TRAFFIC_METER_)
            clear_meter_related();
#endif
            pRomeCfgParam->pptpCfgParam.uptime = 0;
            pRomeCfgParam->pptpCfgParam.manualHangup = 0;
            pRomeCfgParam->pptpCfgParam.manualStart = 1;
            pptp_start(0);
        }
            else if (strcmp(ptr, "2") == 0) {   /* pptp disconnect */
            printf("pptp disconnect\n");
#if defined(_TRAFFIC_METER_)
            clear_meter_related();
#endif

            pRomeCfgParam->pptpCfgParam.uptime = 0;
            if (pRomeCfgParam->pptpCfgParam.dialState == PPTPCFGPARAM_DIALSTATE_DIALED_SUCCESS) {
#if defined(EXTENSION_RU) 
	    	pRomeCfgParam->pptpCfgParam.manualHangup = 1;
	    	pRomeCfgParam->pptpCfgParam.manualStart = 0;
#endif            	
                send_signal("/var/run/pppd0.pid",1);
                send_signal("/var/run/ppp0.pid",1);
                sleep(1);
                send_signal("/var/run/pptp.pid",9);
#if defined(DHCPPPTP)
				if (pRomeCfgParam->ifCfgParam[0].connType != IFCFGPARAM_CONNTYPE_DHCPPPTP)
#endif/*DHCPPPTP*/
                pRomeCfgParam->pptpCfgParam.dialState=PPTPCFGPARAM_DIALSTATE_OFF;
                //pRomeCfgParam->pptpCfgParam.manualHangup = 1;
                //if ((pRomeCfgParam->pptpCfgParam.demand==1)) {
                //  pRomeCfgParam->pptpCfgParam.dialState=PPPOECFGPARAM_DIALSTATE_DIALED_WAITING;
                //  AddDemandRoute(4,1); // 4: used for signal 44.
                //}
            }
        }
        else if (strcmp(ptr, "3") == 0) {   /* l2tp connect */
            printf("l2tp connect\n");
            pRomeCfgParam->l2tpCfgParam.manualHangup = 0;
            pRomeCfgParam->l2tpCfgParam.manualStart = 1;
            l2tp_start(0);
        }
            else if (strcmp(ptr, "4") == 0) {   /* l2tp disconnect */
            printf("l2tp disconnect\n");
            if (pRomeCfgParam->l2tpCfgParam.dialState == PPTPCFGPARAM_DIALSTATE_DIALED_SUCCESS) {
#if defined(EXTENSION_RU) 
		pRomeCfgParam->l2tpCfgParam.manualHangup = 1;
		pRomeCfgParam->l2tpCfgParam.manualStart = 0;	    
#endif            	
                send_signal("/var/run/pppd0.pid",1);
                send_signal("/var/run/ppp0.pid",1);
                sleep(1);
                send_signal("/var/run/l2tpd.pid",9);
                pRomeCfgParam->l2tpCfgParam.dialState=PPTPCFGPARAM_DIALSTATE_OFF;
                //pRomeCfgParam->pptpCfgParam.manualHangup = 1;
                //if ((pRomeCfgParam->l2tpCfgParam.demand==1)) {
                //  pRomeCfgParam->l2tpCfgParam.dialState=PPPOECFGPARAM_DIALSTATE_DIALED_WAITING;
                //  AddDemandRoute(6,1); // 6: used for signal 46.
                //}
            }
        }
    }

    ptr = websGetVar(wp, "page", "");
    strcat(path, ptr);
    //websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
    websHeader(wp);
    websWrite(wp, (char_t *) apply_refresh_header, 3, path);
    websFooter(wp);

}

//Yili 2006/1/10
#if defined(DNI)
static void
validate_mnu_login_action(webs_t wp, char *value, struct variable *v)
{
    char *ptr;
    int i;

    struct variable fields[] = {

        { name: "logo_enable", longname: "Logo enable", nvram_name: "logo_enable" },
        { name: "ver_check_enable", longname: "Version check enable", nvram_name: "ver_check_enable" },

    };

    for (i = 0; i < ARRAYSIZE(fields); i++) {
        ptr = websGetVar(wp, fields[i].name, NULL);
        if (ptr) {
            nvram_set(fields[i].nvram_name, ptr);
        }
    }

    mnu_login_cfg_save();

}
#endif
//end

//static void
static int
validate_qos_adding(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3;
    int i,ret;
    char tmp[200], nvram_name[20];
    vs_variable *vs_ptr=NULL;


    ptr = websGetVar(wp, "qosService", NULL);
    if (ptr == NULL)
        return TRUE;
    i = atoi(ptr);
    if (i > ARRAYSIZE(app_ports))
        return TRUE;

    vs_ptr = &(app_ports[i-1]);

    ptr = strtok(vs_ptr->ports_ptr, ",:");

    //Add by Timo wu
    ret = check_qos_entry( vs_ptr->name, ptr , "BLANK" , MAX_QOS , "BLANK");
    switch(ret)
    {
        case FORWARD_PORT_CONFLICT:
        case NAME_CONFLICT:
            return NAME_CONFLICT;
        default:
            break;
    }

    i = 0;
    while (ptr) {
        i = get_free_qos_entry();
        if ((i < 0) || (i >= MAX_QOS)) {
            printf("no free qos entry\n");
            return TRUE;
        }
        sprintf(nvram_name, "qos_port%d", i);
        sprintf(tmp, "%s,%s,on,%d,%s",
                ptr,
                ((vs_ptr->protocol == _TCP+_UDP) ? "both" :
                    ((vs_ptr->protocol == _TCP) ? "tcp" : "udp")),
                1,      /* always setting is high priority */
                vs_ptr->name);
        nvram_set(nvram_name, tmp);
        ptr = strtok(NULL, ",:");
        i++;
    }
    qos_cfg_save();
    qos_tableDriverAccess();

    return TRUE;
}


extern int qos_cur_index;

/* d-d,proto,on/off,priority(0/1),desc */

static void
validate_qos_editing(webs_t wp, char *value, struct variable *v)
{

    char *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5, *ptr6, *ptr7;
    int i, ret;
    char tmp[100], nvram_name[20];

    sprintf(nvram_name, "qos_port%d", qos_cur_index);

    ptr = websGetVar(wp, "GlobalPort", NULL);   /* start port */
    ptr1 = websGetVar(wp, "LocalPort", NULL);   /* end port */
    ptr2 = websGetVar(wp, "ServiceType", NULL);

    //ptr6 = websGetVar(wp, "VirSerAct", NULL);
    //ptr7 = websGetVar(wp, "ServiceType", NULL);

    if (!ptr || !ptr1 || !ptr2)
        return;

    //Add by Timo wu
    ret = check_qos_entry( "BLANK", ptr, ptr1, qos_cur_index, "edit");
    switch(ret)
    {
        case NAME_CONFLICT:
        case FORWARD_PORT_CONFLICT:
        {
            char path[64] = "/www/error_port_serviceport_exist.html";

            ptr = websGetVar(wp, "page", "");
            if (!ptr)
                    sprintf(return_page, "/%s", DirectoryIndex);
                else
                    sprintf(return_page, "/%s", ptr);

                noread = 1;
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            action = NOREPLY; //only reply error page, not reply previous page in apply_cgi
            return;
        }
        default:
            break;
    }

    sprintf(tmp, "%s-%s,%s,%s,%d,%s",
            ptr,
            ptr1,
            ((atoi(ptr2) == 6) ? "tcp" : ((atoi(ptr2) == 17) ? "udp" : "both")),
            "on",               /* always on */
            1,
            websGetVar(wp, "Remark", "")
            );
    if (qos_cur_index >= MAX_QOS) {
        i = get_free_qos_entry();
        if ((i < 0) || (i >= MAX_QOS)) {
            printf("no free qos entry\n");
            return;
        }
        sprintf(nvram_name, "qos_port%d", i);
    }

    nvram_set(nvram_name, tmp);
    qos_cfg_save();
    qos_tableDriverAccess();
    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

}




int qos_cur_index = MAX_QOS;

static void
validate_qos_setting(webs_t wp, char *value, struct variable *v)
{
    char *ptr, *ptr1, *ptr2, *ptr3;
    int i;
    char tmp[200], nvram_name[20];
    vs_variable *vs_ptr=NULL;


    ptr = websGetVar(wp, "Button", NULL);
    if (ptr == NULL)
        return;
    if (ptr) {
        if (!strcmp(ptr, "Add_Qos")) {

            //Add by Timo Wu
            if( validate_qos_adding(wp, value, v)== NAME_CONFLICT)
            {
                //printf("validate_service_port_adding is FALSE\n");
                char path[64] = "/www/error_port_servicename_exist.html";
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                    else{
                        sprintf(return_page, "/%s", ptr);
                    }
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                return;
            }
            else
            {
                //printf("validate_service_port_adding is TRUE\n");
            ptr = websGetVar(wp, "page", NULL);
            if (!ptr) {
                sprintf(tmp, "/www/%s", DirectoryIndex);
            }
            else
                sprintf(tmp, "/www/%s", ptr);

            websDefaultHandler(wp, urlPrefix, webDir, arg, url, tmp , query);
            }
        }
        //else if (strstr(ptr, "Add")) {
        else if (!strcmp(ptr, "Add")) {
            /* add service block */
            char path[64] = "/www/qos_custom.html";

            qos_cur_index = MAX_QOS;

            websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html

        }
        //else if (strstr(ptr, "Modify")) {
        else if (!strcmp(ptr, "Modify")) {
            /* edit service */
            char path[64] = "/www/qos_custom.html";

            ptr = websGetVar(wp, "R1", NULL);
            if (ptr) {
                qos_cur_index = atoi(ptr);

                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);//Brian+ show html
            }
            else {
                ptr = websGetVar(wp, "page", "");
                if (!ptr)
                        sprintf(return_page, "/%s", DirectoryIndex);
                        else
                            sprintf(return_page, "/%s", ptr);
                strcpy(path, "/www/");
                strcat(path, "edit_forward_err.html");
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
            }

        }
        //else if (strstr(ptr, "Delete")) {
        else if (!strcmp(ptr, "Delete")) {
            char path[64] = "/www/qos_custom.html";

            ptr = websGetVar(wp, "onlydel", "0");
            if (strcmp(ptr, "1") == 0) {
                /* delete this service */
                ptr = websGetVar(wp, "R1", NULL);
                if (ptr) {
                    i = atoi(ptr);

                    sprintf(nvram_name, "qos_port%d", i);
                    nvram_set(nvram_name, "");
                    ptr = websGetVar(wp, "page", "");
                    strcpy(path, "/www/");
                    if (!ptr)
                            strcat(path, DirectoryIndex);
                            else
                                strcat(path, ptr);
                            //strcpy(return_page, path);
                    if (!ptr)
                            sprintf(return_page, "/%s", DirectoryIndex);
                            else
                                sprintf(return_page, "/%s", ptr);

                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);
                    //websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/test_conn1.html" , NULL);

                    qos_cfg_save();
                    qos_tableDriverAccess();
                        /* avoid already opens port still work (HNAPT) */
                    //action = RESTART;
                    //sysReinit(pRomeCfgParam->ifCfgParam[0].connType);

                }
                else {

                    ptr = websGetVar(wp, "page", "");
                    if (!ptr)
                            sprintf(return_page, "/%s", DirectoryIndex);
                            else
                                sprintf(return_page, "/%s", ptr);
                    strcpy(path, "/www/");
                    strcat(path, "del_forward_err.html");
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, path , NULL);

                }

            }
        }
    }


}

#if defined(_TRAFFIC_METER_)
extern traffic_meter WAN_meter;

void check_meter_wan_restart(void)
{

    //if ((WAN_meter.thismonth.reached_hi_water & 2) && pRomeCfgParam->trafficMeterCfgParam.led_on) {
    if (pRomeCfgParam->trafficMeterCfgParam.led_on) {
        if (get_wanstatus())
            setLedLight(1/*WAN_LED*/, 0 /*LED_GREEN*/);
        else
            setLedLight(1/*WAN_LED*/, 1 /*LED_AMBER*/);
    }
    //if (WAN_meter.thismonth.limited_actived && pRomeCfgParam->trafficMeterCfgParam.block_on) {
    if (WAN_meter.thismonth.limited_actived) {
        clear_meter_status(&(WAN_meter.thismonth));
        httphijack_del();
#if 1
        clear_meter_related();
        send_signal("/var/run/boa.pid",71);     /* start wan */
        //printf("start_wan(), because restart counters !!!!!\n");
        //start_wan();
#else
        char *ptr = nvram_get_wan_proto("wan_proto");
        if (strcmp(ptr, "pppoe") == 0) {
            printf("pppoe connect\n");
            clear_meter_related();
            pRomeCfgParam->pppoeCfgParam[0].uptime = 0;
            pRomeCfgParam->pppoeCfgParam[0].manualHangup = 0;
            pRomeCfgParam->pppoeCfgParam[0].manualStart = 1;
            pppd_start(0);
        }
        else if (strcmp(ptr, "pptp") == 0) {
            printf("pptp connect\n");
            clear_meter_related();
            pRomeCfgParam->pptpCfgParam.uptime = 0;
            pRomeCfgParam->pptpCfgParam.manualHangup = 0;
            pRomeCfgParam->pptpCfgParam.manualStart = 1;
            pptp_start(0);
        }
#endif      /* #if 1 */

    }


}

extern int need_update_statistics_flash;
void restart_meter_counter(void)
{
    printf("%s\n", __FUNCTION__);
    /* restart counter */
    memset((char *)&WAN_meter.today, 0x0, sizeof(traffic_meter_unit));
    memset((char *)&WAN_meter.yesterday, 0x0, sizeof(traffic_meter_unit));
    memset((char *)&WAN_meter.thisweek, 0x0, sizeof(traffic_meter_unit));
    nvram_set("watermark_reached", "0");
    nvram_set("watermark_showed", "0");
    nvram_set("limited_reached", "0");
    nvram_set("limited_actived", "0");
    check_meter_wan_restart();

    memset((char *)&WAN_meter.thismonth, 0x0, sizeof(traffic_meter_unit));
    WAN_meter.thismonth.restarted_counter = 1;
    //need_update_statistics_flash = 1;
}


static void
validate_traffic_meter(webs_t wp, char *value, struct variable *v)
{
    char *ptr;
    int control_enable = 0, control_type = 0, control_dir = 0, block_on = 0;
    int new_control_enable = 0, new_control_type = 0, new_control_dir = 0, new_block_on = 0;
    int vol_limit = 0, time_limit = 0;
    int wan_restart = 0;

    ptr = websGetVar(wp, "Button", NULL);
    if (!ptr)
        return;
    if (strstr(ptr, "Restart")) {
#if 0
        memset((char *)&WAN_meter.today, 0x0, sizeof(traffic_meter_unit));
        memset((char *)&WAN_meter.yesterday, 0x0, sizeof(traffic_meter_unit));
        memset((char *)&WAN_meter.thisweek, 0x0, sizeof(traffic_meter_unit));
        if ((WAN_meter.thismonth.reached_hi_water & 2) && pRomeCfgParam->trafficMeterCfgParam.led_on) {
            if (get_wanstatus())
                setLedLight(1/*WAN_LED*/, 0 /*LED_GREEN*/);
            else
                setLedLight(1/*WAN_LED*/, 1 /*LED_AMBER*/);
        }
        nvram_set("watermark_reached", "0");
        nvram_set("watermark_showed", "0");
        nvram_set("limited_reached", "0");
        nvram_set("limited_actived", "0");
        if (WAN_meter.thismonth.limited_actived) {
            send_signal("/var/run/boa.pid",71);     /* start wan */
            //printf("start_wan(), because restart counters !!!!!\n");
            //start_wan();
        }
        memset((char *)&WAN_meter.thismonth, 0x0, sizeof(traffic_meter_unit));
#endif
        restart_meter_counter();
        meter_statistics_update_flash(&WAN_meter);
    }
    else {  /* apply */
        int i;
        char *ptr, *ptr1, *ptr2, *ptr3;

        struct variable fields[] = {

            { name: "traffic_on", longname: "traffic_on", nvram_name: "traffic_control_enable" },
            { name: "contrl_type", longname: "contrl_type", nvram_name: "traffic_control_type" },
            { name: "contrl_dir", longname: "contrl_dir", nvram_name: "traffic_control_dir" },

            { name: "volume_monthly_limit", longname: "volume_monthly_limit", nvram_name: "volume_monthly_limit" },
            { name: "round_up_volume", longname: "round_up_volume", nvram_name: "round_up_volume" },
            { name: "conntime_monthly_limit", longname: "conntime_monthly_limit", nvram_name: "conntime_monthly_limit" },
            { name: "waterMark", longname: "waterMark", nvram_name: "hi_water" },
            { name: "led_on", longname: "led_on", nvram_name: "traffic_led_on" },
            { name: "block_on", longname: "block_on", nvram_name: "traffic_block_on" },



        };

        control_enable = atoi(nvram_get("traffic_control_enable"));
        control_type = atoi(nvram_get("traffic_control_type"));
        control_dir = atoi(nvram_get("traffic_control_dir"));
        block_on = atoi(nvram_get("traffic_block_on"));

        vol_limit = atoi(nvram_get("volume_monthly_limit"));
        time_limit = atoi(nvram_get("conntime_monthly_limit"));

        trafficMeterCfgParam_t tmp;
        memcpy((char *)&tmp, (char*)&(pRomeCfgParam->trafficMeterCfgParam), sizeof(pRomeCfgParam->trafficMeterCfgParam));

        for (i = 0; i < ARRAYSIZE(fields); i++) {
            ptr = websGetVar(wp, fields[i].name, NULL);
            if (ptr) {
                nvram_set(fields[i].nvram_name, ptr);
            }
        }

        ptr = websGetVar(wp, "hour", NULL); /* 0-11 */
        ptr1 = websGetVar(wp, "min", NULL); /* 0-59 */
        ptr2 = websGetVar(wp, "ampm_sel", NULL);    /* am/pm */
        ptr3 = websGetVar(wp, "dayofmonth", NULL);  /* 1st, 2nd ... */
        if ((ptr == NULL) || (ptr1 == NULL) || (ptr2 == NULL) || (ptr3 == NULL)) {
            ;   /*  return; */
        }
        else {
            printf("hour=%s, min=%s, ampm_sel=%s, dayofmonth=%s\n", ptr, ptr1, ptr2, ptr3);
            pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour = atoi(ptr);
            if (ptr2[0] == '1') {   /* pm */
                if (pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour < 12)
                    pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour += 12;
                if (pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour >= 24)
                    pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour  = 0;
            }
            else
                if (pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour == 12)
                    pRomeCfgParam->trafficMeterCfgParam.tm.tm_hour = 0;

            pRomeCfgParam->trafficMeterCfgParam.tm.tm_min = atoi(ptr1);

            pRomeCfgParam->trafficMeterCfgParam.tm.tm_mday = atoi(ptr3);
            printf("pRomeCfgParam->trafficMeterCfgParam.tm.tm_mday = %d, ptr3=%s\n",
                pRomeCfgParam->trafficMeterCfgParam.tm.tm_mday, ptr3);
        }
        traffic_meter_cfg_save();

        //if (atoi(nvram_get("traffic_control_enable")) == 0) {
        //  wan_restart = 1;
        //}

        //if (atoi(nvram_get("traffic_control_type")) == 0) {
        //  if (atoi(nvram_get("traffic_control_dir")) == 0) {
        //      wan_restart = 1;
        //  }
        //}
        new_control_enable = atoi(nvram_get("traffic_control_enable"));
        new_control_type = atoi(nvram_get("traffic_control_type"));
        new_control_dir = atoi(nvram_get("traffic_control_dir"));
        new_block_on = atoi(nvram_get("traffic_block_on"));

        if ((new_control_enable == 0) || ((new_control_type == 0) && (new_control_dir == 0))) {
            //clear_meter_status(&(WAN_meter.thismonth));
        }

        if (control_enable != new_control_enable)
            wan_restart = 1;

        if (control_type != new_control_type)
            wan_restart = 1;

        if (control_dir != new_control_dir)
            wan_restart = 1;

        if (block_on != new_block_on) {
            wan_restart = 1;
        }

        if ((new_control_enable == 1) && ((new_control_type == 0) && (new_control_dir != 0))) {
            if (vol_limit != atoi(nvram_get("volume_monthly_limit")))
                wan_restart = 1;
        }

        if ((new_control_enable == 1) && (new_control_type == 1)) {
            if (time_limit != atoi(nvram_get("conntime_monthly_limit")))
                wan_restart = 1;
        }


        if (wan_restart) {
            if (block_on && !pRomeCfgParam->trafficMeterCfgParam.block_on && pRomeCfgParam->trafficMeterCfgParam.limited_reached) {
                nvram_set("watermark_reached", "0");
                nvram_set("watermark_showed", "0");
                nvram_set("limited_reached", "0");
                nvram_set("limited_actived", "0");
                //WAN_meter.thismonth.over_total_adjust_k = 0;
                //WAN_meter.thismonth.over_download_adjust_k = 0;
                check_meter_wan_restart();
            }
            else if (!block_on && pRomeCfgParam->trafficMeterCfgParam.block_on && pRomeCfgParam->trafficMeterCfgParam.limited_reached) {
                clear_meter_status(&(WAN_meter.thismonth));
                check_traffic_meter(0);
#if 0
                int ret = check_hi_water(&(WAN_meter.thismonth), 0);
                if ((ret & 0x2) == 0) {
                    nvram_set("watermark_reached", "0");
                    nvram_set("watermark_showed", "0");
                    nvram_set("limited_reached", "0");
                    nvram_set("limited_actived", "0");
                    //WAN_meter.thismonth.over_total_adjust_k = 0;
                    //WAN_meter.thismonth.over_download_adjust_k = 0;
                    check_meter_wan_restart();
                }
#endif
            }
            else if (pRomeCfgParam->trafficMeterCfgParam.block_on && pRomeCfgParam->trafficMeterCfgParam.limited_reached) {
            //if (pRomeCfgParam->trafficMeterCfgParam.block_on && WAN_meter.thismonth.limited_actived) {
                int ret = check_hi_water(&(WAN_meter.thismonth), 0);
                if ((ret & 0x2) == 0) {
                    nvram_set("watermark_reached", "0");
                    nvram_set("watermark_showed", "0");
                    nvram_set("limited_reached", "0");
                    nvram_set("limited_actived", "0");
                    //WAN_meter.thismonth.over_total_adjust_k = 0;
                    //WAN_meter.thismonth.over_download_adjust_k = 0;
                    check_meter_wan_restart();
                }
            }
            else {
                clear_meter_status(&(WAN_meter.thismonth));
                check_traffic_meter(0);
            }
        }

        /* will need to check cfg changed action */
        //if (pRomeCfgParam->trafficMeterCfgParam.enable == 0) {
        //  httphijack_del();
        //}

    }
}
#endif  /* #if defined(_TRAFFIC_METER_) */



/*
 * Variables are set in order (put dependent variables later). Set
 * nullok to TRUE to ignore zero-length values of the variable itself.
 * For more complicated validation that cannot be done in one pass or
 * depends on additional form components or can throw an error in a
 * unique painful way, write your own validation routine and assign it
 * to a hidden variable (e.g. filter_ip).
 *
 * EZC_FLAGS_READ : implies the variable will be returned for ezconfig read request
 * EZC_FLAGS_WRITE : allows the variable to be modified by the ezconfig tool
 *
 * The variables marked with EZConfig have to maintain backward compatibility.
 * If they cannot then the ezc_version has to be bumped up
 *
 */
struct variable variables[] = {
    { "L", "LAN Setting", validate_lan_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, "lan_ipaddr" },
    { "W", "WAN Setting", validate_wan_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, "wan_proto" },
    { "K", "FW Keyword Block Setting", validate_fw_site_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "I", "FW Packet Block Setting", validate_fw_service_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "I1", "FW Packet Block Setting", validate_fw_service_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "F", "FW Schedule Setting", validate_fw_schedule_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "E", "FW Port Trigger Setting", validate_fw_porttrigger_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "E1", "FW Port Trigger Setting", validate_fw_porttrigger_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "EF", "Email Setting", validate_email_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "A", "Password Setting", validate_password_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "P", "WAN_Other Setting", validate_wan_other_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "R", "Management Setting", validate_management_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "VSP", "Service Forward Port Adding", validate_service_port_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "VSA", "Service Forward Port Adding", validate_service_port_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "VSI", "Service Forward Port Adding", validate_service_port_editing, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "D", "DDNS setting", validate_ddns_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "B", "UPNP_setting", validate_upnp_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},

    { "S", "Static Route Adding", validate_static_route_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "S1", "Static Route Adding", validate_static_route_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},

    { "H", "Dhcpc reservation adding", validate_dhcpc_res_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},

    { "LD", "Log Action", validate_log_action_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},

    { "POT", "POT Action", validate_pot_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "QA", "Qos Adding", validate_qos_adding, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "QS", "QoS Adding", validate_qos_setting, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
    { "QE", "QoS Adding", validate_qos_editing, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},

#if defined(_TRAFFIC_METER_)
    { "TM", "Traffic Meter", validate_traffic_meter, NULL, FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE | EZC_FLAGS_NVRAM_NAME, NULL},
#endif

};


int
variables_arraysize(void)
{
    return ARRAYSIZE(variables);
}

static void
validate_cgi(webs_t wp)
{
    struct variable *v;
    char *value;

    //websBufferInit(wp);

    value = websGetVar(wp, "Load", NULL);
    if (value) {
        for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {

            if (strcmp(value, v->name) != 0)
                continue;
            if (v->validate) {
                v->validate(wp, value, v);
                return;
            }
        }
    }

    /* Validate and set variables in table order */
    for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
        if (!(value = websGetVar(wp, v->name, NULL)))
        {
            continue;
        }
        if ((!*value && v->nullok) || !v->validate)
        {

            if( !strcmp(v->name, "fpl") || !strcmp(v->name, "macfl") )
                v->validate(wp, value, v);
            else {
                if (v->ezc_flags & EZC_FLAGS_NVRAM_NAME)
                  nvram_set(v->nvram_name, value);
                else
                  nvram_set(v->name, value);
            }
        }
        else
        {
            v->validate(wp, value, v);
        }
    }

}


/* Copy each token in wordlist delimited by space into word */
#define foreach(word, wordlist, next) \
    for (next = &wordlist[strspn(wordlist, " ")], \
         strncpy(word, next, sizeof(word)), \
         word[strcspn(word, " ")] = '\0', \
         word[sizeof(word) - 1] = '\0', \
         next = strchr(next, ' '); \
         strlen(word); \
         next = next ? &next[strspn(next, " ")] : "", \
         strncpy(word, next, sizeof(word)), \
         word[strcspn(word, " ")] = '\0', \
         word[sizeof(word) - 1] = '\0', \
         next = strchr(next, ' '))

extern char *nvram_get(char *name);


static char * get_arg(char *args, char **next);
static void call(char *func, webs_t stream);

//#define DPRINTF       printf
#define DPRINTF


void
do_ej(char *path, webs_t *stream)
{
    FILE *fp = NULL;
    int c;
    char pattern[100], *asp = NULL, *func = NULL, *end = NULL;
    int len = 0;


    if (!(fp = fopen(path, "r")))
    {
        printf/*DPRINTF*/( "%s:: fopen [%s] error [%s]\n", __FUNCTION__, path, strerror( errno) );
        return;
    }

    while ((c = getc(fp)) != EOF)
    {
        /* Add to pattern space */
        pattern[len++] = c;
        pattern[len] = '\0';
        if (len == (sizeof(pattern) - 1))
            goto release;


        /* Look for <% ... */
        if (!asp && !strncmp(pattern, "<%", len))
        {
            if (len == 2)
                asp = pattern + 2;
            continue;
        }

        /* Look for ... %> */
        if (asp)
        {
            if (unqstrstr(asp, "%>"))
            {
                for (func = asp; func < &pattern[len]; func = end)
                {
                    /* Skip initial whitespace */
                    for (; isspace((int)*func); func++);
                    if (!(end = unqstrstr(func, ";")))
                        break;
                    *end++ = '\0';

                    /* Call function */
                    call(func, stream);
                }
                asp = NULL;
                len = 0;
            }
            continue;
        }

release:
        /* Release pattern space */
        //websWrite(stream, pattern);
        req_write(stream, pattern);
        asp = NULL;
        len = 0;
    }
    req_flush(stream);

    if (fp != NULL)
        fclose(fp);
}



extern client_req_t client_req;
char spoof_mac[20]={0};
//char macstr[17] ={0};
//#define ADD_VLAN            2
/*Add BlankState by Kevin.Chang@DNI  */
extern int rtl8651_checkEthernetPortStatus(unsigned int port, int *up,int *speed, int *duplex, int *range);
extern int SetSmtWzdDone(int Done);
void
TraceWanDet_cgi(webs_t wp, char_t *path, char_t **query)
{
    char *value;
    char *page;
       char strURL[] ="/Blank_detwan.html";
       char *path;

    wizard_wan_spoofmac_detected = 0;
    int curState, past_connType = GetCurWanConnType();

    init_form_cgi(wp, query);

    page = websGetVar(wp, "page", "");
    curState = atoi(nvram_get("SmtWzdStep"));
    if(strstr(page,"basicsetting") ) {
#if defined(CONFIG_WAN_LAN_MATCH_DETECTION)
        int WanLanFlag = atoi(nvram_get("WL_ip_match"));
        if(WanLanFlag) {
            Set_Flag_WAN_LAN_MATCH_DET(0);
        }
#endif
        SetSmtWzdDone(1);
        websHeader(wp);
           websWrite(wp, (char_t *) apply_refresh_header, 0, "/rel_index.html");
        websFooter(wp);
        return;
    }
    if(strstr(page,"ErrorAuth") ) {
        client_req.authed_send = 0;
#if 0
        websWrite(wp,"<html><head>");
        websWrite(wp,"<script language=\"JavaScript\">");
        websWrite(wp,"top.location.href=\"/ErrorAuthCancel.html\"");
        websWrite(wp,"</script>");
        websWrite(wp,"</head><body></body></html>");
#else
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/ErrorAuthCancel.html" , NULL);
#endif
        return;
    }

    if(strstr(page,"wizard_select") ) {
        char *DetSel;
        DetSel = websGetVar(wp, "WANDetc","");
        if(!strcmp(DetSel,"AutoDetc")) {
                nvram_set("SmtWzdStep","1");
                curState=1;
        } else {
            //nvram_set("SmtWzdState","0");
            SetSmtWzdDone(1);
            websHeader(wp);
               websWrite(wp, (char_t *) apply_refresh_header, 0, "/basicform1.html");
            websFooter(wp);
            return;
        }
    }

    if(!strcmp(page,"wizard_select.html")) {
        curState=1;
    }

    value = websGetVar(wp, "SW_Step", NULL);
    if(value != NULL) {
        curState = atoi(value);
        if(curState == 5) {
            curState = atoi(nvram_get("SmtBDetWanType"));
        }
        if(curState == 6) {
            curState = atoi(nvram_get("SmtBTestWanConn"));
        }
        //printf("@KKDEBUG@:%s: page is %s,value is %s, and SmtWzdState is %d\n",
        //  __FUNCTION__,
        //  page, value, curState);
    }

    switch(curState) {
        case 1:
            if(strstr(page,"welcome") || strstr(page,"wan_discon") || strstr(page,"wizard_select")) {

                {   //Detected WAN port is plugin?

                    int RetVal, up, speed, duplex, range;
                    RetVal = rtl8651_checkEthernetPortStatus(0, &up, &speed, &duplex, &range);
                    if(!up) {
                        websHeader(wp);
                           websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_wan_discon.html");
                        websFooter(wp);
                        return;
                    }
                }

#if 0
                //Pop the detecting page
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detectwan.html");
                websFooter(wp);
#else
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_detectwan.html", NULL);
#endif
                nvram_set("SmtWzdStep","2");
                nvram_set("SmtBDetWanType","2");

                if(!fork()) {
                    //Run WAN detect wizard
                    int SmtWzdState=0;
                    int i;

                    strcpy(spoof_mac, wizard_getclonemac(wp->fd));
                    close(wp->fd);
                    upgrading = 1;
                    rtl8651a_setDosStatus(0,0);
		    pRomeCfgParam->pptpCfgParam.manualHangup = 1;                    
		    pRomeCfgParam->pppoeCfgParam[0].manualHangup = 1;
		    pRomeCfgParam->telCfgParam.manualHangup = 1;		    
                    stop_wan();
                    sleep(2);
                    {
                    	int tmp;
                    	
		    	pRomeCfgParam->pptpCfgParam.manualHangup = 0;                    
		    	pRomeCfgParam->pppoeCfgParam[0].manualHangup = 0;
		    	pRomeCfgParam->telCfgParam.manualHangup = 0;                    	
		        tmp = pRomeCfgParam->ifCfgParam[0].connType;                    
                        nvram_set("wan_proto", "dhcp");
    		        start_wan(); 
                        rtl8651a_setDosStatus(0,0);    		    
    		        sleep(2); 
			pRomeCfgParam->ifCfgParam[0].connType = tmp;    		        
    		    }                     
             

                    SmtWzdState=wizard_go();
                    upgrading = 0;
                    nvram_set("wizard_wan_spoofmac_detected", "0");
                    //printf("SmtWzdState = %d\n", SmtWzdState);
                    switch(SmtWzdState) {
                        case 1: //DHCP
                            nvram_set("SmtWzdStep","10");
                            nvram_set("SmtBDetWanType","10");
                            break;
                        case 2: //PPPOE
                            nvram_set("SmtWzdStep","20");
                            nvram_set("SmtBDetWanType","20");
                            break;
                        case 3: //PPTP
                            nvram_set("SmtWzdStep","30");
                            nvram_set("SmtBDetWanType","30");
                            break;
                        case 4: //FixedIP
                            nvram_set("SmtWzdStep","40");
                            nvram_set("SmtBDetWanType","40");
                            break;
                        case 5: //BigPond
                            nvram_set("SmtWzdStep","50");
                            nvram_set("SmtBDetWanType","50");
                            break;
                        case 6: //spoofmac DHCP
                            nvram_set("SmtWzdStep","10");
                            nvram_set("SmtBDetWanType","10");
                            nvram_set("wizard_wan_spoofmac_detected", "1");
                            break;
                        case 7: //spoofmac Bigpond
                            nvram_set("SmtWzdStep","50");
                            nvram_set("SmtBDetWanType","50");
                            nvram_set("wizard_wan_spoofmac_detected", "1");
                            break;
                        default: //FixedIP
                            nvram_set("SmtWzdStep","40");
                            nvram_set("SmtBDetWanType","40");
                            break;
                    }
//                  nvram_set("SmtBDetWanType","40");
//                  nvram_set("SmtBDetWanType","10");
                    exit(1);
                }
            } else {
                websWrite(wp,"<html><head><title></title><body>");
                websWrite(wp,"TraceWanDet.cgi error....");
                websWrite(wp,"current state is %d  ",curState);
                websWrite(wp,"current page is %s  ",page);
                websWrite(wp,"</body></html>");
            }

            break;

        case 2:
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_detectwan.html", NULL);

            /*if(strstr(page,"detectwan")) {
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detectwan.html");
                websFooter(wp);
            } else {
                websWrite(wp,"<html><head><title></title><body>");
                websWrite(wp,"TraceWanDet.cgi error....");
                websWrite(wp,"current state is %d  ",curState);
                websWrite(wp,"current page is %s  ",page);
                websWrite(wp,"</body></html>");
            } */
            break;

        case 3:
            if(strstr(page,"detectwan")) {
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detectwan.html");
                websFooter(wp);
            } else {
                websWrite(wp,"<html><head><title></title><body>");
                websWrite(wp,"TraceWanDet.cgi error....");
                websWrite(wp,"current state is %d  ",curState);
                websWrite(wp,"current page is %s  ",page);
                websWrite(wp,"</body></html>");
            }
            break;
        case 10:        //DHCP
            if(strstr(page, "is_dhcp")) {
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);

                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","12");
                    if ( atoi(nvram_get("wizard_wan_spoofmac_detected")) ){
                        wizard_wan_spoofmac_detected = 1;
                        eth0_with_wan_spoofmac_up();//change wan0 mac to pc mac
                    }
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(g_pastWanType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_is_dhcp.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_is_dhcp.html" , NULL);
#endif
            }
            break;

        case 12:
            nvram_set("SmtWzdStep","14");
            nvram_set("SmtBTestWanConn","14");
#if 0
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detwan.html");
            websFooter(wp);
#else
            /* modified by lucas */
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_detwan.html" , NULL);

#endif

            if(!fork()) {
                extern int h_errno;
                struct hostent *NameInfo;
                char UrlName[]={"WWW.NETGEAR.COM"};

                close(wp->fd);
                NameInfo = gethostbyname((char*)UrlName);
                if(h_errno) {
                    sleep(1);
                    NameInfo = gethostbyname((char*)UrlName);
                }
                if(h_errno) {
                    nvram_set("SmtWzdStep","16");
                    nvram_set("SmtBTestWanConn","16");
                } else {
                    nvram_set("SmtWzdStep","101");
                    nvram_set("SmtBTestWanConn","101");
                }
                exit(1);
            }
            break;

        case 14:
#if 0
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detwan.html");
            websFooter(wp);
            break;
#else
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_detwan.html" , NULL);
#endif
        case 16:
            if(strstr(page, "ipdynamic")) {
#if 0               //Kevin:06Feb17:TestButton
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
#else
                char* Test;
                Test = websGetVar(wp, "DoTest","");

                if(!strcmp(Test,"yes")) {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Test_update_proc.html" , NULL);
                } else {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
                }
#endif

                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","17");
                    nvram_set("SmtBDetWanType","17");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_wizard_ipdynamic.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_wizard_ipdynamic.html" , NULL);
#endif
            }
            break;

        case 17:
            nvram_set("SmtWzdStep","14");
            nvram_set("SmtBTestWanConn","14");
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_detwan.html" , NULL);

            if(!fork()) {
                extern int h_errno;
                struct hostent *NameInfo;
                char UrlName[]={"WWW.NETGEAR.COM"};

                close(wp->fd);
                NameInfo = gethostbyname((char*)UrlName);
                if(h_errno) {
                    sleep(1);
                    NameInfo = gethostbyname((char*)UrlName);
                }
                if(h_errno) {
                    nvram_set("SmtWzdStep","100");
                    nvram_set("SmtBTestWanConn","100");
                } else {
                    nvram_set("SmtBTestWanConn","101");
                    nvram_set("SmtWzdStep","101");
                }
                exit(1);
            }
            break;


        case 20:        //PPPOE
            if(strstr(page, "pppoe")) {
#if 0               //Kevin:06Feb17:TestButton
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
#else
                char* Test = websGetVar(wp, "DoTest","");

                if(!strcmp(Test,"yes")) {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Test_update_proc.html" , NULL);
                } else {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
                }

#endif

                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","98");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_is_pppoe.html");
                websFooter(wp);
#else
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_is_pppoe.html" , NULL);
#endif
            }
            break;

        case 30:        //PPTP
            if(strstr(page, "PPTP")) {
#if 0               //Kevin:06Feb17:TestButton
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
#else
                char* Test;
                Test = websGetVar(wp, "DoTest","");

                if(!strcmp(Test,"yes")) {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Test_update_proc.html" , NULL);
                } else {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
                }
#endif

                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","98");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_is_PPTP.html");
                websFooter(wp);
#else
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_is_PPTP.html" , NULL);
#endif
            }
            break;

        case 40:        //FixIP
            if(strstr(page, "ipfixed")) {
#if 0               //Kevin:06Feb17:TestButton
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
#else
                char* Test;
                Test = websGetVar(wp, "DoTest","");

                if(!strcmp(Test,"yes")) {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Test_update_proc.html" , NULL);
                } else {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
                }

#endif
                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","98");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_is_ipfixed.html");
                websFooter(wp);
#else
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_is_ipfixed.html" , NULL);
#endif
            }
            break;

        case 50:        //BigPond
            if(strstr(page, "Telstra")) {
#if 0               //Kevin:06Feb17:TestButton
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
#else
                char* Test;
                Test = websGetVar(wp, "DoTest","");

                if(!strcmp(Test,"yes")) {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Test_update_proc.html" , NULL);
                } else {
                    websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/Blank_update_proc.html" , NULL);
                }
#endif
                if(!fork()) {
                    close(wp->fd);
                    nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","98");
                    if ( atoi(nvram_get("wizard_wan_spoofmac_detected")) ){
                        wizard_wan_spoofmac_detected = 1;
                        eth0_with_wan_spoofmac_up();//change wan0 mac to pc mac
                    }
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    /* g_pastWanType remember in sysInit */
                    sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_is_Telstra.html");
                websFooter(wp);
            }
            break;

        case 98:
            nvram_set("SmtWzdStep","99");
            nvram_set("SmtBTestWanConn","99");
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detwan.html");
            websFooter(wp);

            if(!fork()) {
                extern int h_errno;
                struct hostent *NameInfo;
                char UrlName[]={"WWW.NETGEAR.COM"};

                NameInfo = gethostbyname((char*)UrlName);
                if(h_errno) {
                    nvram_set("SmtWzdStep","100");
                    nvram_set("SmtBTestWanConn","100");
                } else {
                    nvram_set("SmtWzdStep","101");
                    nvram_set("SmtBTestWanConn","101");
                }
                exit(1);
            }
            break;
        case 99:
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detwan.html");
            websFooter(wp);
            break;
        case 100:
            nvram_set("SmtWzdStep","0");
            SetSmtWzdDone(1);
#if 1
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/basicform1.html" , NULL);
#else
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Error_Conn_Basic.html");
            websFooter(wp);
#endif
            break;
        case 101:
            nvram_set("SmtWzdStep","0");
            nvram_set("auto_connect","on");
            SetSmtWzdDone(1);
#if 1
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/basicform1.html" , NULL);
#else
            websWrite(wp,"<html><head>");
            websWrite(wp,"<script language=\"JavaScript\">");
            websWrite(wp,"location.href=\"/VER_sel.html\"");
            websWrite(wp,"</script>");
            websWrite(wp,"</head><body></body></html>");
#endif
            break;
        default:
            websWrite(wp,"<html><head><title></title><body>");
            websWrite(wp,"TraceWanDet.cgi error....");
            websWrite(wp,"current state is %d  ",curState);
            websWrite(wp,"current page is %s  ",page);
            websWrite(wp,"</body></html>");
            break;
    }

}

extern int g_bBlankStateLogin;

void
CA_cgi(webs_t wp, char_t *path, char_t **query)
{
    char *value;
    char *page;
       char strURL[] ="/Blank_detwan.html";
       char *path;
    char *chkLogout;

    int curState, cur_connType, past_connType = GetCurWanConnType();

    wizard_wan_spoofmac_detected = 0;
    init_form_cgi(wp, query);

    page = websGetVar(wp, "page", "");
    curState = atoi(nvram_get("SmtWzdStep"));
    chkLogout=websGetVar(wp, "logout", "");
    if(!strcmp(chkLogout,"LOGOUT")) {
        g_bBlankStateLogin=0;
        websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/logout.html" , NULL);
        return;
    }

    if(strstr(page,"basicsetting") ) {
        nvram_set("SmtWzdStep","0");
        SetSmtWzdDone(1);
        websHeader(wp);
           websWrite(wp, (char_t *) apply_refresh_header, 0, "/rel_index.html");
        websFooter(wp);
        return;
    }
    if(strstr(page,"ErrorAuth") ) {
        websHeader(wp);
           websWrite(wp, (char_t *) apply_refresh_header, 0, "/ErrorAuthCancel.html");
        websFooter(wp);
        return;
    }

    if(strstr(page,"wizard_select") ) {
        char *DetSel;
        DetSel = websGetVar(wp, "WANDetc","");
        if(!strcmp(DetSel,"AutoDetc")) {
                nvram_set("SmtWzdStep","1");
                curState=1;
        } else {
            nvram_set("SmtWzdStep","0");
            SetSmtWzdDone(1);
            websHeader(wp);
               websWrite(wp, (char_t *) apply_refresh_header, 0, "/basicform1.html");
            websFooter(wp);
            return;
        }
    }

    if(strstr(page,"HiddenPage")) {
        char *SmtWzdConfig;
        SmtWzdConfig=websGetVar(wp,"ConfigSmtWzd","");

        if(strstr(SmtWzdConfig,"Enable") ) {
            SetSmtWzdDone(0);
        } else {
            SetSmtWzdDone(1);
        }
        websWrite(wp,"<html><head>");
        websWrite(wp,"<script language=\"JavaScript\">");
        websWrite(wp,"top.location.href=\"http://www.routerlogin.com/\"");
        websWrite(wp,"</script>");
        websWrite(wp,"</head><body></body></html>");
        return;
    }
    if(!strcmp(page,"CA_Welcome.html")) {
        curState=1;
    }

    value = websGetVar(wp, "CA_Step", NULL);
    if(value != NULL) {
        curState = atoi(value);
        if(curState == 5) { curState = atoi(nvram_get("SmtBDetWanType")); }
    //  printf("@KKDEBUG@:%s: page is %s,value is %s, and SmtWzdState is %d\n",
    //      __FUNCTION__,
    //      page, value, curState);
    }

    switch(curState) {
        case 1:
            if(strstr(page,"Welcome") || strstr(page,"WAN_Retry")) {
                nvram_set("auto_connect","on");
                {   //Detected WAN port is plugin?

                    int RetVal, up, speed, duplex, range;
#ifdef CONFIG_WAN_LAN_MATCH_DETECTION
                    int CurWanLanFlag, PostWanLanFlag;
                    CurWanLanFlag = atoi(nvram_get("WL_ip_match"));
                    //PostWanLanFlag = atoi(websGetVar(wp,"MatchSubnet", NULL));
                    PostWanLanFlag = atoi(websGetVar(wp,"MatchSubnet", ""));
                    if(CurWanLanFlag && !PostWanLanFlag) {
                        //nvram_set("WL_ip_match","0");
                        Set_Flag_WAN_LAN_MATCH_DET(0);
                    }
#endif
                    RetVal = rtl8651_checkEthernetPortStatus(0, &up, &speed, &duplex, &range);
                    if(!up) {
                        websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_WAN_Retry.html" , NULL);
                        //websHeader(wp);
                           //websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_WAN_Retry.html");
                        //websFooter(wp);
                        return;
                    }
                }

                //Pop the detecting page
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Detecting_WAN_Connection.html");
                websFooter(wp);
#else
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Detecting_WAN_Connection.html" , NULL);

#endif
                nvram_set("SmtWzdStep","2");
                nvram_set("SmtBDetWanType","2");

                if(!fork()) {
                    //Run WAN detect wizard
                    int SmtWzdState=0;
                    int i;
                    //char macstr[17];

                    strcpy(spoof_mac, wizard_getclonemac(wp->fd));

                    close(wp->fd);  /* child no used */
                    upgrading = 1;
                    rtl8651a_setDosStatus(0,0);
		    pRomeCfgParam->pptpCfgParam.manualHangup = 1;                    
		    pRomeCfgParam->pppoeCfgParam[0].manualHangup = 1;
		    pRomeCfgParam->telCfgParam.manualHangup = 1;	                    
                    stop_wan();
                    sleep(2); 
                    {
                    	int tmp;
                    	
		    	pRomeCfgParam->pptpCfgParam.manualHangup = 0;                    
		    	pRomeCfgParam->pppoeCfgParam[0].manualHangup = 0;
		    	pRomeCfgParam->telCfgParam.manualHangup = 0;	                    	
		        tmp = pRomeCfgParam->ifCfgParam[0].connType;                    
                        nvram_set("wan_proto", "dhcp");
    		        start_wan(); 
                        rtl8651a_setDosStatus(0,0);    		    
    		        sleep(2); 
			pRomeCfgParam->ifCfgParam[0].connType = tmp;    		        
    		    }                      

                    SmtWzdState=wizard_go();
                    upgrading = 0;
                    nvram_set("wizard_wan_spoofmac_detected", "0");
                    switch(SmtWzdState) {
                        case 1: //DHCP
                            nvram_set("SmtWzdStep","10");
                            nvram_set("SmtBDetWanType","10");
                            break;
                        case 2: //PPPOE
                            nvram_set("SmtWzdStep","20");
                            nvram_set("SmtBDetWanType","20");
                            break;
                        case 3: //PPTP
                            nvram_set("SmtWzdStep","30");
                            nvram_set("SmtBDetWanType","30");
                            break;
                        case 4: //FixedIP
                            nvram_set("SmtWzdStep","40");
                            nvram_set("SmtBDetWanType","40");
                            break;
                        case 5: //BigPond
                            nvram_set("SmtWzdStep","50");
                            nvram_set("SmtBDetWanType","50");
                            break;
                        case 6: //spoofmac DHCP
                            nvram_set("SmtWzdStep","10");
                            nvram_set("SmtBDetWanType","10");
                            nvram_set("wizard_wan_spoofmac_detected", "1");
                            break;
                        case 7: //spoofmac Bigpond
                            nvram_set("SmtWzdStep","50");
                            nvram_set("SmtBDetWanType","50");
                            nvram_set("wizard_wan_spoofmac_detected", "1");
                            break;
                        default: //FixedIP
                            nvram_set("SmtWzdStep","40");
                            nvram_set("SmtBDetWanType","40");
                            break;
                    }
//                  nvram_set("SmtWzdStep","50");
                    exit(1);
                }
            } else {
                websWrite(wp,"<html><head><title></title><body>");
                websWrite(wp,"CA.cgi error....");
                websWrite(wp,"current state is %d  ",curState);
                websWrite(wp,"current page is %s  ",page);
                websWrite(wp,"</body></html>");
            }

            break;

        case 2:
            if(strstr(page,"Detecting")) {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Detecting_WAN_Connection.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Detecting_WAN_Connection.html" , NULL);
#endif
            } else {
                websWrite(wp,"<html><head><title></title><body>");
                websWrite(wp,"CA.cgi error....");
                websWrite(wp,"current state is %d  ",curState);
                websWrite(wp,"current page is %s  ",page);
                websWrite(wp,"</body></html>");
            }
            break;

        case 10:        //DHCP
            if(strstr(page, "Dhcp")) {
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Change_Admin_PassWord.html" , NULL);


                if(!fork()) {
                    close(wp->fd);      /* child no used */
                    //060705:kevin nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","70");
                    if ( atoi(nvram_get("wizard_wan_spoofmac_detected")) ){
                        wizard_wan_spoofmac_detected = 1;
                        eth0_with_wan_spoofmac_up();//change wan0 mac to pc mac
                    }
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_ConnType_Dhcp.html" , NULL);
//              websHeader(wp);
//                 websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_ConnType_Dhcp.html");
//              websFooter(wp);
            }
            break;


        case 20:        //PPPOE
            if(strstr(page, "PPPoE")) {
#if 0
                websHeader(wp);
                   //websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Update_Settings.html");
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Change_Admin_PassWord.html");
                websFooter(wp);
#endif
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Change_Admin_PassWord.html" , NULL);

                if(!fork()) {
                    close(wp->fd);      /* child no used */

                    //060705:kevin nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","70");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_ConnType_PPPoE.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_ConnType_PPPoE.html" , NULL);
#endif
            }
            break;

        case 30:        //PPTP
            if(strstr(page, "PPTP")) {
#if 0
                websHeader(wp);
                   //websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Update_Settings.html");
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Change_Admin_PassWord.html");
                websFooter(wp);
#endif
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Change_Admin_PassWord.html" , NULL);

                if(!fork()) {
                    close(wp->fd);      /* child no used */
                    //060705:kevin nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","70");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_ConnType_PPTP.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_ConnType_PPTP.html" , NULL);
#endif
            }
            break;

        case 40:        //FixIP
            if(strstr(page, "StaticIP")) {
#if 0
                websHeader(wp);
                   //websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Update_Settings.html");
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Change_Admin_PassWord.html");
                websFooter(wp);
#else
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Change_Admin_PassWord.html" , NULL);
#endif
                if(!fork()) {
                    close(wp->fd);      /* child no used */
                    //060705:kevin nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","70");
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_ConnType_StaticIP.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_ConnType_StaticIP.html" , NULL);
#endif
            }
            break;

        case 50:        //BigPond
            if(strstr(page, "Telstra")) {
#if 0
                websHeader(wp);
                   //websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Update_Settings.html");
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Change_Admin_PassWord.html");
                websFooter(wp);
#else
                /* modified by lucas */
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Change_Admin_PassWord.html" , NULL);
#endif
                if(!fork()) {
                    close(wp->fd);      /* child no used */
                    //060705:kevin nvram_set("SmtWzdDone","1");
                    nvram_set("SmtWzdStep","70");
                    if ( atoi(nvram_get("wizard_wan_spoofmac_detected")) ){
                        wizard_wan_spoofmac_detected = 1;
                        eth0_with_wan_spoofmac_up();//change wan0 mac to pc mac
                    }
                    validate_cgi(wp);
                    //sysReinit(past_connType);
                    //wizard_wan_restart(past_connType);
                    exit(1);
                }
            } else {
#if 0
                websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_ConnType_Telstra.html");
                websFooter(wp);
#else
                websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_ConnType_Telstra.html" , NULL);
#endif
            }
            break;

        case 70:
            if(strstr(page,"PassWord")) {
                int idx=0;
                char *OptionSel=websGetVar(wp, "OptionsSel", "");
                char *OptionSelHtml[] ={ "/www/CA_Summary_Page.html",
                                        "/www/CA_Change_Admin_PassWord_Input.html"
                                        };
                char *ptr;

                ptr=websGetVar(wp, "passwd", "");
                nvram_set("http_passwd",ptr);
                if( !strstr(page, "Input") && !strstr(OptionSel,"default")) {
                    idx=1;
                }
#if 0
            websHeader(wp);
                   websWrite(wp, (char_t *) apply_refresh_header, 0, OptionSelHtml[idx]);
            websFooter(wp);
#else
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, OptionSelHtml[idx] , NULL);
#endif
            return;
            }
#if 0
            websHeader(wp);
                websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Update_Settings.html");
            websFooter(wp);
#else
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Update_Settings.html" , NULL);
#endif
            if(!fork()) {
                close(wp->fd);
                //mgmt_cfg_save();
                cur_connType = GetCurWanConnType();
                nvram_set("SmtWzdDone","1"); //060705:kevin
                nvram_set("SmtWzdStep","98");
                WanConfigSaveToFlash(cur_connType);
                //sysReinit(past_connType);
                /* g_pastWanType remember in sysInit */
                sysReinit(g_pastWanType);
//          nvram_set("SmtWzdStep","98");
                exit(1);
            }
            break;
        case 98:
            {
                extern int h_errno;
                struct hostent *NameInfo;
                char UrlName[]={"WWW.NETGEAR.COM"};

                NameInfo = gethostbyname((char*)UrlName);
                if(h_errno) {
                    sleep(1);
                    NameInfo = gethostbyname((char*)UrlName);
                }
                if(h_errno) {
                    nvram_set("SmtWzdStep","1");
                    SetSmtWzdDone(0);
#if 0
                    websHeader(wp);
                    websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Failure.html");
                    websFooter(wp);
#else
            websDefaultHandler(wp, NULL, NULL, NULL, NULL, "/www/CA_Failure.html" , NULL);
#endif
                } else {
                    nvram_set("SmtWzdStep","0");
                    nvram_set("auto_connect","on");
                    SetSmtWzdDone(1);
                    websWrite(wp,"<html><head>");
                    websWrite(wp,"<script language=\"JavaScript\">");
                    websWrite(wp,"top.location.href=\"http://www.netgear.com/success/rp614v4.php\"");
                    websWrite(wp,"</script>");
                    websWrite(wp,"</head><body></body></html>");
                }
            }
            break;
/*      case 99:
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/Blank_detwan.html");
            websFooter(wp);
            break; */
        case 100:
            nvram_set("SmtWzdStep","1");
            SetSmtWzdDone(0);
            websHeader(wp);
            websWrite(wp, (char_t *) apply_refresh_header, 0, "/CA_Failure.html");
            websFooter(wp);
            break;
        case 101:
            nvram_set("SmtWzdStep","0");
            SetSmtWzdDone(1);
            websWrite(wp,"<html><head>");
            websWrite(wp,"<script language=\"JavaScript\">");
            websWrite(wp,"top.location.href=\"http://WWW.NETGEAR.COM/\"");
            websWrite(wp,"</script>");
            websWrite(wp,"</head><body></body></html>");
            break;
        default:
            websWrite(wp,"<html><head><title></title><body>");
            websWrite(wp,"CA.cgi error....");
            websWrite(wp,"current state is %d  ",curState);
            websWrite(wp,"current page is %s  ",page);
            websWrite(wp,"</body></html>");
            break;
    }

}

/* iptables setting */
#define IPTABLES    "/bin/iptables"
#define LAN_IFNAME  "eth1"
#define WAN_IFNAME  "eth0"
int firewall_starting(void)
{
#ifdef N_DNS_HIJACK     /* PR only */
    return 0;

#else
    char buf[200];
    char *ptr;
    char wan_ifname[10]=WAN_IFNAME;

    /* clear iptable */
    sprintf(buf, "%s -F", IPTABLES);
    system(buf);
    sprintf(buf, "%s -X", IPTABLES);
    system(buf);
    sprintf(buf, "%s -Z", IPTABLES);
    system(buf);

#if defined(DUALNAT)
    sprintf(buf, "%s -t nat -F", IPTABLES);
    system(buf);
    sprintf(buf, "%s -t nat -X", IPTABLES);
    system(buf);
    sprintf(buf, "%s -t nat -Z", IPTABLES);
    system(buf);
#endif

    strcpy(wan_ifname, get_wan_ifname());
    /* # Inbound defaults */
    sprintf(buf, "%s -P INPUT DROP", IPTABLES);
    system(buf);

    sprintf(buf, "%s -A INPUT -i %s -m state --state NEW -j ACCEPT", IPTABLES, LAN_IFNAME);
    system(buf);
    sprintf(buf, "%s -A INPUT -m state --state ESTABLISHED,RELATED -j ACCEPT", IPTABLES);
    system(buf);

    ptr = nvram_get("wan_ping");
    if (ptr && (ptr[0] == '1')) {
        sprintf(buf, "%s -A INPUT -i %s -p icmp --icmp-type 8 -j ACCEPT", IPTABLES, wan_ifname);
        system(buf);
    }

    ptr = nvram_get("fw_remoteenable");
    if (ptr && (ptr[0] == '1')) {
        sprintf(buf, "%s -A INPUT -i %s -p tcp --dport 80 -j ACCEPT", IPTABLES, wan_ifname);
        system(buf);
    }

#if 1
    ptr = nvram_get("wan_proto");
    if (strcmp(ptr, "dhcp") == 0) {
        sprintf(buf, "%s -A INPUT -i %s -p udp --sport 67 --dport 68 -j ACCEPT", IPTABLES, wan_ifname);
        system(buf);
    }
#endif

#if defined(DUALNAT)
    ptr = nvram_get("wan_proto");
    if ( (strcmp(ptr, "pptp") == 0)/* && (strcmp(nvram_get("wan_pptp_mode"), "1") == 0)*/) {
        sprintf(buf, "%s -t nat -D POSTROUTING -o %s -j MASQUERADE", IPTABLES, WAN_IFNAME);
        system(buf);
        sprintf(buf, "%s -t nat -A POSTROUTING -o %s -j MASQUERADE", IPTABLES, WAN_IFNAME);
        system(buf);

        FILE *f=fopen("/proc/sys/net/ipv4/ip_forward","w+");
        if (f) {
            fprintf(f,"%d",1);
            fclose(f);
        }
    }
    else {
        FILE *f=fopen("/proc/sys/net/ipv4/ip_forward","w+");
        if (f) {
            fprintf(f,"%d",0);
            fclose(f);
        }
    }

#endif

    sprintf(buf, "%s -A INPUT -m state --state INVALID -j DROP", IPTABLES);
    system(buf);

    /* # Inbound defaults */
    sprintf(buf, "%s -P INPUT DROP", IPTABLES);
    system(buf);
    return 0;
#endif
}


